// CSBuildDoc.h : interface of the CCSBuildDoc class
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_CSBUILDDOC_H__7F7EED2A_C6A0_11D6_A264_8CD155B1182C__INCLUDED_)
#define AFX_CSBUILDDOC_H__7F7EED2A_C6A0_11D6_A264_8CD155B1182C__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class SKIN_CACHE
{
  i32  m_level;           // -1 if cache is empty;
  i8   m_columns[16][64]; //Records from Expool.
  i8   m_size[16];        //Record sizes; (-1 need to look) (0 no record exists)
  i32  Load(i32 level, i32 x);
public:
  SKIN_CACHE(void){m_level=-1;};
  ~SKIN_CACHE(void){};
  ui8  GetSkin(i32 level, i32 x, i32 y);
  ui8  GetSkin(const LOCATIONREL& locr);
  ui8  GetDefaultSkin(i32 level);
  void SetSkin(i32 level, i32 x, i32 y, i8 skinNum);
  void SetDefaultSkin(i32 level, ui8 skinNum);
};

extern SKIN_CACHE skinCache;

class EXTRACTED_LEVEL
{
  friend class CCSBuildDoc;
  int       m_origLevel;
  ui16      m_DSALevelIndex[32];
  LEVELDESC m_levelDescriptor;
  int       m_cellflagSize; //Number of bytes
                            //Including decorations, monsters, etc.
  CELLFLAG *m_pCellflags;
  int       m_contentsSize; //Number of RN entries.
  RN       *m_pContents;
public:
  EXTRACTED_LEVEL(void);
  ~EXTRACTED_LEVEL(void);
};

class OBJECT_CHECKPOINT
{
  friend class CELL_CHECKPOINT;
  OBJECT_CHECKPOINT *m_nextObject;
  OBJECT_CHECKPOINT *m_firstPossession;
  DBTYPE             m_dbType;
  ACTUATORCLASS      m_acClass;
  unsigned char      m_position;
  unsigned char      m_objectData[16]; //room for monster entry
  ui16              *m_text;
  void ConstructDB0(RN obj, ROOMTYPE rt);
  void ConstructDB1(RN obj, ROOMTYPE rt);
  void ConstructDB2(RN obj, ROOMTYPE rt);
  void ConstructDB3(RN obj, LOCATIONREL *plocr, ROOMTYPE rt);
  void ConstructDB4(RN obj, ROOMTYPE rt);
  void ConstructDB5(RN obj, ROOMTYPE rt);
  void ConstructDB6(RN obj, ROOMTYPE rt);
  void ConstructDB7(RN obj, ROOMTYPE rt);
  void ConstructDB8(RN obj, ROOMTYPE rt);
  void ConstructDB9(RN obj, ROOMTYPE rt);
  void ConstructDB10(RN obj, ROOMTYPE rt);
  void ConstructDB14(RN obj, ROOMTYPE rt);
  void ConstructDB15(RN obj, ROOMTYPE rt);
  RN RebuildDB0(RN obj=RNnul)const;
  RN RebuildDB1(RN obj=RNnul)const;
  RN RebuildDB2(RN obj=RNnul)const;
  RN RebuildDB3(RN obj=RNnul)const;
  RN RebuildDB4(RN obj=RNnul)const;
  RN RebuildDB5(RN obj=RNnul)const;
  RN RebuildDB6(RN obj=RNnul)const;
  RN RebuildDB7(RN obj=RNnul)const;
  RN RebuildDB8(RN obj=RNnul)const;
  RN RebuildDB9(RN obj=RNnul)const;
  RN RebuildDB10(RN obj=RNnul)const;
  RN RebuildDB14(RN obj=RNnul)const;
  RN RebuildDB15(RN obj=RNnul)const;
//  bool NotEqualDB0(RN obj)const;
//  bool NotEqualDB1(RN obj)const;
//  bool NotEqualDB2(RN obj)const;
//  bool NotEqualDB3(LOCATIONREL *plocr,RN obj)const;
//  bool NotEqualDB4(RN obj)const;
//  bool NotEqualDB5(RN obj)const;
//  bool NotEqualDB6(RN obj)const;
//  bool NotEqualDB7(RN obj)const;
//  bool NotEqualDB8(RN obj)const;
//  bool NotEqualDB9(RN obj)const;
//  bool NotEqualDB10(RN obj)const;
//  bool NotEqualDB14(RN obj)const;
//  bool NotEqualDB15(RN obj)const;
  void AddToPossessions(RN obj, ROOMTYPE rt); // A list of objects
  void RebuildPossessions(RN *pRN) const;
public:
  OBJECT_CHECKPOINT(RN obj, ROOMTYPE rt, LOCATIONREL *plocr);
  ~OBJECT_CHECKPOINT(void);
  void Dump(void);
  bool operator ==(const OBJECT_CHECKPOINT &oc) const;
  DBTYPE dbType(void)const{return m_dbType;};
  RN RebuildObject(RN obj=RNnul);
  int Position(void){return m_position;};
};

#define MAXCELLPERCHECKPOINT 2

enum CELL_CHECKPOINT_FLAGS
{
  CCF_NextOneGoesWithThisOne = 0x001,
  CCF_PrevOneGoesWithThisOne = 0x002
};

class CELL_CHECKPOINT
{
  friend class CELL_CHECKPOINT_LIST;
  CELL_CHECKPOINT   *m_nextCell;
  CELL_CHECKPOINT   *m_prevCell;
  OBJECT_CHECKPOINT *m_firstObject; 
  LOCATIONREL        m_locr;
  CELLFLAG           m_cellFlag;
  unsigned int       m_flags;
public:
  CELL_CHECKPOINT(void);
  ~CELL_CHECKPOINT(void);
  bool operator ==(const CELL_CHECKPOINT& cc) const;
  CELLFLAG CellFlag(void){return m_cellFlag;};
  void CellFlag(CELLFLAG cf){m_cellFlag = cf;};
  void Location(LOCATIONREL locr){m_locr = locr;};
  LOCATIONREL Location(void){return m_locr;};
  void SetFlag(CELL_CHECKPOINT_FLAGS f);
  bool Flags(CELL_CHECKPOINT_FLAGS f);
  void AddObject(OBJECT_CHECKPOINT *pOC);
  OBJECT_CHECKPOINT *FirstObject(void) const {return m_firstObject;};
  OBJECT_CHECKPOINT *NextObject(const OBJECT_CHECKPOINT *pOC)const
  {
    return pOC->m_nextObject;
  };
  void Dump(void);
};

class CELL_CHECKPOINT_LIST
{
  CELL_CHECKPOINT *m_firstCell;
  CELL_CHECKPOINT *m_lastCell;
  CELL_CHECKPOINT *m_currentCursor;
  unsigned int     m_maxLength;
  unsigned int     m_curLength;
public:
  CELL_CHECKPOINT_LIST(void);
  ~CELL_CHECKPOINT_LIST(void);
  void WriteAtCursor(CELL_CHECKPOINT *pCC);
  CELL_CHECKPOINT *Replace(CELL_CHECKPOINT *pCCnew);
  CELL_CHECKPOINT *ReadCellAtCursor(void);
  bool IsRedoAvailable(void);
  void Clear(void);
  bool Backspace(void);
  CELL_CHECKPOINT *ForwardSpace(void);
};

enum EDIT_OPT
{
  //Stairways during editing.  Does not affect 'Check'|'Stairs'
  EDIT_OPT_AllowIllegalStairs         = 0x00000001,

  //Pits during editing.  Does not affect 'Check'|'Pits'
  EDIT_OPT_AllowIllegalPits           = 0x00000002,

  //The "edge-triggered" and/or gate options
  EDIT_OPT_AllowEnhancedAndOr         = 0x00000004,

  //Allow "unimporant objects"
  EDIT_OPT_AllowUnimportantObjects    = 0x00000008,

  //Allow Orphaned Objects without warning
  EDIT_OPT_AllowOrphanedObjects       = 0x00000010,

  //Allow Cloned (more than one reference to) Objects
  //This is for viewing only..We will never CREATE such 
  //a thing intentionally!
  EDIT_OPT_AllowClonedObjects         = 0x00000020,

  //Allow objects like Apples in stairwells
  EDIT_OPT_AllowObjectsOnStairs       = 0x00000040,

  //Allow Pew, Ra, and Kath Bombs
  EDIT_OPT_AllowPewRaKath             = 0x00000080,

  //Some objects have different names in DM.
  EDIT_OPT_UseDMObjectNames           = 0x00000100,

  //and some collections of OPTIONS
  EDIT_OPT_SafeOptions           = 0x00000010,
  EDIT_OPT_DefaultOptions        = 0x0000000c,
  EDIT_OPT_ExpertOptions         = 0x000000df
};


struct EXTENDEDFEATURESBLOCK
{
  char sentinel[20];
  int           dataMapLength;
  unsigned int  dataTypeMapChecksum;
  unsigned int  dataIndexMapChecksum;
  unsigned int  extendedFeaturesChecksum;
  char          version;
  char          flags;
enum FLAGS
{
  LevelDSAInfoPresent  = 0x01,
  EditingOptionsPresent= 0x02,
  SimpleEncryption     = 0x04,
  IndirectText         = 0x08,
  AutoRecord           = 0x10,
  ExpandedPortraits    = 0x20,
  DMRules              = 0x40,
  BigActuators         = 0x80
};
  ui16          numDSA;
  unsigned int  editingOptions;
  unsigned int  gameInfoSize;
  unsigned int  cellFlagArraySize;
  unsigned int  graphicsSignature1;
  unsigned int  graphicsSignature2;
  unsigned int  spellFilterLocation;
                    //Bit 31 set if location is valid.
  ui32      extendedFlags; //Why did I think 8 would suffice?
enum EXTENDEDFLAGS
{
  OverlayActive       = 0x00000001,
  InvisibleMonsters   = 0x00000002,
  DrawAsSize4Monsters = 0x00000004
};
  i32       overlayOrdinal; //overlay Number + 1
  i32       overlayP1;
  i32       overlayP2;
  i32       overlayP3;
  i32       overlayP4;
  ui32      CSBgraphicsSignature1;
  ui32      CSBgraphicsSignature2;
  char      hintKey[8];
  char fill[512-104];
  EXTENDEDFEATURESBLOCK(void);
  ~EXTENDEDFEATURESBLOCK(void);
};



class SEARCHPKT  // To search the dungeon for objects
{
public:

  SEARCHPKT(void){object = RNnul;};
  unsigned int mask; //Example: (1<<dbTELEPORTER) | (1<<dbMISC)
  int level;
  int col;
  int row;
  RN  object; //The most recently found. RNnul=start RNeof=end
};


class CCSBuildDoc : public CDocument
{
protected: // create from serialization only
	CCSBuildDoc();
	DECLARE_DYNCREATE(CCSBuildDoc)
  void SetModifiedFlag(void);
  BOOL DoSave(LPCTSTR lpszPathName, BOOL bReplace);

// Attributes
private:
  bool   m_dungeonFormat;
  bool   m_compressed;
  bool   m_indirectText;
  bool   m_bigActuators;
  unsigned char m_saveExpansionTable[20];
  char  *m_expandedBuffer; //For reading dungeon-formatted files
  int    m_expandedBufferSize;
  int    m_expandedBufferIndex;
  int    m_inStreamLeft;
  char  *m_inStreamNext;
  bool   m_validDungeon;
  RN     m_ExtendedFeaturesText;
  CELL_CHECKPOINT_LIST m_cellUndoList;
  GAMEBLOCK1 m_block1; // 512 bytes
  GAMEBLOCK2 m_block2; // 128 bytes
  ITEM16 *m_Item16;
  i16    m_MaxITEM16;
  DSA    *m_DSAIndex[256];
  ui16   m_DSALevelIndex[64][32];
  unsigned int m_editingOptions;
  unsigned int m_cellflagArraySize;
  //
  // ************* NOTE WELL !! *************
  // 3328 bytes read into this part of the structure.
  // This includes the four characters and 128 additional
  // bytes of global data.
  CHARDESC m_Characters[4];
  i16 brightness;
  i8  SeeThruWalls;
  i8  Byte13279;   
  i16 Word13278;       //Party shield
  i16 Word13276;       //Spell effect
  i16 Word13274;       //Spell effect
  i8  NumHistEnt;      // #entries in 13268???
  ui8 uByte13271;      //Life frozen time
  i8  Byte13270;       //index of entry in 13268
  i8  Byte13269;       //index of entry in 13268
  i16 Word13268[24];   //history of moves???
                       // bits 0-4 = mapX
                       // bits 5_9 = mapY
                       // bits 10_15 = loaded level
  i8  Byte13220[24];   //parallels 13268??  
                       // Is this a once-only kind of thing?
  i8  Invisible;       // 13196;
  i8  Byte13195[41];
  /////////////  End of Character portion of save file /////////////
  //  ***************************************************
  //
  TIMER           *m_Timers;
  ui16            *m_timerQue;
  DUNGEONDATINDEX *m_dungeonDatIndex;
  LEVELDESC       *m_pLevelDescriptors;
  ui16            *m_objectLevelIndex; //offset in objectColumnIndex
                                       // for first column in
                                       //  each level
  ui16            *m_objectColumnIndex;//offset in objectIndex for
                                       //each column in dungeon.
  RN              *m_objectList;
  CELLFLAG        *m_cellFlagArray;
  int             *m_indirectTextIndex; //ui16 offsets in compressedText
  int              m_sizeOfCompressedText;
  ui16            *m_compressedText;
  ui8             *m_dataTypeMap;
  ui16            *m_dataIndexMap;
  int              m_dataMapLength;
  CELLFLAG ***m_pppdPointer10450;
  i16    m_MaxTimers;
  ui32   m_totalMoveCount;
  ui32   m_RandomGameID;
  ui8    m_Byte22808[132];
  ui32   m_Time;
  ui32   m_Long11732;
  ui32   m_RandomNumber;
  i16    m_NumCharacter;
  i16    m_partyX;
  i16    m_partyY;
  i16    m_partyFacing;
  i16    m_partyLevel;
  i16    m_HandChar;
  ui16   m_numTimer;
  i16    m_NextAvailTimer;
  i16    m_ITEM16QueLen;
  i16    m_MagicCaster;
  ui32   m_Long12950;
  ui32   m_Long12954;
  ui16   m_Word13268[24];
  ui16   m_Word11710;
  ui16   m_Word11712;
  ui16   m_Word11714;
  //ui16   m_Brightness;
  RN     m_objectInHand;
  ui16   m_Word22592;
  int    m_numColumnPointers;
  ui16    m_DBEntrySize[16];
  LOCATIONREL m_spellFilterLocation;
public:
  int    m_currentDisplayLevel;
  bool   m_littleEndian;
  bool   m_targetMode;
  bool   m_PCFormattedBlock1;

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CCSBuildDoc)
	public:
	virtual BOOL OnNewDocument();
	virtual void Serialize(CArchive& ar);
	virtual BOOL OnSaveDocument(LPCTSTR lpszPathName);
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~CCSBuildDoc();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

void Sorry(const char *msg);
void Warning(const char *msg);
void Information(const char *msg);
bool DoYouReallyWantTo(const char *msg);
void FreeAllMemory(void);
bool LoadDungeon(CArchive *ar);
void MakeBigActuators(void);
void InitializeDungeonDatIndex(void);
unsigned int EditingOptions(void);
void EditingOptions(unsigned int);
//*******************************
// A few functions to "Write/Read w/Check Sum" (WCS/RCS)
unsigned int WCS(CArchive& ar);
unsigned int RCS(CFile *f);
void WCS(char *, int count=1);
void RCS(char *, int count=1);
void WCS(ui16 *, int count=1);
void RCS(ui16 *, int count=1);
void WCS(ui8 *, int count=1);
void RCS(ui8 *, int count=1);
void WCS(int *, int count=1);
void RCS(int *, int count=1);
void WCS(unsigned int *, int count=1);
void RCS(unsigned int *, int count=1);
unsigned int m_WCSChecksum;
CArchive *m_pWCSar;
CFile *m_pRCSCFile;
//*******************************
bool Compressed(void){return m_compressed;};
void Compressed(bool b){m_compressed=b;};
bool DungeonFormat(void){return m_dungeonFormat;};
void DungeonFormat(bool b){m_dungeonFormat=b;};
void ExpandData(char *src,char *dst,int dstSize);
bool AddLevel(void);
bool AddColumn(int level);
bool AddRow(int level);
bool VerifyLevelData(void);
EXTRACTED_LEVEL *RemoveLevel(int level);
bool InsertLevel(int level, EXTRACTED_LEVEL *pExtLev);
unsigned int CellFlagArraySize(void){return m_cellflagArraySize;};
void CellFlagArraySize(unsigned int n){MODIFY m_cellflagArraySize = n;};
void RecomputePppd10450(void);
int CurrentDisplayLevel(void){return m_currentDisplayLevel;};
STORESTATUS StoreDungeon(CArchive& ar);
STORESTATUS WriteAndChecksum(
                     CArchive& ar,
                     ui8 *buf,
                     ui16 *pRunningChecksum, //updated
                     i32 numByte);
LOADSTATUS ReadAndUnscrambleBlockLittleEndian(CFile *f, ui16 *buf);
LOADSTATUS ReadAndUnscrambleBlockBigEndian(CFile *f, ui16 *buf);
ui16 Unscramble(ui16 *buf, ui16 initialHash, i32 numword);
LOADSTATUS UnscrambleStream(
                  CFile *f,
                  ui16 *dest,        //data destination
                  i32 size,          //# of words
                  ui16 initialHash,   
                  ui16 Checksum);    //expected checksum
void swapBlock1(GAMEBLOCK1 *block1);
void swapPCBlock1(PCGAMEBLOCK1 *pBlock);
void swapBlock2(GAMEBLOCK2 *block2);
void extractBlock1Data(GAMEBLOCK1 *pBlock);
void extractPCBlock1Data(PCGAMEBLOCK1 *pBlock);
void initializeBlock1Data(void);
void extractBlock2Data(void);
void initializeBlock2Data(void);
void InitializeItem16(void);
void InitializeTimers(void);
void swapTimers(void);
void DeleteMonsterMovementTimers(LOCATIONREL locr);
void swapTimerQue(void);
void swapITEM16s(void);
void DetachItem16(i32 P1);
void DeleteItem16Entries(void);
void swapLevelDescriptors(int numLevel);
void swapCharacter(int i);
void swapCharacterData(void);
void InitializeCharacters(void);
void translatePCcharacters(void);
void swapTextArray(ui16 *text);
void swapDungeonDatIndex(void);
void swapPointer10454(void);
void swapObjectList(i32 num);
int NumLevel(void) {return m_dungeonDatIndex->NumLevel();};
void NumLevel(int n) {m_dungeonDatIndex->NumLevel(n);};
int PartyX(void){return m_partyX;};
void PartyX(int x);
int PartyY(void){return m_partyY;};
void PartyY(int y);
int PartyFacing(void){return m_partyFacing;};
void PartyFacing(int facing);
int PartyLevel(void){return m_partyLevel;};
void PartyLevel(int level);
LOCATIONREL SpellFilterLocation(void) const;
void SpellFilterLocation(const LOCATIONREL& locr);
LOCATIONREL MonsterAttackFilterLocation(void) const;
void MonsterAttackFilterLocation(const LOCATIONREL& locr);
LOCATIONREL PartyAttackFilterLocation(void) const;
void PartyAttackFilterLocation(const LOCATIONREL& locr);
LOCATIONREL FeedingFilterLocation(void) const;
void FeedingFilterLocation(const LOCATIONREL& locr);
LOCATIONREL AdjustSkillsFilterLocation(void) const;
void AdjustSkillsFilterLocation(const LOCATIONREL& locr);
LOCATIONREL DeleteMonsterFilterLocation(void) const;
void DeleteMonsterFilterLocation(const LOCATIONREL& locr);
LOCATIONREL IndirectActionLocation(void) const;
void IndirectActionLocation(const LOCATIONREL& locr);
LOCATIONREL ViewingFilterLocation(void) const;
void ViewingFilterLocation(const LOCATIONREL& locr);
LOCATIONREL SoundFilterLocation(void) const;
void SoundFilterLocation(const LOCATIONREL& locr);
LOCATIONREL MonsterMoveFilterLocation(bool *sameLevel, unsigned int *maxDistance) const;
void MonsterMoveFilterLocation(const LOCATIONREL& locr, bool sameLevel, unsigned int maxDistance);
LOCATIONREL MonsterMoveFilterLocation(bool *sameLevel, unsigned int *maxDistance, int level) const;
void MonsterMoveFilterLocation(const LOCATIONREL& locr, bool sameLevel, unsigned int maxDistance, int level);
int StartingX(void){return m_dungeonDatIndex->StartingPartyX();};
void StartingX(int x)
{
  if (x != m_dungeonDatIndex->StartingPartyX())
  {
    m_dungeonDatIndex->StartingPartyX(x);
    SetModifiedFlag();
  };
};
int StartingY(void){return m_dungeonDatIndex->StartingPartyY();};
void StartingY(int y)
{
  if (y != m_dungeonDatIndex->StartingPartyY())
  {
    m_dungeonDatIndex->StartingPartyY(y);
    SetModifiedFlag();
  };
};
int StartingFacing(void){return m_dungeonDatIndex->StartingPartyFacing();};
void StartingFacing(int f)
{
  if (f != m_dungeonDatIndex->StartingPartyFacing())
  {
    m_dungeonDatIndex->StartingPartyFacing(f);
    SetModifiedFlag();
  };
};
i16 Brightness(void) {return brightness;};
void Brightness(i16 b) {brightness = b;};
//int  IsHintable(void);
//void IsHintable(bool hintable);
DSA *getDSA(int index){return m_DSAIndex[index];};
ui16 getDSA(int level, int index)
{
  return m_DSALevelIndex[level][index];
};
void setDSA(int index, DSA *pDSA);
void setDSA(int level, int index, ui16 dsaIndex);
ui16 compressDSAs(void);
void WriteDSAs(CArchive& ar);
LOADSTATUS ReadDSAs(CFile *f, int numDSA);
void ClearDSALevelIndex(void);
void ReadDSALevelIndex(CFile *f);
void WriteDSALevelIndex(CArchive& ar);
ui8 DefaultSkin(i32 level);
void DefaultSkin(i32 level, i32 skinNum);
int ExperienceMultiplier(int level)
{
  return m_pLevelDescriptors[level].experienceMultiplier();
};
void ExperienceMultiplier(int level, int em)
{
  if (em!=m_pLevelDescriptors[level].experienceMultiplier())
  {
    m_pLevelDescriptors[level].experienceMultiplier(em);
    SetModifiedFlag();
  };
}
void VerifyListOfObjects(RN *pRN, int level, int x, int y, int depth);
void VerifyDatabases(void);
void ConvertListOfObjects(RN *pRN, bool DuplicateOK,
                          int level, int x, int y);
void ConvertDataBases(void);
void convertScrolls();
void ConvertCharacters(void);
void ConvertTimers(void);
void ExtendPortraits(void);
int  ReadExtendedFeatures(CFile *f);
void WriteExtendedFeatures(CArchive& ar);
LOADSTATUS ReadDatabases(CFile *f);
LOADSTATUS readNewScrollDatabase(CFile *f, i16 *pCheckSum);
LOADSTATUS readOldScrollDatabase(CFile *f, i16 *pCheckSum);
LOADSTATUS readScrollDatabase(CFile *f, i16 *pCheckSum);
LOADSTATUS FetchDataBytes(
                CFile *f, ui8 *buf,i16 *checksum, i32 numByte);
LOADSTATUS StreamInput(CFile *f, ui8 *buf, int numByte);
STORESTATUS WriteDataFile(CArchive& ar,
                          ui8 *pwbuf,
                          i32 length);
i16 GenChecksum(i16 *buf, int initSum, i32 numWord);
STORESTATUS ScrambleAndWrite(CArchive& ar, ui16 *pwbuf);
STORESTATUS WriteScrambled(
              CArchive& ar,
              ui16 *buf,
              i32 byteCount,
              i16 initialHash,
              ui16 *checksum);

int Time(void){return m_Time;};
int NumTimers(void){return m_numTimer;};
int MaxTimers(void){return m_MaxTimers;};
TIMER *Timer(int index);
ui16   TimerQueue(int index);
void SetDBSize(int dbNum, int size);
int GetDBSize(int dbNum);
int LastRow(int level);
int LastColumn(int level);
int OffsetX(int level);
void OffsetX(int level, int v);
int OffsetY(int level);
void OffsetY(int level, int v);
int CellOffset(int level);
int NumMonsterType(int level);
int NumDoorTypes(void){return 4;};
int NumWallDecorations(int level);
int NumRandomWallDecorations(int level);
void NumRandomWallDecorations(int level, int num);
int NumFloorDecorations(int level);
int NumDoorDecorations(int level);
int NumRandomFloorDecorations(int level);
void NumRandomFloorDecorations(int level, int num);
void RotateDecoration(int level, int offset, int index);
WALLTYPE WallDecorationType(int level, int index);
bool IsValidWallDecorationOrdinal(int level, int ordinal);
bool IsLegalWallDecorationIndex(int level, int ordinal);
void AddWallDecoration(int level, WALLTYPE wt);
void ReplaceWallDecoration(int level, WALLTYPE wt, int index);
MONSTERTYPE MonsterType(int level, int index);
bool IsValidMonsterType(int level, MONSTERTYPE mt);
void AddMonsterType(int level, MONSTERTYPE mt);
void ReplaceMonsterType(int level, MONSTERTYPE mt, int index);
FLOORTYPE FloorDecorationType(int level, int ordinal);
bool IsLegalFloorDecorationIndex(int level, int ordinal);
bool IsValidFloorDecorationOrdinal(int level, int ordinal);
FLOORTYPE FloorDecorationType(LEVELINFO *levelInfo, int ordinal);
void AddFloorDecoration(int level, FLOORTYPE ft);
void ReplaceFloorDecoration(int level, FLOORTYPE ft, int ordinal);
ORNATETYPE DoorDecorationType(int level, int index);
void AddDoorDecoration(int level, ORNATETYPE ot);
void ReplaceDoorDecoration(int level, ORNATETYPE ot, int index);
int DoorType(int level, int index);
void DoorType(int level, int index, int type);
bool IsValid(void) {return m_validDungeon;};
void MarkInvalid(void){m_validDungeon = false;};
void MarkValid(void){m_validDungeon = true;};
CELLFLAG GetCellFlag(int level, int x, int y);
CELLFLAG GetCellFlag(LOCATIONREL locr);
CELLFLAG GetCellFlag(LOCATIONABS *loca);
void SetCellFlag(int level, int x, int y, CELLFLAG cf);
void SetCellFlag(LOCATIONREL locr, CELLFLAG cf);
void SetCellFlag(LOCATIONABS *loca, CELLFLAG cf);
bool GetExtendedCellFlag(LOCATIONREL locr, int bit);
void SetExtendedCellFlag(LOCATIONREL locr, int bit, bool b);
void InsertByteIntoCellflagArray(int index, ui8 value);
void SetRoomType(int level, int col, int row, ROOMTYPE rt);
void SetRoomType(LOCATIONREL locr, ROOMTYPE rt);
ROOMTYPE GetRoomType(LOCATIONREL locr){return RoomType(GetCellFlag(locr));};
ROOMTYPE GetLogicalRoomType(LOCATIONREL locr);
int *DetermineActuatorLocations(void);
ROOMTYPE *CCSBuildDoc::DetermineActuatorRoomTypes(void);
void DetermineActuatorRoomTypesList(RN, ROOMTYPE, ROOMTYPE *result);
void SetStairsUp(LOCATIONREL locr, bool up);
void SetStairsDirection(LOCATIONREL locr, int direction);
void SetFalseWallOpen(LOCATIONREL locr, bool open);
void SetFalseWallVisible(LOCATIONREL locr, bool visible);
void SetDoorState(LOCATIONREL locr, unsigned int s);
void SetDoorNS(LOCATIONREL locr, bool NS);
void SetTeleporterActive(LOCATIONREL locr, bool active);
DBTYPE DataType    (int indirectIndex);
ui8    DataPosition(int indirectIndex);
void   DataPosition(int indirectIndex, int pos);
void   ClearReferenceAndCloneFlags(void);
bool   IsIndirectIndexReferenced(int indirectIndex) const;
bool   IsIndirectIndexCloned(int indirectIndex) const;
void   MarkIndirectIndexReferenced(int indirectIndex);
void   MarkIndirectIndexCloned(int indirectIndex);
ui16   DataDBIndex (int indirectIndex);   
int    dataMapLength(void){return m_dataMapLength;};
ui16   AllocateIndirectIndex(DBTYPE dbType, int dbIndex);
void   ReleaseObjectIndirectIndex(int objectIndirectIndex);
void   DumpLevelInfo(FILE *f);
void   DumpObjectIndirectIndexTable(FILE *f);
void   DumpDatabases(FILE *f);
void   DumpRawText(FILE *f);
RN FindFirstObject(int level, int x, int y) const;
RN FindFirstObject(const LOCATIONREL locr) const;
RN FindFirstActuator(int level, int column, int row);
RN FindFirstActuator(const LOCATIONREL locr);
RN FindNextActuator(RN obj);
RN FindFirstMonster(const LOCATIONREL locr);
RN FindNextMonster(RN obj);
RN FindFirstTeleporter(int level, int column, int row);
RN FindNextTeleporter(RN obj);
RN FindFirstTeleporter(LOCATIONREL locr);
RN FindFirstTeleporter(LOCATIONABS *loca);
RN FindNextTeleporter(const LOCATIONREL to, LOCATIONABS *from);
RN FindFirstDoor(int level, int column, int row);
RN FindFirstDoor(LOCATIONREL locr);
RN FindFirstDoor(LOCATIONABS *loca);
RN FindFirstTextRecord(const LOCATIONREL locr) const;
RN FindNextTextRecord(RN obj);
void FindNextObject(SEARCHPKT *pkt);
RN GetDBRecordLink(RN recName) const;
bool IsDBIndexOK(int indirectIndex);
bool IsIndirectIndexOK(int indirectIndex) const;
DBCOMMON *GetCommonAddress(DBTYPE dbType, int dbIndex) const;
DBCOMMON *GetCommonAddress(RN recDesc) const;
DB0  *GetRecordAddressDB0(RN object);
DB1  *GetRecordAddressDB1(RN object);
DB2  *GetRecordAddressDB2(RN object);
DB2  *GetRecordAddressDB2(int dbIndex);
DB3  *GetRecordAddressDB3(RN object);
DB3  *GetRecordAddressDB3(int dbIndex);
DB4  *GetRecordAddressDB4(RN object);
DB5  *GetRecordAddressDB5(RN object);
DB6  *GetRecordAddressDB6(RN object);
DB7  *GetRecordAddressDB7(RN object);
DB8  *GetRecordAddressDB8(RN object);
DB9  *GetRecordAddressDB9(RN object);
DB10 *GetRecordAddressDB10(RN object);
DB14 *GetRecordAddressDB14(RN object);
DB15 *GetRecordAddressDB15(RN object);

//void ReplaceDB0 (RN obj, DB0  *pNewDB0);
//void ReplaceDB1 (RN obj, DB1  *pNewDB1);
//void ReplaceDB3 (RN obj, DB3  *pNewDB3);
//void ReplaceDB4 (RN obj, DB4  *pNewDB4);
//void ReplaceDB5 (RN obj, DB5  *pNewDB5);
//void ReplaceDB9 (RN obj, DB9  *pNewDB9);
//void ReplaceDB10(RN obj, DB10 *pNewDB10);

private:
//RN FindEmptyDBEntry(DBTYPE dbType);
void   EnlargeDataMap(void);
public:
RN FindEmptyDB0Entry();
RN FindEmptyDB1Entry();
RN FindEmptyDB2Entry(TEXTTYPE tt, bool assignText=true);
RN FindEmptyDB3Entry(ACTUATORCLASS actuatorClass);
RN FindEmptyDB4Entry(MONSTERTYPE monsterType = MONSTERTYPE());
RN FindEmptyDB5Entry(WEAPONTYPE weaponType = WEAPONTYPE());
RN FindEmptyDB6Entry(CLOTHINGTYPE clothingType = CLOTHINGTYPE());
RN FindEmptyDB7Entry();
RN FindEmptyDB8Entry(POTIONTYPE potionType=POTIONTYPE());
RN FindEmptyDB9Entry();
RN FindEmptyDB10Entry(MISCTYPE miscType=MISCTYPE());
RN FindEmptyDB14Entry();
RN FindEmptyDB15Entry();

RN MakeMoreRoomInDB(DBTYPE dbType);
int GetObjectListIndex(LOCATIONREL locr); //Return -1 if illegal location
void RemoveObjectFromRoom(LOCATIONREL locr, 
                          RN object, 
                          bool Release,
                          bool clearFwdLink);
void AddObjectToRoom(RN newObj, LOCATIONREL locr);
void AddObjectToList(RN newObj, RN *pList, RN after);
void DeleteObjectFromList(RN obj, RN *pList, 
                          bool discard,
                          bool clearObjLink=true);

int ExpandText(char *result,DB2 *pDB2,int flags, int max);
int ExpandText(char *result,RN objTEXT,int flags, int max);
int EncodeAndCompressText(CString& text, ui16 *dest, TEXTTYPE flags);
int GetCompressedTextIndex(RN text);
CString GetRawText(RN objDB2, TEXTTYPE flags);
void Delete(RN object, int password); //Delete indirectIndex and DBentry
void SetText(RN objDB2, CString text, TEXTTYPE flags);
int  ComputeTextLength(ui16 *text);
int  ComputeTextLength(int textIndex);
ui16 *FindRawText(int textIndex)const;
bool IsValidIndirectTextIndex(int indirectTextIndex)const;
bool IsValidCompressedTextIndex(int compressedTextIndex)const;
int  CompressedTextIndex(int indirectTextIndex)const;
int  SizeOfCompressedText(void) const;
void InsertText(RN objDB2, CString text, TEXTTYPE flags);
void RemoveText(DB2 *pDB2);
void DeleteText(int textIndex);
int AddText(ui16 *text, int length);
int NumWordsInTextArray(void) const
            {return m_dungeonDatIndex->NumWordsInTextArray();};
LOADSTATUS ConvertToIndirectText(ui16 *text);
bool CheckCharacterPossessions(bool DOIT);
RN   CharacterPossession(int character, int loc);
void CharacterPossession(int character, int loc, RN possession);
RN  *pCharacterPossession(int character, int loc);
RN   ObjectInHand(void){return m_objectInHand;};
RN  *pObjectInHand(void){return &m_objectInHand;};
int  NumCharacter(void){return m_NumCharacter;};
char *Portrait(int characterNum){return (char *)&m_Characters[characterNum].portrait[0];};
bool CheckObjectInHand(bool DOIT);
bool CheckMissilesInFlight(bool DOIT);
void DeleteTimer(TIMER *pTimer);
int  SetTimer(TIMER *pNewTimer);
int  FindTimerPosition(i32 timerIndex);
void RepositionTimerEntry(i32 timerQPosition);
bool TestTimerLess(TIMER *timer1, TIMER *timer2);
CELL_CHECKPOINT *CheckpointCell(LOCATIONREL locr);
OBJECT_CHECKPOINT *CheckpointObject(RN obj, LOCATIONREL *plocr);
void WriteCellUndo(CELL_CHECKPOINT *pCC);
bool BackspaceCellUndo(void);
CELL_CHECKPOINT *ReplaceCellUndo(CELL_CHECKPOINT *pCCnew);
CELL_CHECKPOINT *ForwardSpaceCellUndo(void);
CELL_CHECKPOINT *ReadCellUndoAtCursor(void);
void ClearCellCheckpoints(void);
bool IsCellUndoAvailable(void);
bool IsCellRedoAvailable(void);
void UndoCell(CELL_CHECKPOINT *pCC);
void RemoveAndDeleteAllObjectsInRoom(LOCATIONREL locr);
void ReplaceAllObjectsInRoom(CELL_CHECKPOINT *pCC);
void ReplaceOneObjectInRoom(LOCATIONREL locr, OBJECT_CHECKPOINT *pOC);
void ReplaceRoomObjectList(LOCATIONREL locr, RN newFirstObj);
bool RepairScrollNonText(CHKERRMSG *pcem);
bool RemoveOrphan(CHKERRMSG *pcem);

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CCSBuildDoc)
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};


extern CCSBuildDoc *pDoc;

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_CSBUILDDOC_H__7F7EED2A_C6A0_11D6_A264_8CD155B1182C__INCLUDED_)
