
#include "stdafx.h"
#include "CSBuildDoc.h"
#include "Clones.h"
#include "Misc.h"

CLONES::CLONES(CWnd *parentWindow)
{
  int i;
  m_parentWindow = parentWindow;
  m_cloneList = NULL;
  for (i=0; i<16; i++) m_dbFlags[i] = NULL;
  Clear();
}

CLONES::~CLONES(void)
{
  Clear();
}

void CLONES::Clear(void)
{
  CLONE *temp;
  int i;
  while (m_cloneList != NULL)
  {
    temp = m_cloneList;
    m_cloneList = temp->m_next;
    delete temp;
  };
  for (i=0; i<16; i++)
  {
    m_sizes[i] = 0;
    if (m_dbFlags[i] != NULL) free(m_dbFlags[i]);
    m_dbFlags[i] = NULL;
  };
}

bool CLONES::IsObjectReferenced(DBTYPE dbType, int index) const
{
  //return (*FindFlag(objIndex) & CLF_1stReference) != 0;
  return (m_dbFlags[dbType][index] & CLF_1stReference) != 0;
}

bool CLONES::IsObjectCloned(DBTYPE dbType, int index) const
{
  //return (*FindFlag(objIndex) & CLF_2ndReference) != 0;
  return (m_dbFlags[dbType][index] & CLF_2ndReference) != 0;
}

bool CLONES::IsIndirectIndexReferenced(int indirectIndex) const
{
  return pDoc->IsIndirectIndexReferenced(indirectIndex);
}

bool CLONES::IsIndirectIndexCloned(int indirectIndex) const
{
  return pDoc->IsIndirectIndexReferenced(indirectIndex);
}

void CLONES::MarkObjectList(RN obj, bool markClones)
{
  //Simply mark all objects with CLF_1stReference if
  // found and CLF_2ndReference if CLF_1stReference
  //  is already set and markClones is true.
  RN possession, contents, flyingObject, wordsOfWisdom;
  DBCOMMON *pObject;
  int objIndex;
  if (obj == RNeof) return;
  if (!obj.IsDBIndexOK()) return;
  for(;obj != RNeof;
        obj = pDoc->GetDBRecordLink(obj))
  {
    objIndex = obj.ConvertToInteger();
    if (pDoc->IsIndirectIndexReferenced(objIndex))
    {
      if (markClones)
        pDoc->MarkIndirectIndexCloned(objIndex);
    }
    else
    {
      pDoc->MarkIndirectIndexReferenced(objIndex);
    };
    if (*FindFlag(objIndex) & CLF_1stReference)
    { //We have seen this object before.
      if (markClones)
      {
        *FindFlag(objIndex) |= CLF_2ndReference;
      };
    };
    *FindFlag(objIndex) |= CLF_1stReference;
    pObject = pDoc->GetCommonAddress(obj);
    switch (obj.dbType())
    {
    case dbDOOR:
      break;
    case dbTELEPORTER:
      break;
    case dbTEXT:
      break;
    case dbACTUATOR:
      break;
    case dbMONSTER:
      possession = pObject->CastToDB4()->possession();
      MarkObjectList(possession, markClones);
      break;
    case dbWEAPON:
      break;
    case   dbCLOTHING:
      break;
    case dbSCROLL:
      wordsOfWisdom = pObject->CastToDB7()->text();
      MarkObjectList(wordsOfWisdom, markClones);
      break;
    case dbPOTION:
      break;
    case dbCHEST:
      contents = pObject->CastToDB9()->contents();
      MarkObjectList(contents, markClones);
      break;
    case dbMISC:
      break;
    case dbEXPOOL:
      break;
    case db12:
    case db13:
      break;
    case dbMISSILE:
      flyingObject = pObject->CastToDB14()->flyingObject();
      if (!flyingObject.IsMagicSpell())
      {
        MarkObjectList(flyingObject, markClones);
      };
      break;
    case dbCLOUD:
      break;
    };
  };
  return;
}

void CLONES::MarkTimerObjects(void)
{
  LOCATIONREL locr;
  locr.l(-1);
  for (i32 i=0; i<pDoc->NumTimers(); i++)
  {
    switch (pDoc->Timer(i)->timerFunction)
    {
    case 24: 
    case 25: //The cloud was found when we scanned the cells.
             break;
    case 48:
    case 49: //The missile was found when we scanned the cells.
             break;
    case 60:
    case 61: MarkObjectList(pDoc->Timer(i)->timerObj8(), true);
             break;
    };
  };
}


void CLONES::MarkCharacterObjects(void)
{
  int i, j;
  LOCATIONREL locr;
  locr.l(-1);
  for (i=0; i<pDoc->NumCharacter(); i++)
  {
    for (j=0; j<30; j++)
    {
      if (pDoc->CharacterPossession(i,j) != RNnul)      
      {
         MarkObjectList(pDoc->CharacterPossession(i,j), true);
      };
    };
  };
  if (pDoc->ObjectInHand() != RNnul)
  {
    MarkObjectList(pDoc->ObjectInHand(), true);
  };

}


void CLONES::MarkCellObjects(void)
{
  short numLev;
  int pl;
  CProgressCtrl progress;
  LOCATIONREL locr;
  RECT prect = {100,100,600,160};
  RN obj;
  numLev=0;
  if (m_parentWindow != NULL)
  {
    progress.Create(WS_CAPTION|WS_VISIBLE|WS_CHILD,prect,m_parentWindow,12);
    progress.SetWindowText("Scanning Dungeon");
    for (locr.First(),pl=locr.l(); locr.l()!=-1; locr.Next())
    {
      if (pl!=locr.l())
      {
        numLev++;
        pl = locr.l();
      };
    };
    progress.SetRange(0,numLev);
  };
  numLev = 0;
  for (locr.First(),pl=locr.l(); locr.l()!=-1; locr.Next())
  {
    if (pl!=locr.l())
    {
      numLev++;
      pl = locr.l();
      if (m_parentWindow != NULL) progress.SetPos(numLev);
    };
    obj = pDoc->FindFirstObject(locr);
    if (obj != RNeof) MarkObjectList(obj, true);
    //if (line == "") continue;
    //loca = locr;
    //warn.Format("%02d(%02d,%02d)%d ",
    //             loca.l,loca.x,loca.y,obj.pos());
    //result += warn;
    //result += line;
  };
}

BYTE *CLONES::FindFlag(int objIndex) const
{
  int dbnum, dbindex;
  static BYTE fakeByte;
  if (!pDoc->IsDBIndexOK(objIndex))
  {
    Strange("Illegal onject index encountered");
    fakeByte = 0;
    return &fakeByte;
  };
  dbnum = pDoc->DataType(objIndex);
  dbindex = pDoc->DataDBIndex(objIndex);
  return m_dbFlags[dbnum] + dbindex;
}


void CLONES::MarkMonsterPossessions(void)
{
  int i;
  DB4 *pDB4;
  for (i=0; i<db.NumEntry(dbMONSTER); i++)
  {
    pDB4 = pDoc->GetCommonAddress(dbMONSTER,i)->CastToDB4();
    if (pDB4->link() == RNnul) continue;
    MarkObjectList(pDB4->possession(), false);
  };
}

void CLONES::MarkScrollTexts(void)
{
  int i;
  DB7 *pDB7;
  for (i=0; i<db.NumEntry(dbSCROLL); i++)
  {
    pDB7 = pDoc->GetCommonAddress(dbSCROLL,i)->CastToDB7();
    if (pDB7->link() == RNnul) continue;
    MarkObjectList(pDB7->text(), false);
  };
}

void CLONES::MarkChestContents(void)
{
  int i;
  DB9 *pDB9;
  for (i=0; i<db.NumEntry(dbCHEST); i++)
  {
    pDB9 = pDoc->GetCommonAddress(dbCHEST,i)->CastToDB9();
    if (pDB9->link() == RNnul) continue;
    MarkObjectList(pDB9->contents(), false);
  };
}

void CLONES::MarkMissileObjects(void)
{
  int i;
  DB14 *pDB14;
  for (i=0; i<db.NumEntry(dbMISSILE); i++)
  {
    pDB14 = pDoc->GetCommonAddress(dbMISSILE,i)->CastToDB14();
    if (pDB14->link() == RNnul) continue;
    if (!pDB14->flyingObject().IsMagicSpell())
    {
      MarkObjectList(pDB14->flyingObject(), false);
    };
  };
}

void CLONES::MarkDungeonObjects(void)
{
  int i;
  Clear();
  pDoc->ClearReferenceAndCloneFlags();
  for (i=0; i<16; i++)
  {
    m_sizes[i] = db.NumEntry(i);
    m_dbFlags[i] = (BYTE *)malloc(m_sizes[i],MALLOC067);
    memset(m_dbFlags[i],0,m_sizes[i]*sizeof(BYTE));
  };
  MarkCellObjects();
  MarkCharacterObjects();
  MarkTimerObjects();
  //Now mark all objects contained in orphaned objects.
  MarkMonsterPossessions();
  MarkScrollTexts();
  MarkChestContents();
  MarkMissileObjects();
}

void CLONES::AnalyzeOneIndirectIndex(int indirectIndex,
                                     CHKERRMSG *result)
{
  LOCATIONREL locr;
  RN cellObj;
  CString line;
  CHKERRMSG *pcem;
  for (pcem = m_cloneParents.FirstError(); 
       pcem != NULL;
       pcem = m_cloneParents.NextError(pcem))
  {
    switch (pcem->ErrorType())
    {
    case CEMT_UNCLONEDPARENTLIST:
    case CEMT_CLONEDPARENTLIST:
    case CEMT_MONSTERPOSSESSION:
    case CEMT_CHESTCONTENTS:
    case CEMT_PARENTCELL:
    case CEMT_SCROLLTEXT:
    case CEMT_CHARACTERPOSSESSION:
      ASSERT(pcem->CloneIndirectIndex != -1);
      if (pcem->CloneIndirectIndex() == indirectIndex)
      {
        result->NewError(pcem);
      };
      break;
    default:
      NotImplemented();
    };
  }
}

bool CLONES::CloneParentAlreadyContains(
                           RN list, 
                           RN clone,
                           CEMT errorType)
{
  int PIndex, CIndex;
  CHKERRMSG *current;
  RN parentObj, cloneObj, linkObj;
  DBCOMMON *pDBParent, *pDBClone;
  for (current = m_cloneParents.LastError();
       current != NULL;
       current = current->Prev())
  {
    if (current->ErrorType() != errorType) continue;
    PIndex = current->ParentIndirectIndex();
    CIndex = current->CloneIndirectIndex();
    cloneObj.ConstructFromInteger(CIndex);
    pDBClone  = pDoc->GetCommonAddress(cloneObj);
    switch (current->ErrorType())
    {
    case CEMT_CLONE: 
      return false;
    case CEMT_UNCLONEDPARENTLIST:
    case CEMT_CLONEDPARENTLIST:
      parentObj.ConstructFromInteger(PIndex);
      pDBParent = pDoc->GetCommonAddress(parentObj);
      linkObj = pDBParent->link();
      break;
    case CEMT_MONSTERPOSSESSION:
      parentObj.ConstructFromInteger(PIndex);
      pDBParent = pDoc->GetCommonAddress(parentObj);
      linkObj = pDoc->GetRecordAddressDB4(parentObj)->possession();
      break;
    case CEMT_CHESTCONTENTS:
      parentObj.ConstructFromInteger(PIndex);
      pDBParent = pDoc->GetCommonAddress(parentObj);
      linkObj = pDoc->GetRecordAddressDB9(parentObj)->contents();
      break;
    case CEMT_SCROLLTEXT:
      parentObj.ConstructFromInteger(PIndex);
      pDBParent = pDoc->GetCommonAddress(parentObj);
      linkObj = pDoc->GetRecordAddressDB7(parentObj)->text();
      break;
    case CEMT_PARENTCELL:
      continue;
    default:
      NotImplemented();
      return false;
    };
    if (   (pDoc->GetCommonAddress(list) == pDBParent)
         &&(pDoc->GetCommonAddress(clone) == pDoc->GetCommonAddress(linkObj)))
    {
      return true;
    };
  };
  return false;
}

void CLONES::FindCloneParentList(
                   RN list,
                   LOCATIONREL locr)
{
  RN nextObj;
  int pindex;
  while (list!=RNeof)
  {
    nextObj = pDoc->GetDBRecordLink(list);
    if (nextObj != RNeof)
    {
      int dbnum, dbindex;
      dbnum = nextObj.dbType();
      dbindex = nextObj.dbIndex();
      if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
      {
        pindex = list.ConvertToInteger();
        if ((*FindFlag(pindex) & CLF_2ndReference) != 0)
        {
          if (CloneParentAlreadyContains(
                       list,nextObj,CEMT_CLONEDPARENTLIST)) 
          {
            list = nextObj;
            continue;
          }
          else
          {
            CString line;
            line.Format("  [%04x] <- Cloned object [%04x]",
                         nextObj.ConvertToInteger(),
                         list.ConvertToInteger());
            m_cloneParents.NewError(line);
            line.Format(" DB%02d(%04x) ",
                         pDoc->DataType(list.ConvertToInteger()),
                         pDoc->DataDBIndex(list.ConvertToInteger()));
            m_cloneParents += line;
            m_cloneParents += list.Name(RoomType(locr));
            m_cloneParents.ErrorType(CEMT_CLONEDPARENTLIST);
            m_cloneParents.ParentIndirectIndex(list.ConvertToInteger());
            m_cloneParents.CloneIndirectIndex(nextObj.ConvertToInteger());
            m_cloneParents.Location(locr);
          };
        }
        else
        {
          CString line;
          LOCATIONABS loca;
          loca = locr;
          line.Format("   [%04x] @ %d(%02d,%02d) ",
                      list.ConvertToInteger(),
                      loca.l,loca.x,loca.y);
          m_cloneParents.NewError(line);
          m_cloneParents += list.Name(RoomType(locr));
          pindex = list.ConvertToInteger();
          m_cloneParents.ErrorType(CEMT_UNCLONEDPARENTLIST);
          m_cloneParents.Location(locr);
          m_cloneParents.ParentIndirectIndex(list.ConvertToInteger());
          m_cloneParents.CloneIndirectIndex(nextObj.ConvertToInteger());
        };
      };
    };
    FindCloneParentContainer(list, locr);
    list = nextObj;
  };
}


void CLONES::FindCloneParentCharacters(void)
{
  int i, n;
  LOCATIONREL locr;
  for (n=0; n<pDoc->NumCharacter(); n++)
  {
    for (i=0; i<30; i++)
    {
      if (pDoc->CharacterPossession(n,i) != RNnul)
      {
        int dbnum, dbindex;
        RN obj;
        obj = pDoc->CharacterPossession(n,i);
        dbnum = obj.dbType();
        dbindex = obj.dbIndex();
        if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
        {
          CString msg;
          msg.Format("  Character %d possessions %d", n, i);
          locr.x(n);
          locr.y(i);
          m_cloneParents.NewError(msg);
          m_cloneParents.ErrorType(CEMT_CHARACTERPOSSESSION);
          m_cloneParents.Location(locr);
          m_cloneParents.CloneIndirectIndex(obj.ConvertToInteger());
        };
        FindCloneParentList(
                         pDoc->CharacterPossession(n,i),
                         locr);
      };
    };
  };
  if (pDoc->ObjectInHand() != RNnul)
  {
    RN obj;
    int dbnum, dbindex;
    obj = pDoc->ObjectInHand();
    dbnum = obj.dbType();
    dbindex = obj.dbIndex();
    if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
    {
      locr.x(-1);
      locr.y(-1);
      m_cloneParents.NewError("  Cursor ");
      m_cloneParents.ErrorType(CEMT_CHARACTERPOSSESSION);
      m_cloneParents.Location(locr);
      m_cloneParents.CloneIndirectIndex(obj.ConvertToInteger());
    };
    FindCloneParentContainer(
                     pDoc->ObjectInHand(),
                     locr);
  };
}

/*
void CLONES::AnalyzeTimerParents(int indirectIndex, 
                                 CHKERRMSG *result)
{
  LOCATIONREL locr;
  locr.l(-1);
  for (i32 i=0; i<pDoc->NumTimers(); i++)
  {
    switch (pDoc->Timer(i)->timerFunction)
    {
    case 24: 
    case 25: AnalyzeContainer(indirectIndex,
                              pDoc->Timer(i)->timerObj8(),
                              result,
                              locr);
             break;
    case 48:
    case 49: AnalyzeContainer(indirectIndex,
                              pDoc->Timer(i)->timerObj6(),
                              result,
                              locr); // The flying object
             break;
    case 60:
    case 61: AnalyzeContainer(indirectIndex,
                              pDoc->Timer(i)->timerObj8(),
                              result,
                              locr);
             break;
    };
  };
}
*/

void CLONES::AnalyzeParents(int dbnum, 
                            int dbindex,
                            CHKERRMSG *result)
{
  int i;
  for (i=1; i<pDoc->dataMapLength(); i++)
  {
    if (   (pDoc->DataDBIndex(i) == dbindex)
         &&(pDoc->DataType(i) == dbnum))
    {
      AnalyzeOneIndirectIndex(i, result);
    };
  };
}

void CLONES::FindCloneParentCells(void)
{
  LOCATIONREL locr;
  RN obj;
  int dbnum, dbindex;

  
  short numLev;
  int pl;
  CProgressCtrl progress;
  RECT prect = {100,100,600,160};
  numLev=0;
  if (m_parentWindow != NULL)
  {
    progress.Create(WS_CAPTION|WS_VISIBLE|WS_CHILD,prect,m_parentWindow,12);
    progress.SetWindowText("Inverting Database");
    for (locr.First(),pl=locr.l(); locr.l()!=-1; locr.Next())
    {
      if (pl!=locr.l())
      {
        numLev++;
        pl = locr.l();
      };
    };
    progress.SetRange(0,numLev);
  };
 
  numLev = 0;  
  for (locr.First(),pl=locr.l(); locr.l()>=0; locr.Next())
  {
    if (pl!=locr.l())
    {
      numLev++;
      pl = locr.l();
      if (m_parentWindow != NULL) progress.SetPos(numLev);
    };
    obj = pDoc->FindFirstObject(locr);
    if (obj == RNeof) continue;
    dbnum = obj.dbType();
    dbindex = obj.dbIndex();
    if (dbindex >= m_sizes[dbnum]) continue;
    if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
    {
      CString msg;
      LOCATIONABS loca;
      loca = locr;
      msg.Format("   Cell %02d(%02d,%02d)",
                     loca.l, loca.x, loca.y);
      m_cloneParents.NewError(msg);
      m_cloneParents.Location(locr);
      m_cloneParents.ErrorType(CEMT_PARENTCELL);
      m_cloneParents.CloneIndirectIndex(obj.ConvertToInteger());
    };
    FindCloneParentList(obj, locr);
  };
}

void CLONES::FindCloneParentTimers()
{
  LOCATIONREL locr;
  locr.l(-1);
  int dbnum, dbindex;
  for (i32 i=0; i<pDoc->NumTimers(); i++)
  {
    switch (pDoc->Timer(i)->timerFunction)
    {
    case 24: 
    case 25: 
      //Nothing to do.  The cloud named in these
      //timers have already been accounted for when we
      //scanned the dungeon cells.
      break;
    case 48:
    case 49: 
      //Nothing to do.  The missile named in these
      //timers have already been accounted for when we
      //scanned the dungeon cells.
      break;
    case 60:
    case 61: 
      dbnum = pDoc->Timer(i)->timerObj8().dbType();
      dbindex = pDoc->Timer(i)->timerObj8().dbIndex();
      if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
      {
        pDoc->Sorry("Timer type 60 clones not implemented");
      };
      FindCloneParentContainer(
                  pDoc->Timer(i)->timerObj8(),
                  locr);
      break;
    };
  };
}


void CLONES::FindCloneParentContainer(
               RN container, //chest, monster, or whatever
               LOCATIONREL locr)
{ //Analyze contents of chest (or whatever) to see if it contains obj.
  DBTYPE obType;
  if (container == RNeof) return;
  obType = container.dbType();
  switch (obType)
  {
  case dbDOOR:
  case dbTELEPORTER:
  case dbCLOTHING:
  case dbWEAPON:
  case dbMISC:
  case dbTEXT:
  case dbPOTION:
  case dbACTUATOR: 
  case dbCLOUD:
    break;
  case dbSCROLL:
    {
      DB7 *pDB7;
      int dbnum, dbindex;
      pDB7 = pDoc->GetRecordAddressDB7(container);
      dbnum = pDB7->text().dbType();
      dbindex = pDB7->text().dbIndex();
      if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
      {
        CString msg;
        LOCATIONABS loca;
        RN text;
        text = pDB7->text();
        if (text == RNeof) break;
        if (CloneParentAlreadyContains(
                  container, 
                  text,
                  CEMT_SCROLLTEXT)) break;
                      
        loca = locr;
        msg.Format("   Contained in scroll [%04x] @ %d(%02d,%02d)",
                    container.ConvertToInteger(),loca.l,loca.x,loca.y);
        m_cloneParents.NewError(msg);
        m_cloneParents.ErrorType(CEMT_SCROLLTEXT);
        m_cloneParents.ParentIndirectIndex(container.ConvertToInteger());
        m_cloneParents.CloneIndirectIndex(text.ConvertToInteger());
        m_cloneParents.Location(locr);
      };
      FindCloneParentContainer(
                      pDB7->text(),
                      locr);
    };
    break;
  case dbMONSTER:
    {
      int dbnum, dbindex;
      DB4 *pDB4;
      RN possessions;
      pDB4 = pDoc->GetRecordAddressDB4(container);
      possessions = pDB4->possession();
      if (possessions == RNeof) break;
      if (!CloneParentAlreadyContains(
                container, 
                possessions,
                CEMT_MONSTERPOSSESSION))
      {
        dbnum = possessions.dbType();
        dbindex = possessions.dbIndex();
        if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
        {
          CString msg;
          CString name;
          MONSTERTYPE mt;
          DB4 *pDB4;
          pDB4 = pDoc->GetRecordAddressDB4(container);
          mt = pDB4->monsterType();
          name = mt.Name();
          LOCATIONABS loca;
          loca = locr;
          msg.Format("  Carried by %s @ %d(%02d,%02d)",
                     name,loca.l, loca.x, loca.y);
          m_cloneParents.NewError(msg);
          m_cloneParents.ErrorType(CEMT_MONSTERPOSSESSION);
          m_cloneParents.CloneIndirectIndex(possessions.ConvertToInteger());
          m_cloneParents.ParentIndirectIndex(container.ConvertToInteger());
          m_cloneParents.Location(locr);
        };
      };
      FindCloneParentList(
                      possessions,
                      locr);
    };
    break;
  case dbCHEST:
    {
      int dbnum, dbindex;
      DB9 *pDB9;
      RN contents;
      pDB9 = pDoc->GetRecordAddressDB9(container);
      contents = pDB9->contents();
      if (contents == RNeof) break;
      if (!CloneParentAlreadyContains(
                container, 
                contents,
                CEMT_CHESTCONTENTS))
      {
        dbnum = contents.dbType();
        dbindex = contents.dbIndex();
        if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
        {
          CString msg;
          LOCATIONABS loca;
          loca = locr;
          msg.Format("  Contained in Chest @ %d(%02d,%02d)",
                     loca.l, loca.x, loca.y);
          m_cloneParents.NewError(msg);
          m_cloneParents.ErrorType(CEMT_CHESTCONTENTS);
          m_cloneParents.CloneIndirectIndex(contents.ConvertToInteger());
          m_cloneParents.ParentIndirectIndex(container.ConvertToInteger());
          m_cloneParents.Location(locr);
        };
      };
      FindCloneParentList(
                      contents,
                      locr);
    };
    break;
  case dbMISSILE:
    {
      DB14 *pDB14;
      RN flyingObject;
      pDB14 = pDoc->GetRecordAddressDB14(container);
      for (flyingObject = pDB14->flyingObject();
           flyingObject != RNeof;
           flyingObject = pDoc->GetDBRecordLink(flyingObject))
      {
        int dbnum, dbindex;
        if (flyingObject.IsMagicSpell()) break;
        dbnum = flyingObject.dbType();
        dbindex = flyingObject.dbIndex();
        if ((m_dbFlags[dbnum][dbindex] & CLF_2ndReference) != 0)
        {
          m_cloneParents.NewError("CLONES Missiles not implemented");
        };
      };
      for (flyingObject = pDB14->flyingObject();
           flyingObject != RNeof;
           flyingObject = pDoc->GetDBRecordLink(flyingObject))
      {
        if (flyingObject.IsMagicSpell()) break;
        FindCloneParentContainer(
                      flyingObject,
                      locr);
      }
    };
    break;
  default:
    NotImplemented();
  };
}

void CLONES::FindCloneParents(void)
{
  //Go through the entire dungeon.  Find every indirectIndex
  //whose associated object points to a cloned object.
  //Add each to the m_cloneParents list along with enough
  //information to identify exactly where the pointer can
  //be found (such as chests...which object in its list).
  //Also find every cell that points to a cloned object.
  //Also find every timer that points to a cloned object.
  //Also find every character possessing a cloned object.
  m_cloneParents.Clear();
  FindCloneParentCells();
  FindCloneParentTimers();
  FindCloneParentCharacters();
  //All done....that was easy!
}

void CLONES::Analyze(CHKERRMSG *result)
{
  int dbnum, i, objID;
  short numClone;
  CString line, name;
  RN obj;
  CProgressCtrl progress;
  RECT prect = {100,100,600,160};
  FindCloneParents();
  numClone = 0;
  for (dbnum=0; dbnum<16; dbnum++)
  {
    for (i=0; i<m_sizes[dbnum]; i++)
    {
      if ((m_dbFlags[dbnum][i] & CLF_2ndReference) != 0)
      {
        numClone++;;
      };
    };
  };
  if (m_parentWindow != NULL)
  {
    progress.Create(WS_CAPTION|WS_VISIBLE|WS_CHILD,prect,m_parentWindow,12);
    progress.SetWindowText("Analyzing Clones");
    progress.SetRange(0,numClone);
  };
  numClone = 0;
  for (dbnum=0; dbnum<16; dbnum++)
  {
    for (i=0; i<m_sizes[dbnum]; i++)
    {
      if ((m_dbFlags[dbnum][i] & CLF_2ndReference) != 0)
      {
        numClone++;
        if (m_parentWindow != NULL) progress.SetPos(numClone);
        obj = RNeof;
        for (objID=1; objID<pDoc->dataMapLength(); objID++)
        {
          if (  (pDoc->DataDBIndex(objID) == i)
              &&(pDoc->DataType(objID) == dbnum))
          {
            obj.ConstructFromInteger(objID);
            break;
          };
        };
        if (obj == RNeof) name = "Unknown";
        else name = obj.Name(roomSTONE);
        line.Format("DB%02d[%04x] cloned %s",dbnum,i,name);
        result->NewError(line);
        result->CloneIndirectIndex(obj.ConvertToInteger());
        result->ErrorType(CEMT_CLONE);
        AnalyzeParents(dbnum,i,result);
      };
    };
  };
}