// EDIT_PressurePad.cpp : implementation file
//

#include "stdafx.h"
//#include "types.h"
#include "CSBuild.h"
#include "Actuator.h"
#include "EDIT_PressurePad.h"
#include "Graphics.h"
#include "Misc.h"
#include "CSBuildDoc.h"
#include "EditTarget.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CString TranslateObjectName(const char *name);

/////////////////////////////////////////////////////////////////////////////
// EDIT_PressurePad dialog


EDIT_PressurePad::EDIT_PressurePad(LOCATIONREL locr,
                                   RN RNpressurePad,
                                   CWnd* pParent /*=NULL*/)
	: CDialog(EDIT_PressurePad::IDD, pParent)
{
	//{{AFX_DATA_INIT(EDIT_PressurePad)
	m_clearConditionMet = FALSE;
	m_clearConditionNotMet = FALSE;
	m_delay = 0;
	m_nothing = -1;
	m_onceOnly = FALSE;
	m_setConditionMet = FALSE;
	m_setConditionNotMet = FALSE;
	m_sound = FALSE;
	m_targetColumn = 0;
	m_targetRow = 0;
	m_toggleConditionMet = FALSE;
	m_toggleConditionNotMet = FALSE;
	m_value = 0;
	m_visible = FALSE;
	m_pressLabel = _T("");
	m_releaseLabel = _T("");
	m_actionTitle = _T("");
	//}}AFX_DATA_INIT
  m_locr = locr;
  m_RNpressurePad = RNpressurePad;
  m_levelInfo = locr;
  m_directionListIndex = -1;
  m_graphicListIndex = -1;
  m_particularObjectType = TranslateObjectName("APPLE");
}

unsigned char legalActions[10][8] = 
{
  { 
    0x00, //set    NO  //0 = Nothing
    0x00, //set    NC
    0x00, //clear  NO
    0x00, //clear  NC
    0x00, //toggle NO
    0x00, //toggle NC
    0x00, //hold   NO
    0x00, //hold   NC
  },
  {
    0x80, //set    NO  //1 = Anything put on switch
    0x40, //set    NC
    0x20, //clear  NO
    0x10, //clear  NC
    0x08, //toggle NO
    0x04, //toggle NC
    0x90, //hold   NO
    0x60, //hold   NC
  },
  {
    0x80, //set    NO  //2 = Party or monster
    0x40, //set    NC
    0x20, //clear  NO
    0x10, //clear  NC
    0x08, //toggle NO
    0x04, //toggle NC
    0x90, //hold   NO
    0x60, //hold   NC
  },
  {
    0x80, //set    NO  //3 = Party possesses object
    0x40, //set    NC
    0x20, //clear  NO
    0x10, //clear  NC
    0x08, //toggle NO
    0x04, //toggle NC
    0x90, //hold   NO
    0x60, //hold   NC
  },
  {
    0x80, //set    NO  //4 = Particular object
    0x40, //set    NC
    0x20, //clear  NO
    0x10, //clear  NC
    0x08, //toggle NO
    0x04, //toggle NC
    0x90, //hold   NO
    0x60, //hold   NC
  },
  {
    0x80, //set    NO  //5 = Monster
    0x40, //set    NC
    0x20, //clear  NO
    0x10, //clear  NC
    0x08, //toggle NO
    0x04, //toggle NC
    0x90, //hold   NO
    0x60, //hold   NC
  },
  {
    0x80, //set    NO  //6 = facing
    0x40, //set    NC
    0x20, //clear  NO
    0x10, //clear  NC
    0x08, //toggle NO
    0x04, //toggle NC
    0x90, //hold   NO
    0x60, //hold   NC
  },
  {
    0x80, //set    NO  //7 = Party Steps on Pad
    0x00, //set    NC
    0x20, //clear  NO
    0x00, //clear  NC
    0x08, //toggle NO
    0x00, //toggle NC
    0x00, //hold   NO
    0x00, //hold   NC
  },
  {
    0x80, //set    NO  //8 = Party Steps on/off Pad
    0x40, //set    NC
    0x20, //clear  NO
    0x10, //clear  NC
    0x08, //toggle NO
    0x04, //toggle NC
    0x90, //hold   NO
    0x60, //hold   NC
  },
};

void EDIT_PressurePad::ExtractDataFromRecord(void)
{
  int acType;
  ACTUATORACTION act;
  bool NC;
  m_particularObjectType = m_pPressurePad->neededObject(m_locr);
  m_objectPossessedType = m_pPressurePad->neededObject(m_locr);
  m_positionListIndex = m_pPressurePad->targetPos();
  m_directionListIndex = m_pPressurePad->value() - 1;
  if (m_directionListIndex < 0) m_directionListIndex = 0;
  acType = m_pPressurePad->actuatorType();
  switch (acType)
  {
  case 0:
    m_nothing = 0; //Nothing
    break;
  case 1:
    m_nothing = 1; //Anything
    break;
  case 2:
    m_nothing = 2; //Party or Monster
    break;
  case 3:
    if (m_pPressurePad->value() == 0) //any direction
    {
      m_nothing = 8; //party moves on or off
    }
    else
    {
      m_nothing = 6; //Party moves or turns (facing)
    };
    break;
  case 4:
    m_nothing = 4; //Particular object
    break;
  case 7:
    m_nothing = 5; //Monster
    break;
  case 8:
    m_nothing = 3; //Party possesses object
    break;
  case 9:
    m_nothing = 7; //Party steps on pad
    break;
  default:
    m_nothing = -1;
  };
  act = m_pPressurePad->action();
  NC = m_pPressurePad->normallyClosed();
  switch (act)
  {
  case AA_HOLD: if (NC) m_clearConditionMet = m_setConditionNotMet = true;
                else m_setConditionMet = m_clearConditionNotMet = true;
                break;
  case AA_SET: if (NC) m_setConditionNotMet=true; else m_setConditionMet=true; break;
  case AA_CLEAR: if (NC) m_clearConditionNotMet=true; else m_clearConditionMet=true; break;
  case AA_TOGGLE: if (NC) m_toggleConditionNotMet=true;else m_toggleConditionMet=true; break;
  };
  m_onceOnly = m_pPressurePad->onceOnly();
//  m_visible = m_pPressurePad->Visible();
  m_targetColumn = m_pPressurePad->targetX() + m_levelInfo.left;
  m_targetRow = m_pPressurePad->targetY() + m_levelInfo.top;
  m_delay = m_pPressurePad->delay();
  m_sound = m_pPressurePad->audible();
  m_graphicListIndex = m_pPressurePad->graphic();
}

void EDIT_PressurePad::EncodeDataIntoRecord(void)
{
  int acType;
  bool NC=false;
  ACTUATORACTION act=AA_SET;
  switch(m_nothing)
  {
  case 0: acType = 0; break;
  case 1: acType = 1; break;
  case 2: acType = 2; break;
  case 3: acType = 8; break;
  case 4: acType = 4; break;
  case 5: acType = 7; break;
  case 6: acType = 3; break;
  case 7: acType = 9; break;
  case 8: acType = 3; break;
  default: InternalError("Bad PressurePad type");
           acType = 1;
           break;
  };
  m_pPressurePad->actuatorType(acType);
  m_pPressurePad->targetPos(m_positionListIndex);
  switch (m_nothing)
  {
  case 3:
    m_objectPossessedType
       = MODIFIEDOBJECTTYPE(m_objectPossessed.GetItemData(m_objectPossessedIndex));
    m_pPressurePad->neededObject(m_objectPossessedType);
    break;
  case 4:
    m_particularObjectType 
       = MODIFIEDOBJECTTYPE(m_particularObject.GetItemData(m_particularObjectIndex));
    m_pPressurePad->neededObject(m_particularObjectType);
    break;
  case 6:
    m_pPressurePad->value(m_directionListIndex+1);
    break;
  case 7:
    m_pPressurePad->value(0);
    break;
  case 8:
    m_pPressurePad->value(0);
    break;
  };
  if (m_setConditionMet && !m_clearConditionNotMet)
  {
    NC = false;
    act = AA_SET;
  };
  if (m_setConditionNotMet && !m_clearConditionMet)
  {
    NC = true;
    act = AA_SET;
  };
  if (m_clearConditionMet && !m_setConditionNotMet)
  {
    NC = false;
    act = AA_CLEAR;
  };
  if (m_clearConditionNotMet && !m_setConditionMet)
  {
    NC = true;
    act = AA_CLEAR;
  };
  if (m_setConditionMet && m_clearConditionNotMet)
  {
    NC = false;
    act = AA_HOLD;
  };
  if (m_setConditionNotMet && m_clearConditionMet)
  {
    NC = true;
    act = AA_HOLD;
  };
  if (m_toggleConditionMet)
  {
    NC = false;
    act = AA_TOGGLE;
  };
  if (m_toggleConditionNotMet)
  {
    NC = true;
    act = AA_TOGGLE;
  };
  m_pPressurePad->action(act);
  m_pPressurePad->normallyClosed(NC);
  m_pPressurePad->onceOnly(m_onceOnly==TRUE);
  m_pPressurePad->targetX(m_targetColumn - m_levelInfo.left);
  m_pPressurePad->targetY(m_targetRow - m_levelInfo.top);
  m_pPressurePad->delay(m_delay);
  m_pPressurePad->audible(m_sound==TRUE);
  m_pPressurePad->graphic(m_graphicListIndex);
//  m_pPressurePad->visible(m_visible);
}

void EDIT_PressurePad::Show(int id, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(id);
  if (pWnd!=NULL) pWnd->ShowWindow(show);
}

bool EDIT_PressurePad::PossibleAction(int mask)
{
  int i;
  for (i=0; i<8; i++)
  {
    if (legalActions[m_nothing][i] & mask) return true;
  };
  return false;
}

void EDIT_PressurePad::HideImpossibleActions(void)
{
  Show(IDC_SetConditionMet,       PossibleAction(0x80));
  Show(IDC_SetConditionNotMet,    PossibleAction(0x40));
  Show(IDC_ClearConditionMet,     PossibleAction(0x20));
  Show(IDC_ClearConditionNotMet,  PossibleAction(0x10));
  Show(IDC_ToggleConditionMet,    PossibleAction(0x08));
  Show(IDC_ToggleConditionNotMet, PossibleAction(0x04));
  Show(IDC_Visible,               false);
}

void EDIT_PressurePad::DoDataExchange(CDataExchange* pDX)
{
//  CWnd *pWnd;
//  int id;
//  char *name;
  int i;
  int numFloorDecoration;
  FLOORTYPE floor;
  if (!pDX->m_bSaveAndValidate)
  {
    switch (m_nothing)
    {
    case 1:
      m_actionTitle = "Anything presses on pad";
      m_pressLabel = "Pad\nPressed";
      m_releaseLabel = "Pad\nReleased";
      break;
    case 2:
      m_actionTitle = "Party or Monster activates pad";
      m_pressLabel = "Steps\nOn";
      m_releaseLabel = "Steps\nOff";
      break;
    case 3:
      m_actionTitle = "Party steps on, off, or turns";
      m_pressLabel = "With\nObject";
      m_releaseLabel = "Without\nObject";
      break;
    case 4:
      m_actionTitle = "Particular object is added\nor removed";
      m_pressLabel = "Add\nObject";
      m_releaseLabel = "Remove\nObject";
      break;
    case 5:
      m_actionTitle = "Monster steps on or off";
      m_pressLabel = "Steps\nOn";
      m_releaseLabel = "Steps\nOff";
      break;
    case 6:
      m_actionTitle = "Party moves or turns";
      m_pressLabel = "On pad and\nfacing";
      m_releaseLabel = "Off pad or\nnot facing";
      break;
    case 7:
      m_actionTitle = "Party steps on pad";
      m_pressLabel = "";
      m_releaseLabel = "";
      break;
    case 8:
      m_actionTitle = "Party steps on/off pad";
      m_pressLabel = "Step\nOn";
      m_releaseLabel = "Step\nOff";
      break;
    default:
      m_actionTitle = "";
      m_pressLabel = "";
      m_releaseLabel = "";
      break;
    };
  };
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EDIT_PressurePad)
	DDX_Control(pDX, IDC_ParticularObject, m_particularObject);
	DDX_Control(pDX, IDC_ObjectPossessed, m_objectPossessed);
	DDX_Control(pDX, IDC_Graphic, m_graphic);
	DDX_Control(pDX, IDC_Direction, m_direction);
	DDX_Control(pDX, IDC_PositionList, m_PositionList);
	DDX_Check(pDX, IDC_ClearConditionMet, m_clearConditionMet);
	DDX_Check(pDX, IDC_ClearConditionNotMet, m_clearConditionNotMet);
	DDX_Text(pDX, IDC_Delay, m_delay);
	DDV_MinMaxUInt(pDX, m_delay, 0, 15);
	DDX_Radio(pDX, IDC_Nothing, m_nothing);
	DDX_Check(pDX, IDC_OnceOnly, m_onceOnly);
	DDX_Check(pDX, IDC_SetConditionMet, m_setConditionMet);
	DDX_Check(pDX, IDC_SetConditionNotMet, m_setConditionNotMet);
	DDX_Check(pDX, IDC_Sound, m_sound);
	DDX_Text(pDX, IDC_TargetColumn, m_targetColumn);
	DDX_Text(pDX, IDC_TargetRow, m_targetRow);
	DDX_Check(pDX, IDC_ToggleConditionMet, m_toggleConditionMet);
	DDX_Check(pDX, IDC_ToggleConditionNotMet, m_toggleConditionNotMet);
	DDX_Text(pDX, IDC_Value, m_value);
	DDX_Check(pDX, IDC_Visible, m_visible);
	DDX_Text(pDX, IDC_PressLabel, m_pressLabel);
	DDX_Text(pDX, IDC_ReleaseLabel, m_releaseLabel);
	DDX_Text(pDX, IDC_ActionTitle, m_actionTitle);
	//}}AFX_DATA_MAP
  if (pDX->m_bSaveAndValidate)
  {
    m_positionListIndex = m_PositionList.GetCurSel();
    numFloorDecoration = pDoc->NumFloorDecorations(m_levelInfo.level);
    if (m_graphic.GetCurSel() <= numFloorDecoration)
    {
      m_graphicListIndex = m_graphic.GetCurSel();
    };
    m_particularObjectIndex = m_particularObject.GetCurSel();
    if (m_particularObjectIndex >= 0) 
    {
      m_particularObjectType = 
          m_particularObject.GetItemData(m_particularObjectIndex);
    };
    m_objectPossessedIndex = m_objectPossessed.GetCurSel();
    if (m_objectPossessedIndex >= 0) 
    {
      m_objectPossessedType = 
          m_particularObject.GetItemData(m_objectPossessedIndex);
    };
    m_directionListIndex = m_direction.GetCurSel();
  };
  if (!pDX->m_bSaveAndValidate)
  {
    m_particularObjectIndex = 
        m_particularObjectType.FillComboBox(&m_particularObject);
    m_objectPossessedIndex = 
        m_objectPossessedType.FillComboBox(&m_objectPossessed);
    Show(IDC_Direction, m_nothing==6);
    Show(IDC_ObjectPossessed, m_nothing==3);
    Show(IDC_ParticularObject,m_nothing==4);
    m_graphic.ResetContent();
    m_graphic.AddString("none");
    numFloorDecoration = pDoc->NumFloorDecorations(m_levelInfo.level);
    for (i=0; i<numFloorDecoration; i++)
    {
      floor = pDoc->FloorDecorationType(m_levelInfo.level,i+1);
      m_graphic.AddString(floor.Name());
    };
    if (m_graphicListIndex > numFloorDecoration)
    {
      m_graphic.AddString("Illegal!");
      m_graphic.SetCurSel(numFloorDecoration+1);
    }
    else
    {
      m_graphic.SetCurSel(m_graphicListIndex);
    };
    m_PositionList.SetCurSel(m_positionListIndex);
    m_objectPossessed.SetCurSel(m_objectPossessedIndex);
    m_particularObject.SetCurSel(m_particularObjectIndex);
    m_direction.SetCurSel(m_directionListIndex);
    HideImpossibleActions();
  };
}


BEGIN_MESSAGE_MAP(EDIT_PressurePad, CDialog)
	//{{AFX_MSG_MAP(EDIT_PressurePad)
	ON_BN_CLICKED(IDC_Nothing, OnOperatedBy)
	ON_BN_CLICKED(IDC_ClearConditionMet, OnClearConditionMet)
	ON_BN_CLICKED(IDC_ClearConditionNotMet, OnClearConditionNotMet)
	ON_BN_CLICKED(IDC_SetConditionMet, OnSetConditionMet)
	ON_BN_CLICKED(IDC_SetConditionNotMet, OnSetConditionNotMet)
	ON_BN_CLICKED(IDC_ToggleConditionMet, OnToggleConditionMet)
	ON_BN_CLICKED(IDC_ToggleConditionNotMet, OnToggleConditionNotMet)
	ON_BN_CLICKED(IDC_Browse, OnBrowse)
	ON_BN_CLICKED(IDC_Anything, OnOperatedBy)
	ON_BN_CLICKED(IDC_Monster, OnOperatedBy)
	ON_BN_CLICKED(IDC_Object, OnOperatedBy)
	ON_BN_CLICKED(IDC_Party, OnOperatedBy)
	ON_BN_CLICKED(IDC_PartyOrMonster, OnOperatedBy)
	ON_BN_CLICKED(IDC_Possession, OnOperatedBy)
	ON_BN_CLICKED(IDC_PartyStepsOn, OnOperatedBy)
	ON_BN_CLICKED(IDC_PartyMoves, OnOperatedBy)
	ON_WM_HELPINFO()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

int ACTUATOR_PressurePad::Edit(LOCATIONREL locr,
                               RN RNpressurePad)
{
  EDIT_PressurePad dlg(locr, RNpressurePad);
  dlg.m_pPressurePad = this;
  return dlg.DoModal();
}

/////////////////////////////////////////////////////////////////////////////
// EDIT_PressurePad message handlers

void EDIT_PressurePad::ValidateActions(int change)
{
  int curAction = 0;
  int i, a;
  curAction |= m_setConditionMet      ?0x80:0;
  curAction |= m_setConditionNotMet   ?0x40:0;
  curAction |= m_clearConditionMet    ?0x20:0;
  curAction |= m_clearConditionNotMet ?0x10:0;
  curAction |= m_toggleConditionMet   ?0x08:0;
  curAction |= m_toggleConditionNotMet?0x04:0;
  for (i=0; i<8; i++)
  {
    if (legalActions[m_nothing][i] == curAction) return;
  };
  // Try to find one that does not modify 'change'
  a = -1;
  for (i=0; i<8; i++)
  {
    if ( legalActions[m_nothing][i] == (curAction^change))
    {
      //Great.  Let's use this one.
      a = legalActions[m_nothing][i];
      break;
    };
  };
  if (a == -1) a = legalActions[m_nothing][0];
  m_setConditionMet       = (a & 0x80) != 0;
  m_setConditionNotMet    = (a & 0x40) != 0;
  m_clearConditionMet     = (a & 0x20) != 0;
  m_clearConditionNotMet  = (a & 0x10) != 0;
  m_toggleConditionMet    = (a & 0x08) != 0;
  m_toggleConditionNotMet = (a & 0x04) != 0;
}

void EDIT_PressurePad::OnOperatedBy() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
  ValidateActions(0);
  UpdateData(false);
}


BOOL EDIT_PressurePad::OnInitDialog() 
{
  CString title;
  LOCATIONABS loca;
  ExtractDataFromRecord();
	CDialog::OnInitDialog();
  loca = m_locr;
  title.Format("Edit Pressure Pad [%04x] %02d(%02d,%02d)",
                  m_RNpressurePad.ConvertToInteger(),
                  loca.l, loca.x, loca.y);
  this->SetWindowText(title);

	
	// TODO: Add extra initialization here
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EDIT_PressurePad::OnOK() 
{
	// TODO: Add extra validation here
  UpdateData(true);
	EncodeDataIntoRecord();
	CDialog::OnOK();
}

void EDIT_PressurePad::UpdateAction(void)
{
  CheckDlgButton(IDC_SetConditionMet,m_setConditionMet);
  CheckDlgButton(IDC_SetConditionNotMet,m_setConditionNotMet);
  CheckDlgButton(IDC_ClearConditionMet,m_clearConditionMet);
  CheckDlgButton(IDC_ClearConditionNotMet,m_clearConditionNotMet);
  CheckDlgButton(IDC_ToggleConditionMet,m_toggleConditionMet);
  CheckDlgButton(IDC_ToggleConditionNotMet,m_toggleConditionNotMet);
}



void EDIT_PressurePad::OnClearConditionMet() 
{
	// TODO: Add your control notification handler code here
	
  m_clearConditionMet = !m_clearConditionMet;
  if (m_clearConditionMet)
  {
    m_setConditionMet = false;
    m_clearConditionNotMet = false;
    m_toggleConditionMet = false;
    m_toggleConditionNotMet = false;
  }
  else
  {
    if (!m_setConditionNotMet) m_setConditionMet = true;
  };
  UpdateAction();
}

void EDIT_PressurePad::OnClearConditionNotMet() 
{
	// TODO: Add your control notification handler code here
	
  m_clearConditionNotMet = !m_clearConditionNotMet;
  if (m_clearConditionNotMet)
  {
    m_setConditionNotMet = false;
    m_clearConditionMet = false;
    m_toggleConditionMet = false;
    m_toggleConditionNotMet = false;
  }
  else
  {
    if (!m_setConditionMet) m_clearConditionMet = true;
  };	
  UpdateAction();
}

void EDIT_PressurePad::OnSetConditionMet() 
{
	// TODO: Add your control notification handler code here
	
  m_setConditionMet = !m_setConditionMet;
  if (m_setConditionMet)
  {
    m_setConditionNotMet = false;
    m_clearConditionMet = false;
    m_toggleConditionMet = false;
    m_toggleConditionNotMet = false;
  }
  else
  {
    if (!m_clearConditionNotMet) m_setConditionNotMet = true;
  };
  UpdateAction();
}

void EDIT_PressurePad::OnSetConditionNotMet() 
{
	// TODO: Add your control notification handler code here
	
  m_setConditionNotMet = !m_setConditionNotMet;
  if (m_setConditionNotMet)
  {
    m_setConditionMet = false;
    m_clearConditionNotMet = false;
    m_toggleConditionMet = false;
    m_toggleConditionNotMet = false;
  }
  else
  {
    if (!m_clearConditionMet) m_setConditionMet = true;
  };
  UpdateAction();
}

void EDIT_PressurePad::OnToggleConditionMet() 
{
	// TODO: Add your control notification handler code here
	
  m_toggleConditionMet = !m_toggleConditionMet;
  if (m_toggleConditionMet)
  {
    m_setConditionMet = false;
    m_setConditionNotMet = false;
    m_clearConditionMet = false;
    m_clearConditionNotMet = false;
    m_toggleConditionNotMet = false;
  }
  else
  {
    m_setConditionMet = true;
  };
  UpdateAction();
}

void EDIT_PressurePad::OnToggleConditionNotMet() 
{
	// TODO: Add your control notification handler code here
	
  m_toggleConditionNotMet = !m_toggleConditionNotMet;
  if (m_toggleConditionNotMet)
  {
    m_setConditionMet = false;
    m_setConditionNotMet = false;
    m_clearConditionMet = false;
    m_clearConditionNotMet = false;
    m_toggleConditionMet = false;
  }
  else
  {
    m_setConditionMet = true;
  };
  UpdateAction();
}


void EDIT_PressurePad::OnBrowse() 
{
	// TODO: Add your control notification handler code here
  LOCATIONABS loca, locaPressurePad;
  LOCATIONREL locrTarget;
  UpdateData(true);
  loca.l = m_locr.l();
  loca.x = m_targetColumn;
  loca.y = m_targetRow;
  locrTarget = loca;
  locaPressurePad = m_locr;
	EditTarget dlg(locrTarget,"Pressure Pad",locaPressurePad,false);
  if (dlg.DoModal() == IDOK)
  {
    locrTarget.l(dlg.m_locr.l());
    locrTarget.x(dlg.m_locr.x());
    locrTarget.y(dlg.m_locr.y());
    loca = locrTarget;
    m_targetColumn = loca.x;
    m_targetRow = loca.y;
    UpdateData(false);
  };	
}

BOOL EDIT_PressurePad::OnHelpInfo(HELPINFO* pHelpInfo) 
{
	// TODO: Add your message handler code here and/or call default
	
	CString s;
  int id;
  switch (pHelpInfo->iCtrlId)
  {
  case IDC_Delay: id = IDH_EditPressurePad_Delay; break;
  default: id = IDH_Context_None;
  };
	s.LoadString(id);
  HelpMsg(s);
  return true;
}
