// EditAndOr.cpp : implementation file
//

#include "stdafx.h"
//#include "types.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "Actuator.h"
#include "EditAndOr.h"
#include "misc.h"
#include "EditTarget.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// EditAndOr dialog


EditAndOr::EditAndOr(LOCATIONREL locr, 
                     RN RNandOr,
                     CWnd* pParent /*=NULL*/) 
	: CDialog(EditAndOr::IDD, pParent)
{
  m_locr = locr;
  m_RNandOr = RNandOr;
	//{{AFX_DATA_INIT(EditAndOr)
	m_column = 0;
	m_delay = 0;
	m_initialValue = 0;
	m_once = FALSE;
	m_row = 0;
	m_targetValue = 0;
	m_clearConditionMet = FALSE;
	m_clearConditionNotMet = FALSE;
	m_setConditionMet = FALSE;
	m_setConditionNotMet = FALSE;
	m_toggleConditionMet = FALSE;
	m_toggleConditionNotMet = FALSE;
	m_rotateConditionMet = FALSE;
	m_rotateConditionNotMet = FALSE;
	m_noEdgeDetect = -1;
	m_skillConditionMet = FALSE;
	m_skillConditionNotMet = FALSE;
	//}}AFX_DATA_INIT
}

enum SPECIFICATIONS
{
  MS = 0x001,  //Set          - condition met
  NS = 0x002,  //Set          - condition Not met
  MC = 0x004,  //Clear        - condition Met
  NC = 0x008,  //Clear        - condition Not met
  MT = 0x010,  //Toggle       - condition Met
  NT = 0x020,  //Toggle       - condition Not met
  MR = 0x040,  //Rotate       - condition Met
  NR = 0x080,  //Rotate       - condition Not met
  MK = 0x100,  //sKill adjust - condition Met
  NK = 0x200   //sKill adjust - condition Not met
};



int legalActions[12] =
{
  MS,   // Set conditon Met
  NS,   // Set condition Not met
  MC,   // Clear condition Met
  NC,   // Clear condition Not met
  MT,   // Toggle condition Met
  NT,   // Toggle condition Not met
  MS+NC,// Hold condition Met
  MC+NS,// Hold condition Not met
  MR,   // Rotate condition Met
  NR,   // Rotate condition Not met
  MK,   // sKill condition Met
  NK,   // sKill condition Not met
};

void EditAndOr::ExtractDataFromRecord(void)
{
  m_action = m_pGate->action();
  m_NC = m_pGate->normallyClosed();
  m_localAction = m_pGate->localAction();
  m_once = m_pGate->onceOnly();
  m_initialValue = m_pGate->initialValue();
  m_targetValue = m_pGate->finalValue();
  m_noEdgeDetect = m_pGate->Edge();
  if (!m_localAction)
  {
    LOCATIONABS loca;
    LOCATIONREL locr;
    locr.x(m_pGate->targetX());
    locr.y(m_pGate->targetY());
    locr.l(m_locr.l());
    locr.p(m_pGate->targetPos());
    loca = locr;
    m_column = loca.x;
    m_row = loca.y;
    m_positionIndex = loca.p;
    m_localActionType = 0;
    m_delay = m_pGate->delay();
  }
  else
  {
    m_column = 0;
    m_row = 0;
    m_positionIndex = 0;
    m_delay = 0;
    m_localActionType = m_pGate->localActionType();
  };
  m_graphicListIndex = m_pGate->wallGraphicOrdinal();
}

void EditAndOr::EncodeDataIntoRecord(void)
{
  ACTUATORACTION ac;
  m_pGate->normallyClosed(false);
  m_localAction = false;
  if (m_setConditionMet && !m_clearConditionNotMet) ac = AA_SET;
  else if (m_clearConditionMet && !m_setConditionNotMet) ac = AA_CLEAR;
  else if (m_toggleConditionMet) ac = AA_TOGGLE;
  else if (m_setConditionMet && m_clearConditionNotMet) ac = AA_HOLD;
  else if (m_skillConditionMet) 
  {
    m_localAction = true;
    m_localActionType = 10;
    ac = AA_SET;
  }
  else if (m_rotateConditionMet) 
  {
    m_localAction = true;
    ac = AA_CLEAR;
    m_localActionType = 2;
  }
  else 
  {
    m_pGate->normallyClosed(true);
    if (!m_clearConditionMet && m_setConditionNotMet) ac = AA_SET;
    else if (!m_setConditionMet && m_clearConditionNotMet) ac = AA_CLEAR;
    else if (m_toggleConditionNotMet) ac = AA_TOGGLE;
    else if (m_skillConditionNotMet)
    {
      m_localAction = true;
      m_localActionType = 10;
      ac = AA_SET;
    }
    else if (m_rotateConditionNotMet)
    {
      m_localAction = true;
      m_localActionType = 2;
      ac = AA_CLEAR;
    }
    else ac = AA_HOLD;
  };
  m_pGate->action(ac);
  m_pGate->onceOnly(m_once != 0);
  m_pGate->initialValue(m_initialValue);
  m_pGate->finalValue(m_targetValue);
  m_pGate->Edge(m_noEdgeDetect);
  if (!m_localAction)
  {
    LOCATIONABS loca;
    LOCATIONREL locr;
    m_pGate->localAction(false);
    loca.x = m_column;
    loca.y = m_row;
    loca.p = m_positionIndex;
    loca.l = m_locr.l();
    locr = loca;
    m_pGate->targetX(locr.x());
    m_pGate->targetY(locr.y());
    m_pGate->targetPos(locr.p());
  }
  else
  {
    m_pGate->localAction(true);
    m_pGate->localActionType(m_localActionType);
  };
  m_pGate->delay(m_delay);
  m_pGate->wallGraphicOrdinal(m_graphicListIndex);
}

void EditAndOr::Show(int ID, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(ID);
  if (pWnd!=NULL) pWnd->ShowWindow(show);
}

void EditAndOr::MarkActions(void)
{
  int j, k;
  j = 2*(m_action&3) + (m_NC?1:0);
  if (m_localAction)
  {
    j = 2 * 4 + (m_NC?1:0);
    if (m_localActionType == 10) j += 2;
  };
  k = legalActions[j];
  m_setConditionMet       = (k&MS) != 0;
  m_setConditionNotMet    = (k&NS) != 0;
  m_clearConditionMet     = (k&MC) != 0;
  m_clearConditionNotMet  = (k&NC) != 0;
  m_toggleConditionMet    = (k&MT) != 0;
  m_toggleConditionNotMet = (k&NT) != 0;
  m_rotateConditionMet    = (k&MR) != 0;
  m_rotateConditionNotMet = (k&NR) != 0;
  m_skillConditionMet     = (k&MK) != 0;
  m_skillConditionNotMet  = (k&NK) != 0;
}

void EditAndOr::DoDataExchange(CDataExchange* pDX)
{
  int i;
  WALLTYPE wall;
  if (!pDX->m_bSaveAndValidate)
  {
    MarkActions();
  };
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EditAndOr)
	DDX_Control(pDX, IDC_Graphic, m_graphic);
	DDX_Control(pDX, IDC_Position, m_position);
	DDX_Text(pDX, IDC_Column, m_column);
	DDV_MinMaxUInt(pDX, m_column, 0, 63);
	DDX_Text(pDX, IDC_Delay, m_delay);
	DDV_MinMaxUInt(pDX, m_delay, 0, 15);
	DDX_Text(pDX, IDC_InitialValue, m_initialValue);
	DDV_MinMaxUInt(pDX, m_initialValue, 0, 15);
	DDX_Check(pDX, IDC_once, m_once);
	DDX_Text(pDX, IDC_Row, m_row);
	DDV_MinMaxUInt(pDX, m_row, 0, 63);
	DDX_Text(pDX, IDC_TargetValue, m_targetValue);
	DDV_MinMaxUInt(pDX, m_targetValue, 0, 15);
	DDX_Check(pDX, IDC_ClearEqual, m_clearConditionMet);
	DDX_Check(pDX, IDC_ClearNotEqual, m_clearConditionNotMet);
	DDX_Check(pDX, IDC_SetEqual, m_setConditionMet);
	DDX_Check(pDX, IDC_SetNotEqual, m_setConditionNotMet);
	DDX_Check(pDX, IDC_ToggleEqual, m_toggleConditionMet);
	DDX_Check(pDX, IDC_ToggleNotEqual, m_toggleConditionNotMet);
	DDX_Check(pDX, IDC_RotateEqual, m_rotateConditionMet);
	DDX_Check(pDX, IDC_RotateNotEqual, m_rotateConditionNotMet);
	DDX_Radio(pDX, IDC_NoEdgeDetect, m_noEdgeDetect);
	DDX_Check(pDX, IDC_SkillEqual, m_skillConditionMet);
	DDX_Check(pDX, IDC_SkillNotEqual, m_skillConditionNotMet);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    m_position.SetCurSel(m_positionIndex);
    m_graphic.ResetContent();
    m_graphic.AddString("none");
    for (i=0; i<pDoc->NumWallDecorations(m_locr.l()); i++)   
    {
      wall = pDoc->WallDecorationType(m_locr.l(),i);
      m_graphic.AddString(wall.Name());
    };
    m_graphic.SetCurSel(m_graphicListIndex);
    Show(IDC_Column,        !m_localAction);
    Show(IDC_ColumnLabel,   !m_localAction);
    Show(IDC_Row,           !m_localAction);
    Show(IDC_RowLabel,      !m_localAction);
    Show(IDC_Position,      !m_localAction);
    Show(IDC_PositionLabel, !m_localAction);
    Show(IDC_Delay,         !m_localAction);
    Show(IDC_DelayLabel,    !m_localAction);
  };
  if (pDX->m_bSaveAndValidate)
  {
    m_graphicListIndex = m_graphic.GetCurSel();
    m_positionIndex = m_position.GetCurSel();
  };
}


BEGIN_MESSAGE_MAP(EditAndOr, CDialog)
	//{{AFX_MSG_MAP(EditAndOr)
	ON_BN_CLICKED(IDC_ClearEqual, OnClearEqual)
	ON_BN_CLICKED(IDC_ClearNotEqual, OnClearNotEqual)
	ON_BN_CLICKED(IDC_SetEqual, OnSetEqual)
	ON_BN_CLICKED(IDC_SetNotEqual, OnSetNotEqual)
	ON_BN_CLICKED(IDC_ToggleEqual, OnToggleEqual)
	ON_BN_CLICKED(IDC_ToggleNotEqual, OnToggleNotEqual)
	ON_BN_CLICKED(IDC_RotateEqual, OnRotateEqual)
	ON_BN_CLICKED(IDC_RotateNotEqual, OnRotateNotEqual)
	ON_BN_CLICKED(IDC_Browse, OnBrowse)
	ON_BN_CLICKED(IDC_SkillEqual, OnSkillEqual)
	ON_BN_CLICKED(IDC_SkillNotEqual, OnSkillNotEqual)
	ON_WM_HELPINFO()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditAndOr message handlers

int ACTUATOR_AndOr::Edit(LOCATIONREL locr, RN RNandOr)
{
  EditAndOr dlg(locr, RNandOr);
  dlg.m_pGate = this;
  return dlg.DoModal();
}


void EditAndOr::OnOK() 
{
	// TODO: Add extra validation here
  UpdateData(true);
	EncodeDataIntoRecord();
	CDialog::OnOK();
}

int EditAndOr::DoModal() 
{
	// TODO: Add your specialized code here and/or call the base class
	ExtractDataFromRecord();
	return CDialog::DoModal();
}

void EditAndOr::UpdateAction(int mask)
{
  int i, k;
  int besti, goodi;
  if (mask != 0) UpdateData(true);
  k = 0;
  if (m_setConditionMet)       k += MS;
  if (m_setConditionNotMet)    k += NS;
  if (m_clearConditionMet)     k += MC;
  if (m_clearConditionNotMet)  k += NC;
  if (m_toggleConditionMet)    k += MT;
  if (m_toggleConditionNotMet) k += NT;
  if (m_rotateConditionMet)    k += MR;
  if (m_rotateConditionNotMet) k += NR;
  if (m_skillConditionMet)     k += MK;
  if (m_skillConditionNotMet)  k += NK;
  for (i=0; i<12; i++)
  {
    if (legalActions[i] != k) continue;
    m_action = (ACTUATORACTION)((i/2) & 3);
    m_NC = (i % 2) != 0;
    m_localAction = i >= 8;
    UpdateData(false);
    return;
  };
  besti = -1;
  goodi = -1;
  for (i=0; i<12; i++)
  {
    if (((legalActions[i] ^ k) & mask) == 0)
    {
      if (goodi == -1) goodi = i;
        if (besti == -1) besti = i;
    };
  };
  i = -1;
  if (goodi != -1) i = goodi;
  if (besti != -1) i = besti;
  if (i != -1)
  {
    m_action      = (ACTUATORACTION)((i/2) & 3);
    m_NC          = (i % 2) != 0;
    m_localAction = i >= 8;
    m_localActionType = (i>=10)?10:2;
    UpdateData(false);
    return;
  };
  m_NC = false;
  m_action = (ACTUATORACTION)0;
  UpdateData(false);
}


void EditAndOr::OnClearEqual() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(MC);	
}

void EditAndOr::OnClearNotEqual() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(NC);
}

void EditAndOr::OnSetEqual() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(MS);
}

void EditAndOr::OnSetNotEqual() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(NS);
}

void EditAndOr::OnToggleEqual() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(MT);
}

void EditAndOr::OnToggleNotEqual() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(NT);
}

void EditAndOr::OnRotateEqual() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(MR);	
}

void EditAndOr::OnRotateNotEqual() 
{
	// TODO: Add your control notification handler code here
	UpdateAction(NR);
}

void EditAndOr::OnSkillEqual() 
{
	// TODO: Add your control notification handler code here
	UpdateAction(MK);
	
}

void EditAndOr::OnSkillNotEqual() 
{
	// TODO: Add your control notification handler code here
	UpdateAction(NK);
	
}

BOOL EditAndOr::OnInitDialog() 
{
  CString title;
  LOCATIONABS loca;
	CDialog::OnInitDialog();
  loca = m_locr;
  title.Format("Edit And/Or Gate [%04X] %02d(%02d,%02d)%d",
                  m_RNandOr.ConvertToInteger(),
                  loca.l, loca.x, loca.y, loca.p);
  SetWindowText(title);

	// TODO: Add extra initialization here
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EditAndOr::OnBrowse() 
{
	// TODO: Add your control notification handler code here
  LOCATIONABS loca, locaAndOr;
  LOCATIONREL locrTarget;
  UpdateData(true);
  loca.l = m_locr.l();
  loca.x = m_column;
  loca.y = m_row;
  locrTarget = loca;
  locaAndOr = m_locr;
	EditTarget dlg(locrTarget,"And/Or Gate",locaAndOr,false);
  if (dlg.DoModal() == IDOK)
  {
    locrTarget.l(dlg.m_locr.l());
    locrTarget.x(dlg.m_locr.x());
    locrTarget.y(dlg.m_locr.y());
    loca = locrTarget;
    m_column = loca.x;
    m_row = loca.y;
    UpdateData(false);
  };	
}


BOOL EditAndOr::OnHelpInfo(HELPINFO* pHelpInfo) 
{
	// TODO: Add your message handler code here and/or call default
	CString s;
  int id;
  switch (pHelpInfo->iCtrlId)
  {
  case IDC_Browse: id = IDH_EditAndOr_Browse; break;
  default: id = IDH_Context_None;
  };
	s.LoadString(id);
  HelpMsg(s);
  return true;
}
