// EditCounter.cpp : implementation file
//

#include "stdafx.h"
//#include "types.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "Actuator.h"
#include "EditCounter.h"
#include "misc.h"
#include "EditTarget.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// EditCounter dialog


EditCounter::EditCounter(LOCATIONREL locr, CWnd* pParent /*=NULL*/)
	: CDialog(EditCounter::IDD, pParent)
{
	//{{AFX_DATA_INIT(EditCounter)
	m_once = FALSE;
	m_initialValue = 0;
	m_row = 0;
	m_delay = 0;
	m_clearConditionNotMet = FALSE;
	m_clearConditionMet = FALSE;
	m_setConditionNotMet = FALSE;
	m_setConditionMet = TRUE;
	m_toggleConditionMet = FALSE;
	m_column = 0;
	m_rotateConditionNotMet = FALSE;
	m_rotateConditionMet = FALSE;
	m_skillConditionNotMet = FALSE;
	m_skillConditionMet = FALSE;
	m_graphicTitle = _T("");
	//}}AFX_DATA_INIT
  m_locr = locr;
  m_levelInfo = m_locr;
}

enum SPECIFICATIONS
{
  MS = 0x001,  //Set          - condition met
  NS = 0x002,  //Set          - condition Not met
  MC = 0x004,  //Clear        - condition Met
  NC = 0x008,  //Clear        - condition Not met
  MT = 0x010,  //Toggle       - condition Met
  NT = 0x020,  //Toggle       - condition Not met
  MR = 0x040,  //Rotate       - condition Met
  NR = 0x080,  //Rotate       - condition Not met
  MK = 0x100,  //sKill adjust - condition Met
  NK = 0x200   //sKill adjust - condition Not met
};

struct LEGALACTION
{
  int            buttons;
  ACTUATORACTION action;
  bool           NC;
  bool           localAction;
  int            localActionType;
};

static LEGALACTION legalActions[] =
{//buttons action    NC    local type
  {MS,     AA_SET,   false,false, -1}, // Set conditon Met
  {MC,     AA_CLEAR, false,false, -1}, // Clear condition Met
  {MT,     AA_TOGGLE,false,false, -1}, // Toggle condition Met
  {MS+NC,  AA_HOLD,  false,false, -1}, // Hold condition Met
  {MC+NS,  AA_HOLD,  true, false, -1}, // Hold condition Not met
  {MR,     AA_SET,   false,true,   2}, // Rotate condition Met
  {MR+NR,  AA_HOLD,  false,true,   2}, // Rotate unconditionally
  {MK,     AA_SET,   false,true,  10}, // sKill condition Met
  {MK+NK,  AA_HOLD,  false,true,  10}  // sKill unconditionally
};

#define NUMLEGAL (sizeof(legalActions)/sizeof(legalActions[0]))

void EditCounter::SetButtons(int k)
{
  m_setConditionMet       = (k & MS) != 0;
  m_setConditionNotMet    = (k & NS) != 0;
  m_clearConditionMet     = (k & MC) != 0;
  m_clearConditionNotMet  = (k & NC) != 0;
  m_toggleConditionMet    = (k & MT) != 0;
  m_skillConditionMet     = (k & MK) != 0;
  m_skillConditionNotMet  = (k & NK) != 0;
  m_rotateConditionMet    = (k & MR) != 0;
  m_rotateConditionNotMet = (k & NR) != 0;
}

int EditCounter::GetButtons(void)
{
  int k;
  k = 0;
  if (m_setConditionMet)       k += MS;
  if (m_setConditionNotMet)    k += NS;
  if (m_clearConditionMet)     k += MC;
  if (m_clearConditionNotMet)  k += NC;
  if (m_toggleConditionMet)    k += MT;
  if (m_rotateConditionMet)    k += MR;
  if (m_rotateConditionNotMet) k += NR;
  if (m_skillConditionMet)     k += MK;
  if (m_skillConditionNotMet)  k += NK;
  return k;
}
void EditCounter::ExtractDataFromRecord(void)
{
  int i, k;
  m_setConditionMet       = true;
  m_setConditionNotMet    = false;
  m_clearConditionMet     = false;
  m_clearConditionNotMet  = false;
  m_toggleConditionMet    = false;
  m_skillConditionMet     = false;
  m_skillConditionNotMet  = false;
  m_rotateConditionMet    = false;
  m_rotateConditionNotMet = false;
  m_action      = m_pCounter->action();
  m_NC          = m_pCounter->normallyClosed();
  m_localAction = m_pCounter->localAction();
  if (m_localAction)
  {
    m_localActionType = m_pCounter->localActionType();
  };
  for (i=0; i<NUMLEGAL; i++)
  {
    if (m_action          != legalActions[i].action)          continue;
    if (m_NC              != legalActions[i].NC)              continue;
    if (m_localAction     != legalActions[i].localAction)     continue;
    if (   m_localAction 
        && (m_localActionType != legalActions[i].localActionType)) continue;
    k = legalActions[i].buttons;
    SetButtons(k);
  };
  m_once = m_pCounter->onceOnly();
  m_initialValue = m_pCounter->counterValue();
  if (m_localAction)
  {
    m_localActionType = m_pCounter->localActionType();
  }
  else
  {
    m_column = m_pCounter->targetX() + m_levelInfo.left;
    m_row = m_pCounter->targetY() + m_levelInfo.top;
    m_pos = m_pCounter->targetPos();
  };
  m_delay = m_pCounter->delay();
  m_graphicListIndex = m_pCounter->wallGraphicOrdinal();
}

void EditCounter::EncodeDataIntoRecord(void)
{
  int i;
  int k;
  k = GetButtons();
  for (i=0; i<NUMLEGAL; i++)
  {
    if (legalActions[i].buttons == k) break;
  };
  if (i >= NUMLEGAL) i = 0;
  m_action          = legalActions[i].action;
  m_NC              = legalActions[i].NC;
  m_localAction     = legalActions[i].localAction;
  m_localActionType = legalActions[i].localActionType;
  m_pCounter->action(m_action);
  m_pCounter->onceOnly(m_once != 0);
  m_pCounter->counterValue(m_initialValue);
  m_pCounter->localAction(m_localAction);
  m_pCounter->normallyClosed(m_NC);
  if (m_localAction)
  {
    m_pCounter->localActionType(m_localActionType);
  }
  else
  {
    m_pCounter->targetX(m_column - m_levelInfo.left);
    m_pCounter->targetY(m_row - m_levelInfo.top);
    m_pCounter->targetPos(m_pos);
  };
  m_pCounter->delay(m_delay);
  m_pCounter->wallGraphicOrdinal(m_graphicListIndex);
}

void EditCounter::Show(int ID, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(ID);
  if (pWnd != NULL) pWnd->ShowWindow(show);
}
 
void EditCounter::DoDataExchange(CDataExchange* pDX)
{
  int i;
  WALLTYPE wall;
  FLOORTYPE floor;
  if (!pDX->m_bSaveAndValidate)
  {
    Show(IDC_Column,     !m_localAction);
    Show(IDC_Row,        !m_localAction);
    Show(IDC_Position,   !m_localAction);
    Show(IDC_ColumnTitle,!m_localAction);
    Show(IDC_RowTitle,   !m_localAction);
    Show(IDC_PosTitle,   !m_localAction);
  };
  if (RoomType(m_locr) == roomSTONE)
  {
    m_graphicTitle = "Wall Decoration";
  }
  else
  {
    m_graphicTitle = "Floor Decoration";
  };
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(EditCounter)
	DDX_Control(pDX, IDC_Graphic, m_graphic);
	DDX_Control(pDX, IDC_Position, m_position);
	DDX_Check(pDX, IDC_once, m_once);
	DDX_Text(pDX, IDC_InitialValue, m_initialValue);
	DDV_MinMaxInt(pDX, m_initialValue, 0, 511);
	DDX_Text(pDX, IDC_Row, m_row);
	DDV_MinMaxUInt(pDX, m_row, 0, 999999);
	DDX_Text(pDX, IDC_Delay, m_delay);
	DDV_MinMaxUInt(pDX, m_delay, 0, 31);
	DDX_Check(pDX, IDC_ClearNonZero, m_clearConditionNotMet);
	DDX_Check(pDX, IDC_ClearZero, m_clearConditionMet);
	DDX_Check(pDX, IDC_SetNonZero, m_setConditionNotMet);
	DDX_Check(pDX, IDC_SetZero, m_setConditionMet);
	DDX_Check(pDX, IDC_ToggleZero, m_toggleConditionMet);
	DDX_Text(pDX, IDC_Column, m_column);
	DDV_MinMaxUInt(pDX, m_column, 0, 99999);
	DDX_Check(pDX, IDC_RotateNonZero, m_rotateConditionNotMet);
	DDX_Check(pDX, IDC_RotateZero, m_rotateConditionMet);
	DDX_Check(pDX, IDC_SkillNonZero, m_skillConditionNotMet);
	DDX_Check(pDX, IDC_SkillZero, m_skillConditionMet);
	DDX_Text(pDX, IDC_GraphicTitle, m_graphicTitle);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    m_position.SetCurSel(m_pos);
    m_graphic.ResetContent();
    m_graphic.AddString("none");
    if (RoomType(m_locr) == roomSTONE)
    {
      for (i=0; i<pDoc->NumWallDecorations(m_levelInfo.level); i++)   
      {
        wall = pDoc->WallDecorationType(m_levelInfo.level,i);
        m_graphic.AddString(wall.Name());
      };
    }
    else
    {
      for (i=1; i<=pDoc->NumFloorDecorations(m_levelInfo.level); i++)   
      {
        floor = pDoc->FloorDecorationType(m_levelInfo.level,i);
        m_graphic.AddString(floor.Name());
      };
    };
    m_graphic.SetCurSel(m_graphicListIndex);
  };
  if (pDX->m_bSaveAndValidate)
  {
    m_graphicListIndex = m_graphic.GetCurSel();
    m_pos = m_position.GetCurSel();
  };
}


BEGIN_MESSAGE_MAP(EditCounter, CDialog)
	//{{AFX_MSG_MAP(EditCounter)
	ON_BN_CLICKED(IDC_SetNonZero, OnSetNonZero)
	ON_BN_CLICKED(IDC_ClearNonZero, OnClearNonZero)
	ON_BN_CLICKED(IDC_ClearZero, OnClearZero)
	ON_BN_CLICKED(IDC_SetZero, OnSetZero)
	ON_BN_CLICKED(IDC_ToggleZero, OnToggleZero)
	ON_BN_CLICKED(IDC_SkillNonZero, OnSkillNonZero)
	ON_BN_CLICKED(IDC_SkillZero, OnSkillZero)
	ON_BN_CLICKED(IDC_RotateNonZero, OnRotateNonZero)
	ON_BN_CLICKED(IDC_RotateZero, OnRotateZero)
	ON_BN_CLICKED(IDC_Browse, OnBrowse)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditCounter message handlers

int ACTUATOR_Counter::Edit(LOCATIONREL locr)
{
  EditCounter dlg(locr);
  dlg.m_pCounter = this;
  return dlg.DoModal();
}

void EditCounter::UpdateAction(int mask)
{
  int i, k;
  int besti, goodi;
  if (mask != 0) UpdateData(true);
  k = GetButtons();
  for (i=0; i<NUMLEGAL; i++)
  {
    if (legalActions[i].buttons != k) continue;
    m_action = legalActions[i].action;
    m_NC = legalActions[i].NC;
    m_localAction = legalActions[i].localAction;
    m_localActionType = legalActions[i].localActionType;
    SetButtons(legalActions[i].buttons);
    UpdateData(false);
    return;
  };
  besti = -1;
  goodi = -1;
  for (i=0; i<NUMLEGAL; i++)
  {
    if (((legalActions[i].buttons ^ k) & mask) == 0)
    {
      if (goodi == -1) goodi = i;
        if (besti == -1) besti = i;
    };
  };
  i = -1;
  if (goodi != -1) i = goodi;
  if (besti != -1) i = besti;
  if (i != -1)
  {
    m_action = legalActions[i].action;
    m_NC = legalActions[i].NC;
    m_localAction = legalActions[i].localAction;
    m_localActionType = legalActions[i].localActionType;
    SetButtons(legalActions[i].buttons);
    UpdateData(false);
    return;
  };
  m_NC = false;
  m_action = (ACTUATORACTION)0;
  UpdateData(false);
}

void EditCounter::OnSetNonZero() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(NS);
}

void EditCounter::OnClearNonZero() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(NC);
}

void EditCounter::OnClearZero() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(MC);
}

void EditCounter::OnSetZero() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(MS);
}

void EditCounter::OnToggleZero() 
{
	// TODO: Add your control notification handler code here	
  UpdateAction(MT);
}

void EditCounter::OnSkillNonZero() 
{
	// TODO: Add your control notification handler code here
  UpdateAction(NK);	
}

void EditCounter::OnSkillZero() 
{
	// TODO: Add your control notification handler code here
	UpdateAction(MK);
}

void EditCounter::OnRotateNonZero() 
{
	// TODO: Add your control notification handler code here
	UpdateAction(NR);
}

void EditCounter::OnRotateZero() 
{
	// TODO: Add your control notification handler code here
	UpdateAction(MR);
}

BOOL EditCounter::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE

}

void EditCounter::OnOK() 
{
	// TODO: Add extra validation here
  CString msg;
  UpdateData(true);
  if (   (m_row < (unsigned)m_levelInfo.top)
      || (m_row > (unsigned)m_levelInfo.bottom) )
  {
    msg.Format("Row must be a number\nbetween %d and %d",
              m_levelInfo.top, m_levelInfo.bottom); 
    Error((LPCTSTR)msg);
    return;
  };
	EncodeDataIntoRecord();
	CDialog::OnOK();
}

int EditCounter::DoModal() 
{
	// TODO: Add your specialized code here and/or call the base class
	ExtractDataFromRecord();
	return CDialog::DoModal();
}


void EditCounter::OnBrowse() 
{
	// TODO: Add your control notification handler code here
	
  LOCATIONABS loca, locaCounter;
  LOCATIONREL locrTarget;
  UpdateData(true);
  loca.l = m_locr.l();
  loca.x = m_column;
  loca.y = m_row;
  locrTarget = loca;
  locaCounter = m_locr;
	EditTarget dlg(locrTarget,"Counter",locaCounter,false);
  if (dlg.DoModal() == IDOK)
  {
    locrTarget.l(dlg.m_locr.l());
    locrTarget.x(dlg.m_locr.x());
    locrTarget.y(dlg.m_locr.y());
    loca = locrTarget;
    m_column = loca.x;
    m_row = loca.y;
    UpdateData(false);
  };	
}
