// EditSwapObject.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "CSBuildDoc.h"
#include "EditSwapObject.h"
#include "Actuator.h"
#include "EditTarget.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// EditSwapObject dialog


EditSwapObject::EditSwapObject(LOCATIONREL locr,
                               CWnd* pParent /*=NULL*/)
	: CDialog(EditSwapObject::IDD, pParent)
{
  m_locr = locr;
	//{{AFX_DATA_INIT(EditSwapObject)
	m_column = 0;
	m_row = 0;
	m_rotate = FALSE;
	m_onceOnly = FALSE;
	m_audible = FALSE;
	m_clear = FALSE;
	m_set = FALSE;
	m_skill = FALSE;
	m_toggle = FALSE;
	m_delay = 0;
	m_position = -1;
	//}}AFX_DATA_INIT
}

enum SPECIFICATIONS
{
  MS = 0x001,  //Set          - condition met
  MC = 0x004,  //Clear        - condition Met
  MT = 0x010,  //Toggle       - condition Met
  MR = 0x040,  //Rotate       - condition Met
  MK = 0x100,  //sKill adjust - condition Met
};

struct LEGALACTION
{
  int            buttons;
  ACTUATORACTION action;
  bool           NC;
  bool           localAction;
  int            localActionType;
};

static LEGALACTION legalActions[] =
{//buttons action    NC    local type
  {MS,     AA_SET,   false,false, -1}, // Set conditon Met
  {MC,     AA_CLEAR, false,false, -1}, // Clear condition Met
  {MT,     AA_TOGGLE,false,false, -1}, // Toggle condition Met
  {MR,     AA_SET,   false,true,   2}, // Rotate condition Met
  {MK,     AA_SET,   false,true,  10}, // sKill condition Met
};

#define NUMLEGAL (sizeof(legalActions)/sizeof(legalActions[0]))


void EditSwapObject::ExtractDataFromRecord(void)
{
  int i, k;
  m_neededObject = m_pSwapObject->neededObject(m_locr);
  m_set       = true;
  m_clear     = false;
  m_toggle    = false;
  m_skill     = false;
  m_rotate    = false;
  m_action      = m_pSwapObject->action();
  m_localAction = m_pSwapObject->localAction();
  if (m_localAction)
  {
    m_localActionType = m_pSwapObject->localActionType();
  };
  for (i=0; i<NUMLEGAL; i++)
  {
    if (m_action          != legalActions[i].action)          continue;
    if (m_localAction     != legalActions[i].localAction)     continue;
    if (   m_localAction 
        && (m_localActionType != legalActions[i].localActionType)) continue;
    k = legalActions[i].buttons;
    SetButtons(k);
  };
  m_onceOnly = m_pSwapObject->onceOnly();
  m_audible  = m_pSwapObject->audible();
  if (m_localAction)
  {
    m_localActionType = m_pSwapObject->localActionType();
  }
  else
  {
    LOCATIONREL locr;
    LOCATIONABS loca;
    locr.l(m_locr.l());
    locr.x(m_pSwapObject->targetX());
    locr.y(m_pSwapObject->targetY());
    locr.p(m_pSwapObject->targetPos());
    loca = locr;
    m_column   = loca.x;
    m_row      = loca.y;
    m_position = loca.p;
  };
  m_delay = m_pSwapObject->delay();
  m_wallGraphicIndex = m_pSwapObject->wallGraphicOrdinal();
}

void EditSwapObject::EncodeDataIntoRecord(void)
{
  m_pSwapObject->delay(m_delay);
  m_pSwapObject->onceOnly(m_onceOnly?true:false);
  m_pSwapObject->audible(m_audible?true:false);
  m_pSwapObject->wallGraphicOrdinal(m_wallGraphicIndex);
  m_pSwapObject->neededObject(m_neededObject);
  m_pSwapObject->action(m_action);
  m_pSwapObject->localAction(m_localAction);
  if (m_localAction)
  {
    m_pSwapObject->localActionType(m_localActionType);
  }
  else
  {
    LOCATIONABS loca;
    LOCATIONREL locr;
    loca.l = locr.l();
    loca.x = m_column;
    loca.y = m_row;
    loca.p = m_position;
    locr = loca;
    m_pSwapObject->targetX(locr.x());
    m_pSwapObject->targetY(locr.y());
    m_pSwapObject->targetPos(locr.p());
  };
}

void EditSwapObject::SetButtons(int k)
{
  m_set       = (k & MS) != 0;
  m_clear     = (k & MC) != 0;
  m_toggle    = (k & MT) != 0;
  m_skill     = (k & MK) != 0;
  m_rotate    = (k & MR) != 0;
}

int EditSwapObject::GetButtons(void)
{
  int k;
  k = 0;
  if (m_set)       k += MS;
  if (m_clear)     k += MC;
  if (m_toggle)    k += MT;
  if (m_rotate)    k += MR;
  if (m_skill)     k += MK;
  return k;
}

void EditSwapObject::Show(int id, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(id);
  if (pWnd != NULL)
  {
    pWnd->ShowWindow(show);
  };
}

void EditSwapObject::DoDataExchange(CDataExchange* pDX)
{
  int i;
	CDialog::DoDataExchange(pDX);


	//{{AFX_DATA_MAP(EditSwapObject)
	DDX_Control(pDX, IDC_ActivatedBy, m_activatedBy);
	DDX_Control(pDX, IDC_Graphic, m_graphic);
	DDX_Text(pDX, IDC_Column, m_column);
	DDX_Text(pDX, IDC_Row, m_row);
	DDX_Check(pDX, IDC_Rotate, m_rotate);
	DDX_Check(pDX, IDC_OnceOnly, m_onceOnly);
	DDX_Check(pDX, IDC_Audible, m_audible);
	DDX_Check(pDX, IDC_Clear, m_clear);
	DDX_Check(pDX, IDC_Set, m_set);
	DDX_Check(pDX, IDC_Skill, m_skill);
	DDX_Check(pDX, IDC_Toggle, m_toggle);
	DDX_Text(pDX, IDC_Delay, m_delay);
	DDV_MinMaxUInt(pDX, m_delay, 0, 15);
	DDX_CBIndex(pDX, IDC_Position, m_position);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    WALLTYPE wall;
    m_graphic.ResetContent();
    m_graphic.AddString("none");
    for (i=0; i<pDoc->NumWallDecorations(m_locr.l()); i++)   
    {
      wall = pDoc->WallDecorationType(m_locr.l(),i);
      m_graphic.AddString(wall.Name());
    };
    m_graphic.SetCurSel(m_wallGraphicIndex);
    m_activatedBy.SetCurSel(m_neededObjectIndex);
    Show(IDC_Row,           !m_localAction);
    Show(IDC_Column,        !m_localAction);
    Show(IDC_Position,      !m_localAction);
    Show(IDC_RowTitle,      !m_localAction);
    Show(IDC_ColumnTitle,   !m_localAction);
    Show(IDC_PositionTitle, !m_localAction);
    Show(IDC_Browse,        !m_localAction);
  };
  if (pDX->m_bSaveAndValidate)
  {
    m_wallGraphicIndex = m_graphic.GetCurSel();
    m_neededObjectIndex = m_activatedBy.GetCurSel();
    m_neededObject = m_activatedBy.GetItemData(m_neededObjectIndex);
  };
}


BEGIN_MESSAGE_MAP(EditSwapObject, CDialog)
	//{{AFX_MSG_MAP(EditSwapObject)
	ON_BN_CLICKED(IDC_Clear, OnClear)
	ON_BN_CLICKED(IDC_Rotate, OnRotate)
	ON_BN_CLICKED(IDC_Set, OnSet)
	ON_BN_CLICKED(IDC_Skill, OnSkill)
	ON_BN_CLICKED(IDC_Toggle, OnToggle)
	ON_BN_CLICKED(IDC_Browse, OnBrowse)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// EditSwapObject message handlers

int ACTUATOR_SwapObject::Edit(LOCATIONREL locr)
{
  EditSwapObject dlg(locr);
  dlg.m_pSwapObject = this;
  return dlg.DoModal();
}

BOOL EditSwapObject::OnInitDialog() 
{
  ExtractDataFromRecord();
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here

  m_neededObjectIndex = m_neededObject.FillComboBox(&m_activatedBy);
  UpdateData(false);
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void EditSwapObject::OnOK() 
{
	// TODO: Add extra validation here
  UpdateData(true);
	EncodeDataIntoRecord();
	CDialog::OnOK();
}

void EditSwapObject::UpdateActions(int mask)
{
  int i, k;
  int besti, goodi;
  if (mask != 0) UpdateData(true);
  k = GetButtons();
  for (i=0; i<NUMLEGAL; i++)
  {
    if (legalActions[i].buttons != k) continue;
    m_action = legalActions[i].action;
    m_localAction = legalActions[i].localAction;
    m_localActionType = legalActions[i].localActionType;
    SetButtons(legalActions[i].buttons);
    UpdateData(false);
    return;
  };
  besti = -1;
  goodi = -1;
  for (i=0; i<NUMLEGAL; i++)
  {
    if (((legalActions[i].buttons ^ k) & mask) == 0)
    {
      if (goodi == -1) goodi = i;
        if (besti == -1) besti = i;
    };
  };
  i = -1;
  if (goodi != -1) i = goodi;
  if (besti != -1) i = besti;
  if (i != -1)
  {
    m_action = legalActions[i].action;
    m_localAction = legalActions[i].localAction;
    m_localActionType = legalActions[i].localActionType;
    SetButtons(legalActions[i].buttons);
    UpdateData(false);
    return;
  };
  m_action = (ACTUATORACTION)0;
  UpdateData(false);
}



void EditSwapObject::OnClear() 
{
	// TODO: Add your control notification handler code here
  UpdateActions(MC);	
}

void EditSwapObject::OnRotate() 
{
	// TODO: Add your control notification handler code here
	UpdateActions(MR);
}

void EditSwapObject::OnSet() 
{
	// TODO: Add your control notification handler code here
	UpdateActions(MS);
}

void EditSwapObject::OnSkill() 
{
	// TODO: Add your control notification handler code here
	UpdateActions(MK);
}

void EditSwapObject::OnToggle() 
{
	// TODO: Add your control notification handler code here
	UpdateActions(MT);
}

void EditSwapObject::OnBrowse() 
{
	// TODO: Add your control notification handler code here
  LOCATIONABS loca, locaAndOr;
  LOCATIONREL locrTarget;
  UpdateData(true);
  loca.l = m_locr.l();
  loca.x = m_column;
  loca.y = m_row;
  locrTarget = loca;
  locaAndOr = m_locr;
	EditTarget dlg(locrTarget,"Swap Object",locaAndOr,false);
  if (dlg.DoModal() == IDOK)
  {
    locrTarget.l(dlg.m_locr.l());
    locrTarget.x(dlg.m_locr.x());
    locrTarget.y(dlg.m_locr.y());
    loca = locrTarget;
    m_column = loca.x;
    m_row = loca.y;
    UpdateData(false);
  };	
}

	
