#include "stdafx.h"
//#include "types.h"
#include "CSBuild.h"
#include "ViewLevelInfo.h"
#include "CSBuildDoc.h"
#include "Misc.h"
#include "Actuator.h"
#include "GetImportLocation.h"




enum PID //Parameter id
{
  PID_Initialize = 1,
  PID_type,
  PID_hp,
  PID_position,
  PID_width,
  PID_height,
  PID_multiplier,
  PID_semiColon,
  PID_address,
  PID_active,
  PID_visible,
  PID_face,
  PID_rotate,
  PID_target,
  PID_rot90,
  PID_rot180,
  PID_rot270,
  PID_north,
  PID_east,
  PID_south,
  PID_west,
  PID_anything,
  PID_monsters,
  PID_partyAndObjects,
  PID_objects,
  PID_num,
  PID_upStairs,
  PID_downStairs,
  PID_eastWest,
  PID_northSouth,
  PID_open,
  PID_invisible,
  PID_Finalize
};

enum PT // Parameter type
{
  PT_none = 0,
  PT_UINT,  //unsigned int
  PT_ELOC,  // [+|1]n([+|-]n,[+|-]n)[n]
  PT_STRING,
  PT_INTARRAY
};

enum CT
{
  CT_COMMANDS = 1,
  CT_PARAMETERS
};

struct DATA;
class IMPORTFILE;

struct PPKT
{
  PID pid;
  PT  pt;
  IMPORTFILE *iFile;
  DATA *pData;
  unsigned int uint;
  signed int sint;
  EXPORTLOC eloc;
};

struct PDEF
{
  char *pName;
  PID  pid;
  PT   pt;
};

struct COMMANDS
{
  char *cmdName;
  PDEF *pDef;
  char *(*processParam)(PPKT *pPkt);
};


static bool Whitespace(char c)
{
  if (c == ' ') return true;
  if (c == '\t') return true;
  return false;
}

class IMPORTFILE
{
  FILE     *m_f;
  bool      m_eof;
  char     *m_inputLine;
  int       m_inputLineLen;
  int       m_lineNum;
  int       m_column;
 
  union
  {
    COMMANDS *m_pCmds[10];
    PDEF     *m_pDefs[10];
  } m_pt; //Parse table
  CT        m_cmdType[10];
  int       m_cmdIndex;
  CString   m_curField;
  bool      m_error;
  PPKT      m_pPkt;
  char   *(*m_processParam)(PPKT *);
  bool      ExtractParameter(PDEF *pDef);
  bool      ConvertInt(bool *isSigned, bool msgIfError);
  bool      ConvertUINT(bool msgIfError);
  bool      ConvertELOC(bool msgIfError);
  bool      ExpectOp(char c, bool msgIfError); //Finishes with NextField();
public:
  IMPORTFILE(FILE *f, DATA *pData);
  ~IMPORTFILE(void);
  void Read(void);
  bool NextField(bool msgIfError);
  CString CurField(void);
  bool Error(char *msg, bool msgIfError);
  bool Parse(COMMANDS *pCmds);
  void PopCommand(void);
  void PushCommand(COMMANDS *pCmd);
};

IMPORTFILE::IMPORTFILE(FILE *f, DATA *pData)
{
  m_f = f;
  m_eof = false;
  m_inputLine = NULL;
  m_lineNum = 0;
  m_column = 0;
  m_cmdIndex = 0;
  m_pPkt.pData = pData;
  m_pPkt.iFile = this;
  m_error = false;
}

IMPORTFILE::~IMPORTFILE(void)
{
  if (m_inputLine != NULL)
  {
    free(m_inputLine);
    m_inputLine = NULL;
  };
}  

bool IMPORTFILE::ConvertUINT(bool msgIfError)
{
  int i;
  unsigned int result = 0;
  for (i=0; i<m_curField.GetLength(); i++)
  {
    if (   (m_curField[i] < '0')
        || (m_curField[i] > '9')
       ) Error("Expected an unsigned integer", msgIfError);
    result = 10 * result + m_curField[i] - '0';
  };
  m_pPkt.uint = result;
  return true;
}

bool IMPORTFILE::ConvertInt(bool *isSigned, bool msgIfError)
{
  int i=0;
  int result = 0;
  int sgn = +1;
  if (m_curField[i] == '+') 
  {
    *isSigned = true;
    i=1;
  };
  if (m_curField[i] == '-') 
  {
    *isSigned = true;
    i=1;
    sgn = -1;
  };
  for (; i<m_curField.GetLength(); i++)
  {
    if (   (m_curField[i] < '0')
        || (m_curField[i] > '9')
       ) Error("Expected an unsigned integer", msgIfError);
    result = 10 * result + m_curField[i] - '0';
  };
  m_pPkt.sint = sgn * result;
  return true;
}

bool IMPORTFILE::ConvertELOC(bool msgIfError)
{
  // [+|-]n([+|-]n,[+|-]n)n
  bool isSigned;
  m_pPkt.eloc.left=-1; //Not relative
  m_pPkt.eloc.top=-1;  //Not relative
  m_pPkt.eloc.level=-1;//Not relative
  if (!ConvertInt(&isSigned, msgIfError)) 
                    return false; 
  m_pPkt.eloc.dl = m_pPkt.sint;
  if (isSigned) m_pPkt.eloc.level=0;//Relative
  if (!ExpectOp('(', msgIfError)) return false;
  if (!ConvertInt(&isSigned, msgIfError)) 
                    return false; 
  m_pPkt.eloc.dx = m_pPkt.sint;
  if (isSigned) m_pPkt.eloc.left=0;//Relative
  if (!ExpectOp(',', msgIfError)) return false;
  if (!ConvertInt(&isSigned, msgIfError)) 
                    return false; 
  m_pPkt.eloc.dy = m_pPkt.sint;
  if (isSigned) m_pPkt.eloc.top=0;//Relative
  if (!ExpectOp(')', msgIfError)) return false;
  if (!ConvertUINT(msgIfError)) 
                    return false; 
  m_pPkt.eloc.pos = m_pPkt.uint;
  if (m_pPkt.uint > 3) return Error("positon > 3", msgIfError);
  return true;
};  
  
void IMPORTFILE::Read()
{
  int lineLen;
  if (m_inputLine == NULL)
  {
    m_inputLineLen = 10;
    m_inputLine = (char *)malloc(m_inputLineLen, MALLOC096);
  };
  do
  {
    lineLen = 0;
    for (;;)
    {
      m_inputLine[m_inputLineLen-1] = 99; //So we can see if it disappears
      if (fgets(m_inputLine+lineLen, m_inputLineLen-lineLen, m_f) == NULL)
      {
        if (lineLen != 0) 
        {
          break; //We are done reading a line.
        }
        else
        {
          m_eof = true;
        };
      };
      if (m_inputLine[m_inputLineLen-1] == 99) break;
      if (m_inputLine[m_inputLineLen-2] == '\n') break;
      lineLen = m_inputLineLen-1;
      m_inputLineLen += 10;
      m_inputLine = (char *)realloc(m_inputLine, m_inputLineLen, MALLOC096);
    };
    m_lineNum++;
  } while (m_inputLine[0] == '#');
  m_column = 0;
  return;
}

bool IMPORTFILE::Error(char *msg, bool msgIfError)
{ //ALWAYS returns false.
  m_error = true;
  if (!msgIfError) return false;
  CString message;
  message.Format("%s\nError occurred at line %d column %d",
                 msg, m_lineNum, m_column+1);
  MessageBox(NULL,
             message,
             "Import Error",
             MB_OK | MB_TASKMODAL);
  return false;
}

CString IMPORTFILE::CurField(void)
{
  return m_curField;
}

bool IMPORTFILE::ExpectOp(char c, bool msgIfError)
{
  if (!NextField(msgIfError)) return false;
  if (m_curField != c)
  {
    if (msgIfError)
    {
      char msg[50];
      sprintf(msg,"Expected '%c'",c);
      Error(msg, true);
    };
    return false;
  };
  return NextField(msgIfError);
}

void IMPORTFILE::PopCommand(void)
{
  if (m_cmdIndex < 0)
  {
    Error("Command Stack underflow", true);
  }
  else
  {
    m_cmdIndex--;
    if (m_cmdType[m_cmdIndex] == CT_COMMANDS)
    {
      m_processParam = m_pt.m_pCmds[m_cmdIndex]->processParam;
    };
    if (m_column != 0) Read();
  };
}

bool IMPORTFILE::NextField(bool msgIfError)
{
  int count;
  char c;
  if (m_eof) return Error("Unexpected End-of-File", msgIfError);
  while (Whitespace(m_inputLine[m_column])) m_column++;
  if (   (m_inputLine[m_column] == 0)
      || (m_inputLine[m_column] == '\n')
     )  return Error("Missing information", msgIfError);
  switch (m_inputLine[m_column])
  {
  case '(':
  case ')':
  case ',':
  case '=':
  case ';':
    m_curField = CString(m_inputLine[m_column], 1);
    m_column++;
    return true;
  };
  count = 0;
  do
  {
    c = m_inputLine[m_column + count];
    count++;
  } while (   (c != 0)
           && (c != '\n')
           && (c != '=')
           && (c != ')')
           && (c != '(')
           && (c != ',')
           && (c != ';')
           && (!Whitespace(c)) );
  count--;
  m_curField = CString(m_inputLine+m_column, count);
  m_curField.MakeLower();
  m_column += count;
  return true;
}

bool IMPORTFILE::ExtractParameter(PDEF *pDef)
{
  switch (pDef->pt)
  {
  case PT_UINT:
    if (!NextField(true)) return false;
    if (m_curField != "=")
    {
      Error("Expected '='", true);
      return false;
    };
    if (!NextField(true)) return false;
    if (!ConvertUINT(true)) return false;
    break;
  case PT_none:
    break;
  case PT_ELOC:
    // [+|-]n([+|-]n,[+|-]n)[n]
    if (!NextField(true)) return false;
    if (m_curField != "=")
    {
      Error("Expected '='", true);
      return false;
    };
    if (!NextField(true)) return false;
    if (!ConvertELOC(true)) return false;
    break;
  case PT_STRING:
    NotImplemented();
    break;
  default: Error("Not Implemented parameter type", true);
           return false;
  };
  return true;
}

bool IMPORTFILE::Parse(COMMANDS *pCmds)
{
  COMMANDS *pCmd;
  char *result;
  PDEF     *pDef;
  m_pt.m_pCmds[0] = pCmds;
  m_cmdType[0] = CT_COMMANDS;
  m_cmdIndex = 0;
  Read();
  do
  {
    if (!NextField(true))
    {
      return false;
    };
    if (m_cmdType[m_cmdIndex] == CT_COMMANDS)
    {
      for (pCmd=m_pt.m_pCmds[m_cmdIndex]; pCmd->cmdName[0]!=0; pCmd++)
      {
        if (pCmd->cmdName == m_curField) break;
      };
      if (pCmd->cmdName[0] == 0)
      {
        Error("Unknown Command", true);
        return false;
      }
      m_pPkt.pid = PID_Initialize;
      m_processParam = pCmd->processParam;
      result = m_processParam(&m_pPkt);
      m_cmdIndex++;
      m_cmdType[m_cmdIndex] = CT_PARAMETERS;
      m_pt.m_pDefs[m_cmdIndex] = pCmd->pDef;
    }
    else
    {
      for (pDef=m_pt.m_pDefs[m_cmdIndex]; pDef->pName[0]!=0; pDef++)
      {
        if (pDef->pName == m_curField) break;
      };
      if (pDef->pName[0] == 0)
      {
        if (m_curField == ";")
        {
          char *(*tp)(PPKT *);
          tp = m_processParam; //save from PopCommand.
          PopCommand();
          m_pPkt.pid = PID_semiColon;
          result = tp(&m_pPkt);
        }
        else
        {
          Error("Unknown Parameter", true);
          return false;
        };
      }
      else
      {
        m_pPkt.pid = pDef->pid;
        if (!ExtractParameter(pDef)) 
        {
          return false;
        };
        result = m_processParam(&m_pPkt);
      };
    };
    if (result != NULL)
    {
      Error(result, true);
      return false;
    };
  } while (m_cmdIndex >= 0);
  return true;
}

void IMPORTFILE::PushCommand(COMMANDS *pCmds)
{
  m_cmdIndex++;
  m_pt.m_pCmds[m_cmdIndex] = pCmds;
  m_cmdType[m_cmdIndex] = CT_COMMANDS;
}

class RNLIST
{
  RNLIST *next;
public:
  RN object;
public:
  RNLIST(RNLIST *old, RN obj){next = old;object=obj;};
};

class TCELL
{
public:
   TCELL *nextCell;
   RNLIST *contents;
   EXPORTLOC address;
   int cf;
   TCELL(TCELL *next);
   ~TCELL(void);
   void SetCF(CELLFLAG cf, int x, int y);
   void AddObj(RN obj);
};

TCELL::TCELL(TCELL *next)
{
  nextCell = next;
  cf = roomILLEGAL<<5;
  contents = NULL;
}

TCELL::~TCELL(void)
{
  TCELL *temp;
  while (nextCell != NULL)
  {
    temp = nextCell;
    nextCell = temp->nextCell;
    temp->nextCell = NULL;
    delete temp;
  };
  if (contents != NULL) delete contents;
  contents = NULL;
}


void TCELL::AddObj(RN obj)
{
  contents = new RNLIST(contents, obj);
}


struct DATA
{
  int level;//destination of relative cell (0,0)
  int difficultyMultiplier;
  int left; //destination
  int top;  //destination
  int width;
  int height;
  DATA(void){width=-1; height=-1;firstCell=NULL;difficultyMultiplier=-1;};
  TCELL *firstCell;
  void NewCell(void);
  void SetCellType(ROOMTYPE rt);
  void SetCellflag(int flag);
  void ClearCellflag(int flag);
  void AddObject(RN obj);
  ROOMTYPE GetCellType(void);
  DB1 *Teleporter(void);
  DB4 *Monster(void);
  ~DATA(void);
};

DATA::~DATA(void)
{
  if (firstCell != NULL) delete firstCell;
  firstCell = NULL;
}

void DATA::NewCell(void)
{
  firstCell = new TCELL(firstCell);
}

void DATA::SetCellType(ROOMTYPE rt)
{
  if (firstCell == NULL) return;
  firstCell->cf &= 0x1f;
  firstCell->cf |= rt << 5;
}

void DATA::SetCellflag(int flag)
{
  if (firstCell == NULL) return;
  firstCell->cf |= flag;
}

void DATA::ClearCellflag(int flag)
{
  if (firstCell == NULL) return;
  firstCell->cf &= ~flag;
}

ROOMTYPE DATA::GetCellType(void)
{
  if (firstCell==NULL) return roomILLEGAL;
  return (ROOMTYPE)(firstCell->cf >> 5);
}

void DATA::AddObject(RN obj)
{
  RN objNext;
  firstCell->AddObj(obj);
}

DB1 *DATA::Teleporter(void)
{
  return pDoc->GetRecordAddressDB1(firstCell->contents->object);
}

DB4 *DATA::Monster(void)
{
  return pDoc->GetRecordAddressDB4(firstCell->contents->object);
}

static PDEF areaDef[] =
{
  {"width"     ,PID_width     ,PT_UINT},
  {"height"    ,PID_height    ,PT_UINT},
  {"multiplier",PID_multiplier,PT_UINT},
  {""}
};

static PDEF endAreaDef[] =
{
  {""},
};

static PDEF teleporterCFDef[] =
{
  {"active" ,PID_active  ,PT_none},
  {"visible",PID_visible ,PT_none},
  {""}
};

static PDEF stairsCFDef[] =
{
  {"up"         ,PID_upStairs   ,PT_none},
  {"down"       ,PID_downStairs ,PT_none},
  {"east/west"  ,PID_eastWest   ,PT_none},
  {"north/south",PID_northSouth ,PT_none},
  {""}
};

static PDEF pitCFDef[] =
{
  {"open"       ,PID_open       ,PT_none},
  {"invisible"  ,PID_invisible  ,PT_none},
  {""}
};


static PDEF teleporterRecDef[] =
{
  {"face"           ,PID_face           ,PT_STRING},
  {"rotate"         ,PID_rotate         ,PT_STRING},
  {"target"         ,PID_target         ,PT_ELOC},
  {"anything"       ,PID_anything       ,PT_none},
  {"monsters"       ,PID_monsters       ,PT_none},
  {"objects"        ,PID_objects        ,PT_none},
  {"partyandobjects",PID_partyAndObjects,PT_none},
  {"rot90"          ,PID_rot90          ,PT_none},
  {"rot270"         ,PID_rot270         ,PT_none},
  {"rot180"         ,PID_rot180         ,PT_none},
  {"north"          ,PID_north          ,PT_none},
  {"east"           ,PID_east           ,PT_none},
  {"south"          ,PID_south          ,PT_none},
  {"west"           ,PID_west           ,PT_none},
  {""}
};

static PDEF monsterRecDef[] =
{
  {"num"            ,PID_num            ,PT_UINT},
  {"type"           ,PID_type           ,PT_STRING},
  {"north"          ,PID_north          ,PT_none},
  {"east"           ,PID_east           ,PT_none},
  {"south"          ,PID_south          ,PT_none},
  {"west"           ,PID_west           ,PT_none},
  {"hp"             ,PID_hp             ,PT_INTARRAY},
  {"position"       ,PID_position       ,PT_INTARRAY},
  {""}
};



static PDEF cellDef[] =
{
  {"address" ,PID_address   ,PT_ELOC},
  {""}
};

static char *pArea(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    if (pPkt->pData->width != -1)
    {
      return "Multiple 'area' commands";
    };
    break;
  case PID_width:
    pPkt->pData->width = pPkt->uint;
    break;
  case PID_height:
    pPkt->pData->height = pPkt->uint;
    break;
  case PID_multiplier:
    pPkt->pData->difficultyMultiplier = pPkt->uint;
    break;
  case PID_semiColon:
    if (   (pPkt->pData->width == -1)
         ||(pPkt->pData->height == -1) )
    {
      return "Area size not specified";
    };
    break;
  default:
    return "Unexpected 'area' parameter";
  };
  return NULL;
}

PDEF endcellDef[] =
{
  {""}
};

PDEF openDef[] =
{
  {""}
};

PDEF stoneDef[] =
{
  {""}
};

static char *pOpen(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    pPkt->pData->SetCellType(roomOPEN);
    break;
  case PID_semiColon:
    break;
  default: return "Unknown 'open' Parameter";
  };
  return NULL;
}

static char *pStone(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    pPkt->pData->SetCellType(roomSTONE);
    break;
  case PID_semiColon:
    break;
  default: return "Unknown 'stone' Parameter";
  };
  return NULL;
}

static char *pCFTeleporter(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    pPkt->pData->SetCellType(roomTELEPORTER);
    break;
  case PID_active:
    pPkt->pData->SetCellflag(0x08);
    break;
  case PID_visible:
    pPkt->pData->SetCellflag(0x04);
    break;
  case PID_semiColon:
    break;
  default: return "Unknown 'teleporter' Parameter";
  };
  return NULL;
}


static char *pCFStairs(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    pPkt->pData->SetCellType(roomSTAIRS);
    break;
  case PID_upStairs:
    pPkt->pData->SetCellflag(0x04);
    break;
  case PID_downStairs:
    pPkt->pData->ClearCellflag(0x04);
    break;
  case PID_eastWest:
    pPkt->pData->ClearCellflag(0x08);
    break;
  case PID_northSouth:
    pPkt->pData->SetCellflag(0x08);
    break;
  case PID_semiColon:
    break;
  default: return "Unknown 'stairs' Parameter";
  };
  return NULL;
}

static char *pCFPit(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    pPkt->pData->SetCellType(roomPIT);
    break;
  case PID_open:
    pPkt->pData->SetCellflag(0x04);
    break;
  case PID_invisible:
    pPkt->pData->ClearCellflag(0x04);
    break;
  case PID_semiColon:
    break;
  default: return "Unknown 'pit' Parameter";
  };
  return NULL;
}


static char *pEndCell(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    if (pPkt->pData->GetCellType() == roomILLEGAL)
    {
      return "Missing room-type declaration";
    };
    break;
  case PID_semiColon:
    pPkt->iFile->PopCommand(); //Pop cellCmds
    break;
  default: return "Unknown 'EndCell' parameter";
  };
  return NULL;
}

static char *BuildImportedArea(PPKT *pPkt)
{
  //First we modify all cell locations to be level-relative
  // just like LOCATIONREL locations;
  // This will set all EXPORTLOC.left and .top to zero.
  TCELL *pCell;
  LOCATIONREL locr;
  int minX, maxX, minY, maxY;
  minX = 32;
  minY = 32;
  maxX = -1;
  maxY = -1;
  for (pCell=pPkt->pData->firstCell;
       pCell!=NULL;
       pCell=pCell->nextCell)
  {

    if (pCell->address.level != -1)
    { // Level is relative
      pCell->address.dl += pPkt->pData->level;
      pCell->address.level = 0;
    };
    if (pCell->address.left != -1)
    {
      pCell->address.dx += pPkt->pData->left;
      pCell->address.left = 0;
    };
    if (pCell->address.top != -1)
    {
      pCell->address.dy += pPkt->pData->top;
      pCell->address.top = 0;
    };
    if (pCell->address.dx < 0) return "Cell lies west of level";
    if (pCell->address.dx > pDoc->LastColumn(pPkt->pData->level))
                  return "Cell lies east of level";
    if (pCell->address.dy < 0) return "Cell lies north of level";
    if (pCell->address.dy > pDoc->LastRow(pPkt->pData->level))
                  return "Cell lies south of level";
    //if (pCell->address.dx > maxX) maxX = pCell->address.dx;
    //if (pCell->address.dx < minX) minX = pCell->address.dx;
    //if (pCell->address.dy > maxY) maxY = pCell->address.dy;
    //if (pCell->address.dy < minY) minY = pCell->address.dy;
    if (  (RoomType(pCell->address) != roomOPEN) && (RoomType(pCell->address) != roomSTONE))
                return "Current celltype is not 'OPEN' or 'STONE'";
    if (pDoc->FindFirstObject(
                        pCell->address.dl,
                        pCell->address.dx,
                        pCell->address.dy) != RNeof)
                return "Current cell is not empty";
  };
  // So now we know:
  // All the cells will fit. 
  // All current cell types for imported area is roomOPEN or roomSTONE.
  // All current cells are empty.
  //
  // Now we go through the TCELL list and set each
  // cell's cellflag and contents.
  //
  for (pCell=pPkt->pData->firstCell;
       pCell!=NULL;
       pCell=pCell->nextCell)
  {
    locr.l(pCell->address.dl);
    locr.x(pCell->address.dx);
    locr.y(pCell->address.dy);
    pDoc->SetCellFlag(locr, (CELLFLAG)pCell->cf);
  };
  if (pPkt->pData->difficultyMultiplier != -1)
  {
    pDoc->ExperienceMultiplier(pPkt->pData->level, pPkt->pData->difficultyMultiplier);
  };
  return NULL;
}

static char *pEndArea(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    break;
  case PID_semiColon:
    pPkt->iFile->PopCommand(); //Pop the area commands..done
    //Everthing went well.  Now we can build the imported area.
    return BuildImportedArea(pPkt);
  default: return "Unknown 'EndCell' parameter";
  };
  return NULL;
}



static char *pRecTeleporter(PPKT *pPkt)
{
  switch(pPkt->pid)
  {
  case PID_Initialize:
    {
      RN  RNteleporter;
      DB1 *pTeleporter;
      RNteleporter = pDoc->FindEmptyDB1Entry();
      pTeleporter = pDoc->GetRecordAddressDB1(RNteleporter);
      pTeleporter->Clear();
      pPkt->pData->AddObject(RNteleporter);
      break;
    };
  case PID_north: 
    pPkt->pData->Teleporter()->facingMode(true);
    pPkt->pData->Teleporter()->rotation(0);
    break;
  case PID_east: 
    pPkt->pData->Teleporter()->facingMode(true);
    pPkt->pData->Teleporter()->rotation(1);
    break;
  case PID_south: 
    pPkt->pData->Teleporter()->facingMode(true);
    pPkt->pData->Teleporter()->rotation(2);
    break;
  case PID_west: 
    pPkt->pData->Teleporter()->facingMode(true);
    pPkt->pData->Teleporter()->rotation(3);
    break;
  case PID_rot90: 
    pPkt->pData->Teleporter()->facingMode(false);
    pPkt->pData->Teleporter()->rotation(1);
    break;
  case PID_rot180: 
    pPkt->pData->Teleporter()->facingMode(false);
    pPkt->pData->Teleporter()->rotation(2);
    break;
  case PID_rot270: 
    pPkt->pData->Teleporter()->facingMode(false);
    pPkt->pData->Teleporter()->rotation(3);
    break;
  case PID_objects:
    pPkt->pData->Teleporter()->what(0);
    break;
  case PID_monsters:
    pPkt->pData->Teleporter()->what(1);
    break;
  case PID_partyAndObjects:
    pPkt->pData->Teleporter()->what(2);
    break;
  case PID_anything:
    pPkt->pData->Teleporter()->what(3);
    break;
  case PID_target:
    pPkt->pData->Teleporter()->dest(pPkt->eloc);
    break;
  case PID_semiColon:
 //   pPkt->iFile->PopCommand();
    break;
  default: return "TeleporterRec parameter not implemented";
  };
  return NULL;
}

static char *pRecMonster(PPKT *pPkt)
{
  switch(pPkt->pid)
  {
  case PID_Initialize:
    {
      RN  RNmonster;
      DB4 *pMonster;
      RNmonster = pDoc->FindEmptyDB4Entry();
      pMonster = pDoc->GetRecordAddressDB4(RNmonster);
      pMonster->Clear();
      pPkt->pData->AddObject(RNmonster);
      break;
    };
  case PID_type:
    NotImplemented();
    break;
  case PID_num:
    pPkt->pData->Monster()->numMonster(pPkt->uint);
    break;
  case PID_north: 
    pPkt->pData->Monster()->facing(0);
    break;
  case PID_east: 
    pPkt->pData->Monster()->facing(1);
    break;
  case PID_south: 
    pPkt->pData->Monster()->facing(2);
    break;
  case PID_west: 
    pPkt->pData->Monster()->facing(3);
    break;
  case PID_hp:
    NotImplemented();
    break;
  case PID_position:
    NotImplemented();
    break;
  case PID_semiColon:
    break;
  default: return "MonsterRec parameter not implemented";
  };
  return NULL;
}

COMMANDS cellCmds[] =
{
  {"open"         ,openDef         ,pOpen},
  {"stone"        ,stoneDef        ,pStone},
  {"stairs"       ,stairsCFDef     ,pCFStairs},
  {"pit"          ,pitCFDef        ,pCFPit},
  {"teleporter"   ,teleporterCFDef ,pCFTeleporter},
  {"teleporterrec",teleporterRecDef,pRecTeleporter},
  {"monsterrec"   ,monsterRecDef   ,pRecMonster},
  {"endcell"      ,endcellDef      ,pEndCell},
  {""}
};

static char *pCell(PPKT *pPkt)
{
  switch (pPkt->pid)
  {
  case PID_Initialize:
    pPkt->pData->NewCell();
    break;
  case PID_address:
    pPkt->pData->firstCell->address = pPkt->eloc;
    break;
  case PID_semiColon:
    pPkt->iFile->PushCommand(cellCmds);
    pPkt->pData->SetCellType(roomILLEGAL);
    break;
  default: return "Cell parameter not implemented";
  };
  return NULL;
}


COMMANDS areaCommands[] =
{
    {"area"   ,areaDef   ,pArea},
    {"cell"   ,cellDef   ,pCell},
    {"endarea",endAreaDef,pEndArea},
    {""}
};


void ViewLevelInfo::ImportLevel(FILE *f) 
{
  //EXPORTLOC eloc;
  GetImportLocation dlg;
  DATA data;
  IMPORTFILE ifile(f, &data);
  if (dlg.DoModal() != IDOK) return;
  data.left = dlg.m_importX;
  data.top = dlg.m_importY;
  data.level = m_currentLevel;
  if (ifile.Parse(areaCommands))
  {
    MessageBox("Import Complete", "Import", MB_OK);
  }
  else
  {
    MessageBox("Import Not Done", "Error", MB_OK);
  };
}