// LocateText.cpp : implementation file
//

#include "stdafx.h"
#include "CSBuild.h"
#include "LocateText.h"
#include "CSBuildDoc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

static int currentFloorText=7;  //floor, wall, or scroll
static int currentMatchExact=1; //Any word
static CString currentSearchWords;

WordList::WordList(void)
{
  m_wordList = NULL;
  m_numWord = 0;
}

WordList::~WordList(void)
{
  Clear();
}

void WordList::Clear(void)
{
  int i;
  for (i=0; i<m_numWord; i++)
  {
    delete m_wordList[i];
    m_wordList[i] = NULL;
  };
  m_numWord = 0;
  if (m_wordList != NULL) free(m_wordList);
  m_wordList = NULL;
}

void WordList::AddWord(const CString& word)
{
  m_numWord++;
  m_wordList = (CString**)realloc(m_wordList,
                                  m_numWord*sizeof(m_wordList[0]),
                                  MALLOC075);
  m_wordList[m_numWord-1] = new CString(word);
}

/////////////////////////////////////////////////////////////////////////////
// LocateText dialog


LocateText::LocateText(CWnd* pParent /*=NULL*/)
	: CDialog(LocateText::IDD, pParent)
{
	//{{AFX_DATA_INIT(LocateText)
	m_floorText = FALSE;
	m_wallText = FALSE;
	m_scrolls = FALSE;
	m_searchWords = _T("");
	m_matchExact = -1;
	//}}AFX_DATA_INIT
}


void LocateText::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(LocateText)
	DDX_Control(pDX, IDC_Result, m_resultCtrl);
	DDX_Check(pDX, IDC_FloorText, m_floorText);
	DDX_Check(pDX, IDC_WallText, m_wallText);
	DDX_Check(pDX, IDC_Scrolls, m_scrolls);
	DDX_Text(pDX, IDC_SearchWords, m_searchWords);
	DDX_Radio(pDX, IDC_MatchExact, m_matchExact);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(LocateText, CDialog)
	//{{AFX_MSG_MAP(LocateText)
	ON_BN_CLICKED(IDC_Find, OnFind)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// LocateText message handlers

void LocateText::RememberSettings(void)
{
  UpdateData(true);
  currentSearchWords = m_searchWords;
  currentFloorText = 0;
  if (m_floorText) currentFloorText |= 1;
  if (m_wallText)  currentFloorText |= 2;
  if (m_scrolls)   currentFloorText |= 4;
  currentMatchExact = m_matchExact;
}

void LocateText::OnOK() 
{
	// TODO: Add extra validation here
  RememberSettings();
	CDialog::OnOK();
}



BOOL LocateText::OnInitDialog() 
{
	CDialog::OnInitDialog();
  UpdateData(true);
  m_matchExact = currentMatchExact;
  m_floorText  = (currentFloorText & 1) != 0;
  m_wallText   = (currentFloorText & 2) != 0;
  m_scrolls    = (currentFloorText & 4) != 0;
  m_searchWords = currentSearchWords;
  UpdateData(false);
	// TODO: Add extra initialization here
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void LocateText::MatchText(RN obj, 
                           const LOCATIONREL& locr,
                           CString& result)
{
  CString raw;
  int i;
  raw = pDoc->GetRawText(obj, TEXTTYPE_SCROLL);
  switch (m_matchExact)
  {
  case 0: //Exact Match
    if (raw.Find(m_searchWords,0) >= 0)
    {
      result += locr.Format();
      result += " ";
      result += raw;
      result += "\015\012";
    };
    break;
  case 1: //Words in any order
    for (i=0; i<m_wordList.NumWord(); i++)
    {
      if (raw.Find(m_wordList.GetWord(i)) < 0)
      {
        break;
      };
    };
    if (i == m_wordList.NumWord())
    {
      result += locr.Format();
      result += " ";
      result += raw;
      result += "\015\012";
    };
    break;
  case 2: //Any word
    for (i=0; i<m_wordList.NumWord(); i++)
    {
      if (raw.Find(m_wordList.GetWord(i)) >= 0)
      {
        result += locr.Format();
        result += " ";
        result += raw;
        result += "\015\012";
        break;
      };
    };
    break;
  };
}

void LocateText::SearchList(RN obj, 
                            const LOCATIONREL& locr,
                            CString& result)
{
  DBTYPE dbType;
  ROOMTYPE rt;
  DB9 *pChest;
  DB7 *pScroll;
  DB4 *pMonster;
  for (;obj != RNeof; obj = pDoc->GetDBRecordLink(obj))
  {
    dbType = obj.dbType();
    switch (dbType)
    {
    case dbDOOR:
    case dbTELEPORTER:
    case dbACTUATOR:
    case dbWEAPON:
    case dbCLOTHING:
    case dbPOTION:
    case dbMISC:
    case dbMISSILE:
    case dbCLOUD:
      break;
    case dbCHEST:
      pChest = pDoc->GetRecordAddressDB9(obj);
      SearchList(pChest->contents(),locr,result);
      break;
    case dbSCROLL:
      pScroll = pDoc->GetRecordAddressDB7(obj);
      if (m_scrolls)
      {
        MatchText(pScroll->text(), locr, result);
      };
      break;
    case dbTEXT:
      rt = RoomType(locr);
      if ((rt == roomSTONE) && m_wallText)
      {
        MatchText(obj, locr, result);
      };
      if ((rt != roomSTONE) && m_floorText)
      {
        MatchText(obj, locr, result);
      };
      break;
    case dbMONSTER:
      pMonster = pDoc->GetRecordAddressDB4(obj);
      SearchList(pMonster->possession(), locr, result);
      break;
    };
  };
}

void LocateText::SplitWords(void)
{
  int col;
  CString word;
  m_wordList.Clear();
  col = 0;
  while (col < m_searchWords.GetLength())
  {
    if (m_searchWords[col] == ' ')
    {
      col++;
      continue;
    };
    word = m_searchWords[col];
    col++;
    while (col < m_searchWords.GetLength())
    {
      if (m_searchWords[col] == ' ') break;
      word += m_searchWords[col];
      col++;
    };
    m_wordList.AddWord(word);
  };
}

void LocateText::SearchAndDisplay(void)
{
  CString result;
  LOCATIONREL locr;
  SplitWords(); //Build m_searchWordList.
  for (locr.First(); locr.l()>=0; locr.Next())
  {
    SearchList(pDoc->FindFirstObject(locr), locr, result);
  };
  if (result == "") result = "Nothing Found";
  m_resultCtrl.SetWindowText(result);
}

void LocateText::OnFind() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
  if ((!m_floorText) && (!m_wallText) && (!m_scrolls))
  {
    pDoc->Sorry("You must check at least one of the boxes labeled\n"
          "'Floor Text', 'Wall Text', or 'Scrolls'");
    return;
  };
  SearchAndDisplay();
	
}

void LocateText::OnCancel() 
{
	// TODO: Add extra cleanup here
	RememberSettings();
	CDialog::OnCancel();
}
