#include "stdafx.h"
//#include "types.h"
#include "misc.h"
#include "CSBuildDoc.h"
#include "CSBuildView.h"

bool CCSBuildView::IsWall(LOCATIONREL locr)
{
  return RoomType(pDoc->GetCellFlag(locr)) == roomSTONE;
}

int CCSBuildView::CountAttachedWalls(LOCATIONREL locr)
{
  int result = 0;
  if (locr.x() == 0) result++;
  else {locr.x(locr.x()-1);if (IsWall(locr)) result++;locr.x(locr.x()+1);};
  if (locr.y() == 0) result++;
  else {locr.y(locr.y()-1);if (IsWall(locr)) result++;locr.y(locr.y()+1);};
  if (locr.x() == pDoc->LastColumn(locr.l())) result++;
  else {locr.x(locr.x()+1);if (IsWall(locr)) result++;locr.x(locr.x()-1);};
  if (locr.y() == pDoc->LastRow(locr.l())) result++;
  else {locr.y(locr.y()+1);if (IsWall(locr)) result++;locr.y(locr.y()-1);};
  return result;
}

int CCSBuildView::FindStairExit(LOCATIONREL locr)
{
  if (locr.x() != 0)
  {
    locr.x(locr.x()-1);if (!IsWall(locr)) return 3;locr.x(locr.x()+1);
  };
  if (locr.y() != 0)
  {
    locr.y(locr.y()-1);if (!IsWall(locr)) return 0;locr.y(locr.y()+1);
  };
  if (locr.x() != pDoc->LastColumn(locr.l()))
  {
    locr.x(locr.x()+1);if (!IsWall(locr)) return 1;locr.x(locr.x()-1);
  };
  if (locr.y() != pDoc->LastRow(locr.l()))
  {
    locr.y(locr.y()+1);if (!IsWall(locr)) return 2;locr.y(locr.y()-1);
  };
  return -1;
}

bool CCSBuildView::AnyAttachedDoors(LOCATIONREL locr, bool msg)
{
  int height, width, dir, col, row;
  CELLFLAG cf;
  ROOMTYPE rt;
  height = pDoc->LastRow(pDoc->m_currentDisplayLevel) + 1;
  width = pDoc->LastColumn(pDoc->m_currentDisplayLevel) + 1;
  for (col=locr.x()-1; col<=locr.x()+1; col++) 
  for (row=locr.y()-1; row<=locr.y()+1; row++)
  {
    if ( (col!=locr.x()) && (row!=locr.y()) ) continue;
    if ( (col==locr.x()) && (row==locr.y()) ) continue;
    if (row<0) continue;
    if (col<0) continue;
    if (col>=width) continue;
    if (row>=height) continue;
    cf = pDoc->GetCellFlag(pDoc->m_currentDisplayLevel, col, row);
    rt = RoomType(cf);
    if (rt != roomDOOR) continue;
    if (col == locr.x()) dir = 0; else dir = 8;
    if (((cf ^ dir) & 8) == 0) 
    {
      if (msg)
      {
        MessageBox("That cell is attached to a door","Error");
      };
      return true;
    };
  };
  return false;
}

bool CCSBuildView::StairwayWouldBeMadeInvalid(LOCATIONREL locr)
{ //See if changing *locr to an open room would
  //leave a stairway dangling in open space.
  locr.x(locr.x()+1);
  if (RoomType(locr) == roomSTAIRS) 
  {
    if(CountAttachedWalls(locr) < 4) return true;
  };
  locr.x(locr.x()-2);
  if (RoomType(locr) == roomSTAIRS)
  {
    if (CountAttachedWalls(locr) < 4) return true;
  };
  locr.x(locr.x()+1);
  locr.y(locr.y()-1);
  if (RoomType(locr) == roomSTAIRS)
  {
    if (CountAttachedWalls(locr) < 4) return true;
  };
  locr.y(locr.y()+2);
  if (RoomType(locr) == roomSTAIRS)
  {
    if (CountAttachedWalls(locr) < 4) return true;
  };
  return false;
}

bool CCSBuildView::ActuatorWouldChange(
                    LOCATIONREL locr,
                    ROOMTYPE oldRT,
                    ROOMTYPE newRT,
                    bool printError)
{
  RN RNac;
  DB3 *pActuator;
  ACTUATORCLASS ac1, ac2;
  for (RNac = pDoc->FindFirstActuator(locr);
       RNac != RNeof;
       RNac = pDoc->GetDBRecordLink(RNac))
  {
    pActuator = pDoc->GetRecordAddressDB3(RNac);
    ac1 = pActuator->actuatorClass(oldRT);
    ac2 = pActuator->actuatorClass(newRT);
    if (ac1 == ac2) continue;
    if (!printError) return true;
    {
      CString name1;
      CString name2;
      CString msg;
      bool result;
      name1 = pActuator->Name(oldRT);
      name2 = pActuator->Name(newRT);
      msg.Format("That would change an actuator located\n"
                 "at %d(%02d,%02d)%d\n"
                 "from a %s\n"
                 "to a %s",
                     locr.l(), locr.x(), locr.y(), RNac.pos(),
                     name1,
                     name2);
      result = pDoc->DoYouReallyWantTo(msg);
      return !result;
    };
  };
  return false;
}

bool CCSBuildView::PlaceOpenRoom(LOCATIONREL locr, 
                                 bool drag,
                                 CELL_CHECKPOINT **ppCCBefore)
{
  CELLFLAG cf;
  RN obj;
  bool result = false;
  if (AnyAttachedDoors(locr, true)) return false;
  cf = pDoc->GetCellFlag(locr);
  switch (RoomType(cf)) //old roomTYPE
  {
  case roomOPEN: //Old roomTYPE
    if (drag)
    {
      result = true;
    };
    break; // nothing to do
  case roomSTONE:
    if (ActuatorWouldChange(locr, roomSTONE, roomOPEN, true))
    {
      return false;
    };
    if (StairwayWouldBeMadeInvalid(locr))
    {
      Error("Opening that cell would\n"
            "make a stairway invalid");
      result = false;
      break;
    };
    pDoc->SetRoomType(locr,roomOPEN);
    Invalidate(false);
    result = true;
    break;
  case roomSTAIRS: //old roomTYPE
    if (drag) break;
    {
      LOCATIONREL newlocr;
      newlocr = locr;
      if (cf & 4) newlocr.Vertical(-1);  //Up
      else newlocr.Vertical(1);          //Down
      pDoc->SetRoomType(locr, roomOPEN);
      result = true;
      Invalidate();
      if (newlocr.l() < 0) 
      {
        MessageBox("Strange! These stairs go nowhere.");
        break;
      };
      if (RoomType(pDoc->GetCellFlag(newlocr)) == roomSTAIRS)
      {
        *ppCCBefore = pDoc->CheckpointCell(newlocr);
        pDoc->SetRoomType(newlocr, roomOPEN);
        result = true;
      }
      else
      {
        MessageBox("Strange!\nThere are no stairs on the other floor");
      };
    };
    break;
  case roomDOOR: //old roomTYPE
    if (drag) break;
    pDoc->SetRoomType(locr, roomOPEN);
    result = true;
    Invalidate();
    break;
  case roomFALSEWALL: //old roomTYPE
  case roomPIT: //old roomTYPE
    pDoc->SetRoomType(locr, roomOPEN);
    Invalidate();
    result = true;
    break;
  case roomTELEPORTER: //old roomTYPE
    if (drag) break;
    pDoc->SetRoomType(locr, roomOPEN);
    result = true;
    Invalidate();
    break;
  default:
    MessageBox("You cannot put an open space there","error");
  }; //switch old roomTYPE
  return result;
}

CString CCSBuildView::IsPitOK(LOCATIONREL locr)
{
  LOCATIONABS belowA;
  LOCATIONREL belowR;
  ROOMTYPE rtBelow;
  if (locr.l() >= pDoc->NumLevel())
  {
    return CString("You are on the bottom floor!");
  };
  belowA = locr; //absolute location
  belowA.l++;  // To the floor below.
  belowR = belowA;
  if (    (belowR.x() > pDoc->LastColumn(belowR.l()))
       || (belowR.y() > pDoc->LastRow(belowR.l())))
  {
    return CString("That pit would go nowhere");
  };
  rtBelow = RoomType(pDoc->GetCellFlag(belowR));
  if (rtBelow == roomSTAIRS)
  {
    return CString("That would fall into a stairwell");
  };
  if (rtBelow == roomSTONE)
  {
    return CString("That would fall into a wall");
  };
  if (rtBelow == roomDOOR)
  {
    return CString("That would fall into a door");
  };
  if (rtBelow == roomFALSEWALL)
  {
    return CString("That would fall into a falsewall");
  };
  return CString("");
}

void CCSBuildView::SetPitOpen(LOCATIONREL locr)
{
  pDoc->SetCellFlag(locr, (ui8)(pDoc->GetCellFlag(locr) | 0x8));
}

void CCSBuildView::SetPitReal(LOCATIONREL locr)
{
  pDoc->SetCellFlag(locr, (ui8)(pDoc->GetCellFlag(locr) & ~0x1));
}

void CCSBuildView::SetPitEasyToSee(LOCATIONREL locr)
{
  pDoc->SetCellFlag(locr, (ui8)(pDoc->GetCellFlag(locr) & ~0x4));
}

bool CCSBuildView::PlacePitRoom(LOCATIONREL locr, 
                                bool drag,
                                CELL_CHECKPOINT **ppCCBefore)
{
  CELLFLAG cf;
  LOCATIONABS belowA;
  LOCATIONREL belowR;
  CString msg;
  cf = pDoc->GetCellFlag(locr);
  switch (RoomType(cf)) //Old roomTYPE
  {
  case roomPIT: //old roomTYPE
    if (!drag) return false;
    return true; //Nothing to do.
  case roomSTONE: //old roomTYPE
    // Check that there is a floor below, etc.
    if (!(pDoc->EditingOptions() & EDIT_OPT_AllowIllegalPits))
    {
      msg = IsPitOK(locr);
    };
    if (msg != "")
    {
      pDoc->Sorry(msg);
      return false;
    };
    if (!PlaceOpenRoom(locr, drag, ppCCBefore)) return false;
    pDoc->SetRoomType(locr, roomPIT);
    SetPitOpen(locr);
    SetPitReal(locr);
    SetPitEasyToSee(locr);
    Invalidate();
    return true;
  case roomOPEN: //old roomTYPE
    // Check that there is a floor below.
    if (!(pDoc->EditingOptions() & EDIT_OPT_AllowIllegalPits))
    {
      msg = IsPitOK(locr);
    };
    if (msg != "")
    {
      pDoc->Sorry(msg);
      return false;
    };
    pDoc->SetRoomType(locr, roomPIT);
    SetPitOpen(locr);
    SetPitReal(locr);
    SetPitEasyToSee(locr);
    Invalidate();
    return true;
  case roomSTAIRS: //old roomTYPE
  case roomTELEPORTER: //ol roomTYPE
  case roomFALSEWALL: //old roomTYPE
  case roomDOOR: //old roomTYPE
    // Check that there is a floor below.
    if (!(pDoc->EditingOptions() & EDIT_OPT_AllowIllegalPits))
    {
      msg = IsPitOK(locr);
    };
    if (msg != "")
    {
      pDoc->Sorry(msg);
      return false;
    };
    if (!PlaceOpenRoom(locr, drag, ppCCBefore)) return false;
    pDoc->SetRoomType(locr, roomPIT);
    SetPitOpen(locr);
    Invalidate();
    return true;
  };
  return false;
}

void CCSBuildView::PlaceTeleporterRoom(LOCATIONREL locr,
                                       CELL_CHECKPOINT **ppCCBefore)
{
  CELLFLAG cf;
  cf = pDoc->GetCellFlag(locr);
  switch (RoomType(cf)) //Old roomTYPE
  {
  case roomSTONE: //old roomTYPE
    if (!PlaceOpenRoom(locr, false, ppCCBefore)) return;
    //Note Fallthrough!!!!
  case roomOPEN: //old roomTYPE
    pDoc->SetRoomType(locr, roomTELEPORTER);
    pDoc->SetTeleporterActive(locr, true);
    Invalidate();
    break;
  case roomTELEPORTER: //old roomTYPE
  case roomSTAIRS: //old roomTYPE
  case roomPIT: //old roomTYPE
  case roomFALSEWALL: //old roomTYPE
  case roomDOOR: //old roomTYPE
    pDoc->Sorry("You can only place a teleporter\nin an open cell");
    break;
  };
}

bool CCSBuildView::PlaceStoneRoom(LOCATIONREL locr, bool drag)
{
  CELLFLAG cf;
  SEARCHPKT srchpkt;
  LOCATIONABS tloca;
  RN objTELEPORT;
  bool result = false;
  cf = pDoc->GetCellFlag(locr);
  //if (AnyAttachedDoors(locr, true)) return;
  switch (RoomType(cf)) //Old roomTYPE
  {
  case roomSTONE: //old roomTYPE
    if (drag) result = true;
    break; //Nothing to do.
  case roomOPEN: //old roomTYPE
    {
      LOCATIONABS loca, upLoca;
      LOCATIONREL upLocr;
      ROOMTYPE uprt;
      loca = locr;
      upLoca = loca;
      upLoca.l--;
      upLocr = upLoca;
      uprt = RoomType(pDoc->GetCellFlag(upLocr));
      if (uprt == roomPIT)
      {
        pDoc->Sorry("There is a pit above that cell");
        return false;
      };
      if (pDoc->FindFirstActuator(locr) != RNeof)
      {
        MessageBox("There is an actuator at that location","error");
        result = false;
        break;
      };
      tloca.l = -1;
      objTELEPORT = pDoc->FindNextTeleporter(locr, &tloca);
      if (objTELEPORT != RNeof)
      {
        char msg[80];
        sprintf(msg,"This square is the destination of \n"
                    "a teleporter at\n"
                    "%d(%02d,%02d)",
                    tloca.l, 
                    tloca.x, 
                    tloca.y);
        MessageBox(msg, "Error");
        return false;
      };
    };
    pDoc->SetRoomType(locr,roomSTONE);
    Invalidate(false);
    result = true;
    break;
  case roomSTAIRS: //old roomTYPE
  case roomPIT: //old roomTYPE
  case roomTELEPORTER: //ol roomTYPE
  case roomFALSEWALL: //old roomTYPE
  case roomDOOR: //old roomTYPE
    pDoc->Sorry("You can only put a wall on an empty cell");
  }; //switch old roomTYPE
  return result;
}

bool CCSBuildView::IsStairsOK(LOCATIONREL locr, int direction)
{
  CELLFLAG cf /*, farcf*/;
  LOCATIONREL farlocr;
  bool illegalStairsOK;
  farlocr = locr;
  farlocr.Vertical(direction);
  cf = pDoc->GetCellFlag(locr);
  illegalStairsOK = (pDoc->EditingOptions() & EDIT_OPT_AllowIllegalStairs)!=0;
  switch (RoomType(cf)) //Old roomTYPE
  {
  case roomOPEN: //old roomTYPE
    if (!illegalStairsOK && (CountAttachedWalls(locr) != 3))
    {
      MessageBox("This end of the stairway\ndoes not have three walls");
      return false;
    };
    break;
  case roomSTAIRS:
  case roomSTONE: //old roomTYPE
  case roomPIT: //old roomTYPE
  case roomTELEPORTER: //ol roomTYPE
  case roomFALSEWALL: //old roomTYPE
  case roomDOOR: //old roomTYPE
  default:
    MessageBox("You can only put stairs in an OPEN cell","Error");
    break;
  };
  return true;
}

void CCSBuildView::PlaceStairsRoom(
                         LOCATIONREL locr, 
                         int direction,
                         CELL_CHECKPOINT **pCCsecondBefore)
{
  CELLFLAG cf;
  bool illegalStairsOK = false;
  LOCATIONABS locabs;
  LOCATIONREL newlocr;
  SEARCHPKT srchpkt;
  char msg[80];
  //xxxx
  cf = pDoc->GetCellFlag(locr);
  newlocr = locr;
  if (pDoc->EditingOptions() & EDIT_OPT_AllowIllegalStairs)
                        illegalStairsOK = true;
  newlocr.Vertical(direction); //-1 if too big or too small.
  locabs.l = -1;
  pDoc->FindNextTeleporter(locr,&locabs);
  if (!illegalStairsOK && (locabs.l >= 0))
  {
    sprintf(msg, "This is the destination of\n"
            "a teleporter at\n"
            "%d(%02d,%02d)",
            locabs.l, 
            locabs.x, 
            locabs.y);
    MessageBox(msg);
    return;
  };
  locabs.l = -1;
  if (newlocr.l() >= 0)
  {
    Invalidate(false);
    pDoc->FindNextTeleporter(newlocr, &locabs);
    if (!illegalStairsOK && (locabs.l >= 0))
    {
      sprintf(msg, "The other end of this stairway"
                  "is the destination of\n"
                  "a teleporter at\n"
                  "%d(%02d,%02d)",
                  locabs.l, 
                  locabs.x, 
                  locabs.y);
      MessageBox(msg);
      return;
    };
  };
  if (newlocr.l() < 0)
  {
    if (!illegalStairsOK)
    {
      MessageBox("Those stairs would go nowhere!");
      return;
    };
  };
  //We are perhaps changing two cells.
  //Make a checkpoint of the other one
  pDoc->SetRoomType(locr, roomSTAIRS);
  pDoc->SetStairsUp(locr, direction == -1);
  pDoc->SetStairsDirection(locr,FindStairExit(locr));
  Invalidate();
  if (newlocr.l() >= 0)
  {
    *pCCsecondBefore = pDoc->CheckpointCell(newlocr);
    pDoc->SetRoomType(newlocr, roomSTAIRS);
    pDoc->SetStairsUp(newlocr, direction == 1);
    pDoc->SetStairsDirection(newlocr,FindStairExit(newlocr));
  };
}

bool CCSBuildView::PlaceFalseWallRoom(LOCATIONREL locr,
                                      bool drag,
                                      CELL_CHECKPOINT **ppCCBefore)
{
  CELLFLAG cf;
  bool result;
  cf = pDoc->GetCellFlag(locr);
  switch (RoomType(cf)) //Old roomTYPE
  {
  case roomSTONE: //old roomTYPE
    if (!PlaceOpenRoom(locr, drag, ppCCBefore)) return false;
    // Note!! Fallthrough
  case roomOPEN: //old roomTYPE
    pDoc->SetRoomType(locr, roomFALSEWALL);
    pDoc->SetFalseWallOpen(locr, false);
    Invalidate(false);
    result = true;
    break;
  case roomFALSEWALL: //old roomTYPE
    if (drag)
    {
      result = true;
      break;
    };
    // Note fallthrough!
  case roomSTAIRS: //old roomTYPE
  case roomPIT: //old roomTYPE
  case roomTELEPORTER: //ol roomTYPE
  case roomDOOR: //old roomTYPE
  default:
    MessageBox("You and only put a false wall\nin an empty cell","Error");
    result = false;
  };
  return result;
}

bool CCSBuildView::IsDoorOK(LOCATIONREL locr, int type)
{
  int dx, dy;
  LOCATIONREL locAdj;
  ROOMTYPE rt1, rt2;
  if (type & 1) //North/south door.
  {
    dx = 1;
    dy = 0;
  }
  else //East/west door
  {
    dx = 0;
    dy = 1;
  };
  //Check that there are two walls adjacent.
  locAdj = locr;
  locAdj.x(locAdj.x()+dx);
  locAdj.y(locAdj.y()+dy);
  rt1 = RoomType(pDoc->GetCellFlag(locAdj));
  locAdj = locr;
  locAdj.x(locAdj.x()-dx);
  locAdj.y(locAdj.y()-dy);
  rt2 = RoomType(pDoc->GetCellFlag(locAdj));
  if ( (rt1!=roomSTONE) || (rt2!=roomSTONE) )
  {
    MessageBox("A door must be attached to two walls");
    return false;
  };
  return true;
};

void CCSBuildView::PlaceDoorRoom(LOCATIONREL locr, 
                                 int type,
                                 CELL_CHECKPOINT **ppCCBefore)
{
  CELLFLAG cf;
  LOCATIONREL locAdj;
  if (!IsDoorOK(locr, type)) return;
  cf = pDoc->GetCellFlag(locr);
  switch (RoomType(cf)) //Old roomTYPE
  {
  case roomSTONE: //old roomTYPE
    if (!PlaceOpenRoom(locr, false,ppCCBefore)) return;
    //Note Fallthrough
  case roomOPEN: //old roomTYPE
    pDoc->SetRoomType(locr, roomDOOR);
    pDoc->SetDoorState(locr, type<2 ? 4 : 0);
    pDoc->SetDoorNS(locr, (type & 1) != 0);
    Invalidate(false);
    break;
  case roomSTAIRS: //old roomTYPE
  case roomPIT: //old roomTYPE
  case roomTELEPORTER: //ol roomTYPE
  case roomFALSEWALL: //old roomTYPE
  case roomDOOR: //old roomTYPE
  default:
    MessageBox("You and only put a door\nin an empty cell","Error");
  };
}
