#include "stdafx.h"
//#include "types.h"
#include "CSBuildDoc.h"
#include "Misc.h"

void RC4_prepare_key(const unsigned char *key_data_ptr,
                     int key_data_len);
void RC4_encipher(unsigned char *buffer_ptr, 
                  int buffer_len,
                  int position);


unsigned int FormChecksum(ui8 *buf, int num);
extern char ExtendedFeaturesVersion;
extern CString encipheredDataFile;
extern bool RecordDesignOption;
extern bool DMRulesOption;
extern bool invisibleMonsters;
extern bool drawAsSize4Monsters;
extern char *gameInfo;
extern ui32 expectedGraphicsSignature1;
extern ui32 expectedGraphicsSignature2;
extern ui32 expectedCSBgraphicsSignature1;
extern ui32 expectedCSBgraphicsSignature2;
extern char hintKey[8];

struct MEMBLOCKS
{
  char *address;
  int  size;
};

MEMBLOCKS memBlocks[12];

void BinaryWrite(CArchive& ar, void *buf, int len)
{
  int pos=0;
  if (encipheredDataFile != "")
  {
    CFile *file;
    ar.Flush();
    file = ar.GetFile();
    pos = file->GetPosition();
    RC4_encipher((unsigned char *)buf, len, pos);
  };
  ar.Write(buf, len);
  if (encipheredDataFile != "")
  { //Undo the damage.
    RC4_encipher((unsigned char *)buf, len, pos);
  };
}

i16 CCSBuildDoc::GenChecksum(i16 *buf, int initHash, i32 numWord)
{
  int result;
  result = initHash;
  while (numWord != 0)
  {
    result +=  LE16(*buf);
    result +=  LE16(*buf) ^ initHash;
    initHash += numWord;
    buf++;
    numWord--;
  };
  return (i16)result;
}

//*********************************************************
// Scramble and write the first 512-byte block of gamesave file.
//*********************************************************
STORESTATUS CCSBuildDoc::ScrambleAndWrite(CArchive& ar, ui16 *pwbuf)
{
  i32 i;
  int hash;
  int sumPart2;
  STORESTATUS status;
  sumPart2 = 0;
  //pwA3 = pwbuf + 128;
  for (i=0; i<128; i++) 
  {
    sumPart2 += LE16(pwbuf[i+128]);
  };
  // sumPart2 is the sum of the second half of the buffer.

  hash = 0;
  //pwA3 = pwbuf;
  for (i=0; i<128; i+=4)
  { 
    //pwBuf[i] = 0;
    hash += LE16(pwbuf[i]);
    //*pwA3 =  0;
    hash ^= LE16(pwbuf[i+1]);
    //*pwA3 =  0;
    hash -= LE16(pwbuf[i+2]);
    if (i != 124)
    {
      //*pwA3 =  0;
      hash ^= LE16(pwbuf[i+3]);
    };
  };
  pwbuf[127] = LE16(sumPart2 ^ hash);
  //Scramble the 2nd 256 bytes.
  Unscramble(pwbuf+128, LE16(pwbuf[29]), 128); // 128 words = 256 bytes
  status = WriteDataFile(ar, (ui8 *)pwbuf, 512);
  //Then unscramble the 2nd 256 bytes again for use.
  Unscramble(pwbuf+128, LE16(pwbuf[29]), 128);
  return status;
}

void shift(unsigned char **out, 
           int *bitsAvail,
           int num, 
           unsigned char value)
{
  int result, shift;
  result = ((**out) << 8) | (value &0xff);
  while (num > 0)
  {
    if (*bitsAvail == 0)
    {
      **out = (unsigned char)(result >> 8);
      (*out)++;
      *bitsAvail = 8;
    };
    shift = num;
    if (num > *bitsAvail) shift = *bitsAvail;
    result <<= shift;
    *bitsAvail -= shift;
    num -= shift;
  };
  **out = (unsigned char)(result >> 8);
  if (*bitsAvail == 0)
  {
    (*out)++;
    *bitsAvail = 8;
  };
}

unsigned int Compress(unsigned char *in, 
                      unsigned int len, 
                      unsigned char *out,
                      unsigned char expTable[20])
{
  int counts[256], table[256];
  int largest;
  unsigned int i, j;
  int k;
  int bitsAvail;
  int byteToEncode;
  unsigned char *dst = out; // Remember the starting address
  // Write the header.
  out[0] = 0x81;
  out[1] = 0x04;
  out[2] = (unsigned char)((len >> 24) & 0xff);
  out[3] = (unsigned char)((len >> 16) & 0xff);
  out[4] = (unsigned char)((len >> 8) & 0xff);
  out[5] = (unsigned char)(len &0xff);
  out[6] = 0;
  out[7] = 0x0c;
  out += 8;
  for (i=0; i<256; i++) 
  {
    table[i] = -1;
    counts[i] = 0;
  };
  if (expTable == NULL)
  {
    for (i=0; i<len; i++)
    {
      counts[in[i]]++;
    };
    for (i=0; i<20; i++)
    {
      largest = -1;
      k = -1;
      for (j=0; j<256; j++)
      {
        if (table[j] != -1) continue;
        if (counts[j] <= largest) continue;
        largest = counts[j];
        k = j;
      };
      table[k] = i;
      *(out++) = (unsigned char)k;
    };
  }
  else
  {
    for (i=0; i<20; i++)
    {
      *(out++) = expTable[i];
      table[expTable[i]] = i;
    };
  };
  //Now for the actual data compression.
  bitsAvail = 8;
  for (;len>0; len--) // While there are more byte so compress;
  {
    byteToEncode = *(in++);
    if (table[byteToEncode] == -1)
    {
      // Encode as 11zzzzzzzz
      shift(&out, &bitsAvail, 2, 0xc0);
      shift(&out, &bitsAvail, 8, (unsigned char)byteToEncode);
      continue;
    };
    if (table[byteToEncode] < 4)
    {
      //Encode as 0xx
      shift(&out, &bitsAvail, 1, 0);
      shift(&out, &bitsAvail, 2, (unsigned char)(table[byteToEncode]<<6));
      continue;
    };
    // Encode as 10yyyy
    shift(&out, &bitsAvail, 2, 0x80);
    shift(&out, &bitsAvail, 4, (unsigned char)((table[byteToEncode] - 4) << 4));
  };
  if (bitsAvail != 8)
  {
    shift(&out, &bitsAvail, bitsAvail, 0);
  };
  if (((out-dst)&1) != 0) shift(&out, &bitsAvail, 8, 0);
  return out-dst;
}

ui16 CCSBuildDoc::compressDSAs(void)
{ //Get rid of any unneeded states/actions
  DSA emptyDSA;
  int i;
  ui16 numDSA = 0;
  for (i=0; i<256; i++)
  {
    if (m_DSAIndex[i] == NULL) continue;
    m_DSAIndex[i]->compress();
    if (*m_DSAIndex[i] != emptyDSA) 
    {
      numDSA++;
      continue;
    };
    delete m_DSAIndex[i];
    m_DSAIndex[i] = NULL;
  };
  return numDSA;
}

void CCSBuildDoc::WriteDSAs(CArchive& ar)
{
  int i;
  unsigned int checksum;
  pDoc->WCS(ar); //Discard old checksum and start new.
  for (i=0; i<256; i++)
  {
    if (m_DSAIndex[i] == NULL) continue;
    pDoc->WCS(&i);
    m_DSAIndex[i]->Write();
  };
  checksum = pDoc->WCS(ar);
  BinaryWrite(ar, &checksum, 4);
}

void CCSBuildDoc::WriteDSALevelIndex(CArchive& ar)
{
  int level, index;
  ui16 DSAindex;
  unsigned char triple[3];
  for (level=0; level<64; level++)
  {
    for (index=0; index<32; index++)
    {
      DSAindex = getDSA(level, index);
      if (DSAindex < 256)
      {
        triple[0] = (unsigned char)level;
        triple[1] = (unsigned char)index;
        triple[2] = (unsigned char)DSAindex;
        BinaryWrite(ar, triple, 3);
      };
    };
  };
  triple[0] = triple[1] = triple[2] = 255;
  BinaryWrite(ar, triple, 3);
}

int WriteGameInfo(CArchive& ar, int size)
{ //Returns size.
  if (size == 0) return 0;
  BinaryWrite(ar, gameInfo, size);
  return size;
}



void CCSBuildDoc::WriteExtendedFeatures(CArchive& ar)
{
  ASSERT(sizeof(EXTENDEDFEATURESBLOCK) == 512);
  EXTENDEDFEATURESBLOCK efb;
  memset(&efb,0,512);
  strcpy(efb.sentinel," Extended Features ");
  efb.dataMapLength = m_dataMapLength;
  efb.dataTypeMapChecksum =
          FormChecksum((ui8 *)m_dataTypeMap, m_dataMapLength);
  efb.dataIndexMapChecksum = 
          FormChecksum((ui8 *)m_dataIndexMap, 2*m_dataMapLength);
  efb.version = ExtendedFeaturesVersion;
  efb.numDSA = compressDSAs();
  efb.editingOptions = pDoc->EditingOptions();
  efb.flags |= EXTENDEDFEATURESBLOCK::LevelDSAInfoPresent;
  efb.flags |= EXTENDEDFEATURESBLOCK::EditingOptionsPresent;
  efb.flags |= EXTENDEDFEATURESBLOCK::IndirectText;
  efb.flags |= EXTENDEDFEATURESBLOCK::ExpandedPortraits;
  efb.flags |= EXTENDEDFEATURESBLOCK::BigActuators;
  efb.gameInfoSize = 0;
  efb.graphicsSignature1 = expectedGraphicsSignature1;
  efb.graphicsSignature2 = expectedGraphicsSignature2;
  efb.CSBgraphicsSignature1 = expectedCSBgraphicsSignature1;
  efb.CSBgraphicsSignature2 = expectedCSBgraphicsSignature2;
  memcpy(efb.hintKey, hintKey, 8);
  if (m_spellFilterLocation.l() < 0)
  {
    efb.spellFilterLocation = 0;
  }
  else
  {
    efb.spellFilterLocation = 
          m_spellFilterLocation.Integer() | 0x80000000;
  };
  efb.cellFlagArraySize = CellFlagArraySize();
  if (gameInfo != NULL) efb.gameInfoSize = strlen(gameInfo);
  if (RecordDesignOption)
          efb.flags |= EXTENDEDFEATURESBLOCK::AutoRecord;
  if (invisibleMonsters)
          efb.extendedFlags |= EXTENDEDFEATURESBLOCK::InvisibleMonsters;
  if (drawAsSize4Monsters)
          efb.extendedFlags |= EXTENDEDFEATURESBLOCK::DrawAsSize4Monsters;
  if (DMRulesOption)
          efb.flags |= EXTENDEDFEATURESBLOCK::DMRules;
  if (   (encipheredDataFile.GetLength() == 1)
      && (encipheredDataFile[0] == 1))
  {
    efb.flags |= EXTENDEDFEATURESBLOCK::SimpleEncryption;
  };
  efb.extendedFeaturesChecksum = 0;
  efb.extendedFeaturesChecksum = 
          FormChecksum((ui8 *)&efb,512);
  BinaryWrite(ar, &efb,512);
  BinaryWrite(ar, m_dataTypeMap, m_dataMapLength);
  BinaryWrite(ar, m_dataIndexMap, 2*m_dataMapLength);
  //*****************************************  
  if (efb.numDSA != 0) WriteDSAs(ar);
  //*****************************************
  WriteGameInfo(ar, efb.gameInfoSize);
  WriteDSALevelIndex(ar);
}

STORESTATUS CCSBuildDoc::StoreDungeon(CArchive& ar)
{
  ui16 Checksum;
  ui16 Hash;
  int dbNum;
  STORESTATUS status;
  int numEnt, entLen;
  //if (d.EmptyHanded==0)
  //{
  //  D1L = GetObjectWeight(d.objectInHand);
  //  LOCAL_4 = D1W;
  //  d.CH16482[d.HandChar].load =
  //        sw(d.CH16482[d.HandChar].load - D1W);//owner of cursor
  //};
  if (encipheredDataFile != "")
  {
    RC4_prepare_key((unsigned char *)(LPCSTR)encipheredDataFile,
                    encipheredDataFile.GetLength());
  };
  WriteExtendedFeatures(ar);
  if (!m_dungeonFormat)
  {
    m_block2.Time = m_Time;
    m_block2.ranseed = m_RandomNumber;
    m_block2.numcharacter = m_NumCharacter;
    m_block2.partyx = m_partyX;
    m_block2.partyy = m_partyY;
    m_block2.partyfacing = m_partyFacing;
    m_block2.partyLevel = m_partyLevel;
    m_block2.handChar = m_HandChar;
    m_block2.MagicCaster = m_MagicCaster;
    m_block2.NumTimer = m_numTimer;
    m_block2.NextAvailTimer = m_NextAvailTimer;
    m_block2.MaxTimers = m_MaxTimers;
    m_block2.ITEM16QueLen = m_ITEM16QueLen;
    m_block2.Long12950 = m_Long12950;
    m_block2.Long12954 = m_Long12954;
    m_block2.Word11710 = m_Word11710;
    m_block2.Word11712 = m_Word11712;
    m_block2.Word11714 = m_Word11714;
    m_block2.ObjectInHand = m_objectInHand.ConvertToInteger();
    m_block2.MaxITEM16 = m_MaxITEM16;
    memset(m_block2.FILL180, 0, 80);
    memset((char *)&memBlocks, 0, 96);
    memBlocks[0].address = (char *)&m_block2;
    memBlocks[0].size = 128;
    memBlocks[1].address = (char *)m_Item16;
    memBlocks[1].size = 16 * m_MaxITEM16;
    memBlocks[2].address = (char *)m_Characters;
    memBlocks[2].size = 3328;
    memBlocks[3].address = (char *)m_Timers;
    memBlocks[3].size = 10 * m_MaxTimers;
    memBlocks[4].address = (char *)m_timerQue;
    memBlocks[4].size = 2 * m_MaxTimers;
    //m_block1.SaveOption = D6W;
    memcpy(m_block1.Byte22808, m_Byte22808, 132);
    m_block1.RandomGameID = m_RandomGameID;
    //m_block1.Byte22596 = (ui8)(m_Word22596);
    //m_block1.Byte22598 = (i8)m_Word22598;
    //m_block1.Word22594 = m_Word22594;
    //m_block1.Word22592 = m_Word22592;
    //m_block1.Block2Hash    = 0; //STRandom();//(i16)STRandom();
    //m_block1.ITEM16Hash    = 0; //STRandom();//(i16)STRandom();
    //m_block1.CharacterHash = 0; //STRandom();//(i16)STRandom();
    //m_block1.TimersHash    = 0; //STRandom();//(i16)STRandom();
    //m_block1.TimerQueHash  = 0; //STRandom();//(i16)STRandom();
    //m_block1.totalMoveCount= totalMoveCount; //STRandom();//(i16)STRandom();
    //m_block1.Hash324       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash326       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash328       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash330       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash332       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash334       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash336       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash338       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash340       = 0; //STRandom();//(i16)STRandom();
    //m_block1.Hash342       = 0; //STRandom();//(i16)STRandom();
    //for (D5W=0; D5W<5; D5W++)
    //{
      //
    swapBlock2(&m_block2);
    m_block1.Block2Checksum = GenChecksum(
                      (i16 *)&m_block2,//
                      m_block1.Block2Hash,
                      128/2);
    swapBlock2(&m_block2);
  };
  if (!m_dungeonFormat)
  {
    swapITEM16s();
    m_block1.ITEM16Checksum = GenChecksum(
                       (i16 *)m_Item16,//
                      m_block1.ITEM16Hash,
                      (16 * m_MaxITEM16)/2);
    swapITEM16s();
  };
  if (!m_dungeonFormat)
  {
    swapCharacterData();
    m_block1.CharacterChecksum = GenChecksum(
                      (i16 *)&m_Characters,//
                      m_block1.CharacterHash,
                      3328/2);
    swapCharacterData();
  };
  if (!m_dungeonFormat)
  {
    swapTimers();
    m_block1.TimersChecksum = GenChecksum(
                      (i16 *)m_Timers,
                      m_block1.TimersHash,
                      (10 * m_MaxTimers)/2);
    swapTimers();
  };
  if (!m_dungeonFormat)
  {
    swapTimerQue();
    m_block1.TimerQueChecksum = GenChecksum(
                      (i16 *)m_timerQue,
                      m_block1.TimerQueHash,
                      (2 * m_MaxTimers)/2);
    swapTimerQue();
  };
  //wordGear(gb1A2+344+2*D5W) = D0W;
  //};
  if (!m_dungeonFormat)
  {
    swapBlock1(&m_block1);
    status = ScrambleAndWrite(ar, (ui16 *)&m_block1);
    swapBlock1(&m_block1);
    if (status != STORE_OK) return status;
    //for (D5W=0; D5W<5; D5W++)
    //{
  };
  if (!m_dungeonFormat)
  {
    swapBlock2(&m_block2);
    status = WriteScrambled(
            ar,
            (ui16 *)&m_block2,
            128,
            m_block1.Block2Hash,
            &Checksum);
    if (status != STORE_OK) return status;
    swapBlock2(&m_block2);
  };
//.........................................................
  if (!m_dungeonFormat)
  {
    swapITEM16s();
    status = WriteScrambled(
            ar,
            (ui16 *)m_Item16,
            16 * m_MaxITEM16,
            m_block1.ITEM16Hash,
            &Checksum);
    if (status != STORE_OK) return status;
    swapITEM16s();
  };
  if (!m_dungeonFormat)
  {
    swapCharacterData();
    status = WriteScrambled(
            ar,
            (ui16 *)m_Characters,
            3328,
            m_block1.CharacterHash,
            &Checksum);
    swapCharacterData();
    if (status != STORE_OK) return status;
  };
//.........................................................
  if (!m_dungeonFormat)
  {
    swapTimers();
    status = WriteScrambled(
            ar,
            (ui16 *)m_Timers,
            10 * m_MaxTimers,
            m_block1.TimersHash,
            &Checksum);
    swapTimers();
  if (status != STORE_OK) return status;
  };
//.........................................................
  if (!m_dungeonFormat)
  {
    swapTimerQue();
    status = WriteScrambled(
            ar,
            (ui16 *)m_timerQue,
            2 * m_MaxTimers,
            m_block1.TimerQueHash,
            &Checksum);
    swapTimerQue();
    if (status != STORE_OK) return status;
  };
//
  //};
  //pwA3 = &Hash;
  if (m_dungeonFormat)
  {
    m_expandedBuffer = (char *)malloc(10,MALLOC000);
    m_expandedBufferSize = 10;
    m_expandedBufferIndex = 0;
  };

  do  //So we can 'break'
  {
    Hash = 0; // initial hash
    swapDungeonDatIndex();
    status = WriteAndChecksum(
                    ar, 
                    (ui8 *)m_dungeonDatIndex, 
                    &Hash, 
                    44);
    swapDungeonDatIndex();
    if (status != STORE_OK) break;
    swapLevelDescriptors(m_dungeonDatIndex->NumLevel());
    status = WriteAndChecksum(
                    ar,
                    (ui8 *)m_pLevelDescriptors,
                    &Hash,
                    16 * m_dungeonDatIndex->NumLevel());
    swapLevelDescriptors(m_dungeonDatIndex->NumLevel());
    if (status != STORE_OK) break;
    swapPointer10454();
    status = WriteAndChecksum(
                    ar,
                    (ui8 *)m_objectColumnIndex,
                    &Hash,
                    2*m_numColumnPointers);
    swapPointer10454();
    if (status != STORE_OK) break;
    swapObjectList(m_dungeonDatIndex->ObjectListLength());
    status = WriteAndChecksum(
                    ar,
                    (ui8 *)m_objectList,
                    &Hash,
                    2*m_dungeonDatIndex->ObjectListLength());
    swapObjectList(m_dungeonDatIndex->ObjectListLength());
    if (status != STORE_OK) break;
    status = WriteAndChecksum(
                    ar,
                    (ui8 *)m_indirectTextIndex,
                    &Hash,
                    4*m_dungeonDatIndex->NumWordsInTextArray());
    if (status != STORE_OK) break;
    status = WriteAndChecksum(
                    ar,
                    (ui8 *)&m_sizeOfCompressedText,
                    &Hash,
                    4);
    if (status != STORE_OK) break;
    status = WriteAndChecksum(
                    ar,
                    (ui8 *)m_compressedText,
                    &Hash,
                    2*m_sizeOfCompressedText);
    if (status != STORE_OK) break;
    for (dbNum=0; dbNum<16; dbNum++)
    {
      numEnt = m_dungeonDatIndex->DBSize(dbNum);//Number of entries
      entLen = m_DBEntrySize[dbNum];
      if (numEnt != 0)
      {
        db.swap(dbNum);
        status = WriteAndChecksum(
                      ar,
                      (ui8 *)db.GetCommonAddress(DBTYPE(dbNum),0),
                      &Hash,
                      numEnt*entLen);
  
        db.swap(dbNum);
      } else status=STORE_OK;
      if (status != STORE_OK) break;
    };
    status = WriteAndChecksum(
                    ar,
                    (ui8 *)m_cellFlagArray,
                    &Hash,
                    CellFlagArraySize());
    if (status != STORE_OK) return status;
    Hash = LE16(Hash);
    status = WriteDataFile(ar,(ui8 *)&Hash, 2); // Write final checksum
    Hash = LE16(Hash);
    if (status != STORE_OK) break;
  } while (status != STORE_OK);
    //if (m_EmptyHanded == 0)
    //{
    //  m_CH16482[m_HandChar].load =
    //        sw(m_CH16482[m_HandChar].load + LOCAL_4);//owner of cursor
    //};
  //m_Word22586 = 1;
  if (m_compressed)
  {
    if (status == STORE_OK)
    {
      char *compressedBuffer;
      int compressedLength;
      compressedBuffer = (char *)malloc(m_expandedBufferIndex+100,MALLOC024);
      compressedLength = Compress((unsigned char *)m_expandedBuffer,
                                  m_expandedBufferIndex,
                                  (unsigned char *)compressedBuffer,
                                  IsModified()?NULL:m_saveExpansionTable);
      BinaryWrite(ar, compressedBuffer, compressedLength);
      free (compressedBuffer);
    };
  };
  if (m_dungeonFormat)
  {
    free (m_expandedBuffer);
  };
  return status;
}

STORESTATUS CCSBuildDoc::WriteDataFile(CArchive& ar,
                                       ui8 *buf,
                                       i32 length)
{
  if (length == 0)
  {
    return STORE_OK;
  };
  if (m_compressed)
  {
    if (m_expandedBufferIndex + length > m_expandedBufferSize)
    {
      m_expandedBufferSize = m_expandedBufferIndex + length;
      m_expandedBuffer = (char *)realloc(m_expandedBuffer,m_expandedBufferSize,MALLOC025);
    };
    memcpy(m_expandedBuffer+m_expandedBufferIndex, buf, length);
    m_expandedBufferIndex += length;
  }
  else
  {
    BinaryWrite(ar, buf,length); //can throw CFileException
  };
  return STORE_OK;
}

STORESTATUS CCSBuildDoc::WriteScrambled(
              CArchive& ar,
              ui16 *buf,
              i32 byteCount,
              i16 initialHash,
              ui16 *checksum)
{
  STORESTATUS status;
  *checksum = Unscramble(buf, initialHash, byteCount/2);
  status = WriteDataFile(ar, (ui8 *)buf, byteCount);
  Unscramble(buf, initialHash, byteCount/2);
  return status;
}

STORESTATUS CCSBuildDoc::WriteAndChecksum(
                     CArchive& ar,
                     ui8 *buf,
                     ui16 *pRunningChecksum, //updated
                     i32 numByte)
{

  if (WriteDataFile(ar, (ui8 *)buf, numByte) != STORE_OK) return STORE_ERROR;
  while (numByte != 0)
  {
    *pRunningChecksum =  (ui16)(*pRunningChecksum + *buf);
    numByte--;
    buf++;
  };
  return STORE_OK;
}
