// ViewLevelInfo.cpp : implementation file
//

#include "stdafx.h"
//#include "types.h"
#include "CSBuild.h"
#include "ViewLevelInfo.h"
#include "CSBuildDoc.h"
#include "Misc.h"
#include "SelectFloorDlg.h"
#include "SelectWall.h"
#include "SelectMonster.h"
#include "SelectDoorDecoration.h"
#include "EditDSAList.h"
#include "Actuator.h"
#include "EditPalette.h"
#include "LevelLocations.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// ViewLevelInfo dialog


ViewLevelInfo::ViewLevelInfo(CWnd* pParent /*=NULL*/)
	: CDialog(ViewLevelInfo::IDD, pParent)
{
	//{{AFX_DATA_INIT(ViewLevelInfo)
	m_Height = 0;
	m_Width = 0;
	m_OffsetX = 0;
	m_OffsetY = 0;
	m_RandomWall = 0;
	m_RandomFloor = 0;
	m_experienceMultiplier = 0;
	m_door1Type = -1;
	m_door2Type = -1;
	m_currentLevel = 0;
	m_defaultSkin = 0;
	//}}AFX_DATA_INIT
}

void ViewLevelInfo::CountMonstersAndDecorations(void)
{
  int i, x, y, lastX, lastY;
  int ordinal;
  LOCATIONREL locr;
  DBTYPE dbtype;
  RN obj;
  DB0 *pDB0;
  DB3 *pDB3;
  DB4 *pDB4;
  ROOMTYPE rt;
  MONSTERTYPE mt;
  ACTUATORCLASS ac;
  for (i=0; i<16; i++)
  {
    m_wallDecorationCount[i] = 0;
    m_floorDecorationCount[i] = 0;
    m_monsterCount[i] = 0;
    m_generatorCount[i] = 0;
    m_doorDecorationCount[i] = 0;
  };
  locr.l(m_currentLevel);
  lastX = pDoc->LastColumn(locr.l());
  lastY = pDoc->LastRow(locr.l());
  for (x = 0; x <= lastX; x++)
  {
    locr.x(x);
    for (y = 0; y <= lastY; y++)
    {
      locr.y(y);
      rt = RoomType(locr);
      for (obj = pDoc->FindFirstObject(locr);
           obj != RNeof;
           obj = pDoc->GetDBRecordLink(obj))
      {
        dbtype = obj.dbType();
        switch (dbtype)
        {
        case dbACTUATOR:
          pDB3 = pDoc->GetRecordAddressDB3(obj);
          ac = pDB3->actuatorClass(rt);
          switch (ac)
          {
          case AC_MONSTERGENERATOR:
            {
              ACTUATOR_MonsterGenerator *pGenerator;
              MONSTERTYPE mt;
              pGenerator = pDB3->Cast2AC_MonsterGenerator(rt);
              mt = pGenerator->monsterType();
              for (i=0; i<pDoc->NumMonsterType(locr.l()); i++)
              {
                if (pDoc->MonsterType(locr.l(),i) == mt)
                {
                  m_generatorCount[i]++;
                };
              };
            };
            break;
          };
          ordinal = pDB3->GraphicOrdinal(locr);
          switch (rt)
          {
          case roomOPEN:
          case roomTELEPORTER:
          case roomPIT:
          case roomSTAIRS:
          case roomDOOR:
          case roomFALSEWALL:
            m_floorDecorationCount[ordinal]++;
            break;
          case roomSTONE:
            m_wallDecorationCount[ordinal]++;
            break;
          };
          break;
        case dbDOOR:
          pDB0 = pDoc->GetRecordAddressDB0(obj);
          ordinal = pDB0->ornateness();
          m_doorDecorationCount[ordinal]++;
          break;
        case dbMONSTER:
          pDB4 = pDoc->GetRecordAddressDB4(obj);
          mt = pDB4->monsterType();
          for (i=0; i<pDoc->NumMonsterType(locr.l()); i++)
          {
            if (pDoc->MonsterType(locr.l(),i) == mt)
            {
              m_monsterCount[i]++;
            };
          };
          break;
        };
      };
    };
  };
};


void ViewLevelInfo::FillWallDecorations(void)
{
  int i;
  WALLTYPE wt;
  CString name;
  int index;
  m_WallDecorations.ResetContent();
  for (i=0; i<pDoc->NumWallDecorations(m_currentLevel); i++)
  {
    if (pDoc->IsLegalWallDecorationIndex(m_currentLevel,i+1))
    {
      wt = pDoc->WallDecorationType(m_currentLevel,i);
      //name = wt.Name();
      name.Format("%s(%d)",wt.Name(),m_wallDecorationCount[i+1]);
    }
    else
    {
      name = "illegal";
    };
    index = m_WallDecorations.AddString(NicelyDecapitalize(name));
    m_WallDecorations.SetItemData(index,wt.Integer());
  };
}

void ViewLevelInfo::FillDoorDecorations(void)
{
  int i;
  int index;
  CString name;
  ORNATETYPE ot; //Ornate Door type
  m_doorDecoration.ResetContent();
  for (i=0; i<pDoc->NumDoorDecorations(m_currentLevel); i++)
  {
    ot = pDoc->DoorDecorationType(m_currentLevel,i);
    name.Format("%s(%d)",ot.Name(),m_doorDecorationCount[i+1]);
    index = m_doorDecoration.AddString(name);
    m_doorDecoration.SetItemData(index, ot.Integer());
  };
}

void ViewLevelInfo::FillFloorDecorations(void)
{
  int i;
  int index;
  FLOORTYPE ft;
  CString name;
  m_FloorList.ResetContent();
  for (i=0; i<pDoc->NumFloorDecorations(m_currentLevel); i++)
  {
    if (pDoc->IsValidFloorDecorationOrdinal(m_currentLevel,i+1))
    {
      ft = pDoc->FloorDecorationType(m_currentLevel,i+1);
      name.Format("%s(%d)",ft.Name(),m_floorDecorationCount[i+1]);
      index = m_FloorList.AddString(name);
      m_FloorList.SetItemData(index, ft.Integer());
    }
    else
    {
      index = m_FloorList.AddString("<none>");
      m_FloorList.SetItemData(index, (DWORD)-1);
    };
  };
}

void ViewLevelInfo::FillMonsterList(void)
{
  int i;
  int index;
  MONSTERTYPE mt;
  CString name;
  m_MonsterList.ResetContent();
  for (i=0; i<pDoc->NumMonsterType(m_currentLevel); i++)
  {
    mt = pDoc->MonsterType(m_currentLevel,i);
    name.Format("%s(%d,%d)",
                  mt.Name(),
                  m_monsterCount[i],
                  m_generatorCount[i]);
    index = m_MonsterList.AddString(name);
    m_MonsterList.SetItemData(index, mt.Integer());
  };
}

void ViewLevelInfo::ExtractDataFromRecord(void)
{
  m_Width = pDoc->LastColumn(m_currentLevel) + 1;
  m_Height = pDoc->LastRow(m_currentLevel) + 1;
  m_OffsetX = pDoc->OffsetX(m_currentLevel);
  m_OffsetY = pDoc->OffsetY(m_currentLevel);
  m_RandomWall = pDoc->NumRandomWallDecorations(m_currentLevel);
  m_RandomFloor = pDoc->NumRandomFloorDecorations(m_currentLevel);
  m_experienceMultiplier = pDoc->ExperienceMultiplier(m_currentLevel);
  m_door1Type = pDoc->DoorType(m_currentLevel, 0);
  m_door2Type = pDoc->DoorType(m_currentLevel, 1);
  m_defaultSkin = pDoc->DefaultSkin(m_currentLevel);
}

void ViewLevelInfo::EncodeDataIntoRecord(void)
{
  pDoc->OffsetX(m_currentLevel, m_OffsetX);
  pDoc->OffsetY(m_currentLevel, m_OffsetY);
  if (m_RandomWall > (unsigned)pDoc->NumWallDecorations(m_currentLevel))
  {
    m_RandomWall = pDoc->NumWallDecorations(m_currentLevel);
  };
  if (m_RandomFloor > (unsigned)pDoc->NumFloorDecorations(m_currentLevel))
  {
    m_RandomFloor = pDoc->NumFloorDecorations(m_currentLevel);
  };
  pDoc->NumRandomWallDecorations(m_currentLevel,m_RandomWall);
  pDoc->NumRandomFloorDecorations(m_currentLevel,m_RandomFloor);
  pDoc->ExperienceMultiplier(m_currentLevel, m_experienceMultiplier);
  pDoc->DoorType(m_currentLevel, 0, m_door1Type);
  pDoc->DoorType(m_currentLevel, 1, m_door2Type);
  pDoc->DefaultSkin(m_currentLevel, m_defaultSkin);
}

void ViewLevelInfo::Show(int id, bool show)
{
  CWnd *pWnd;
  pWnd = GetDlgItem(id);
  if (pWnd == NULL) return;
  pWnd->ShowWindow(show);
}

void ViewLevelInfo::DoDataExchange(CDataExchange* pDX)
{
  if (!pDoc->IsValid()) return;
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(ViewLevelInfo)
	DDX_Control(pDX, IDC_DoorDecoration, m_doorDecoration);
	DDX_Control(pDX, IDC_FloorList, m_FloorList);
	DDX_Control(pDX, IDC_MonsterList, m_MonsterList);
	DDX_Control(pDX, IDC_WallDecorations, m_WallDecorations);
	DDX_Text(pDX, IDC_Height, m_Height);
	DDV_MinMaxUInt(pDX, m_Height, 1, 32);
	DDX_Text(pDX, IDC_Width, m_Width);
	DDV_MinMaxUInt(pDX, m_Width, 1, 32);
	DDX_Text(pDX, IDC_OffsetX, m_OffsetX);
	DDX_Text(pDX, IDC_OffsetY, m_OffsetY);
	DDX_Text(pDX, IDC_RandomWall, m_RandomWall);
	DDX_Text(pDX, IDC_RandomFloor, m_RandomFloor);
	DDX_Text(pDX, IDC_ExperienceMultiplier, m_experienceMultiplier);
	DDV_MinMaxUInt(pDX, m_experienceMultiplier, 0, 15);
	DDX_Radio(pDX, IDC_Door1Type, m_door1Type);
	DDX_Radio(pDX, IDC_Door2Type, m_door2Type);
	DDX_Text(pDX, IDC_CurrentLevel, m_currentLevel);
	DDV_MinMaxUInt(pDX, m_currentLevel, 0, 63);
	DDX_Text(pDX, IDC_DefaultSkin, m_defaultSkin);
	DDV_MinMaxUInt(pDX, m_defaultSkin, 0, 255);
	//}}AFX_DATA_MAP
  if (!pDX->m_bSaveAndValidate)
  {
    CountMonstersAndDecorations();
    FillWallDecorations();
    FillMonsterList();
    FillFloorDecorations();
    FillDoorDecorations();
    Show(IDC_NewWall,   pDoc->NumWallDecorations(m_currentLevel)<15);
    Show(IDC_NewFloor,  pDoc->NumFloorDecorations(m_currentLevel)<15);
    Show(IDC_NewDoor,   pDoc->NumDoorDecorations(m_currentLevel)<15);
    Show(IDC_NewMonster,pDoc->NumMonsterType(m_currentLevel)<15);
  };
}


BEGIN_MESSAGE_MAP(ViewLevelInfo, CDialog)
	//{{AFX_MSG_MAP(ViewLevelInfo)
	ON_BN_CLICKED(IDC_AddColumn, OnAddColumn)
	ON_BN_CLICKED(IDC_AddRow, OnAddRow)
	ON_BN_CLICKED(IDC_NewFloor, OnNewFloor)
	ON_BN_CLICKED(IDC_NewWall, OnNewWall)
	ON_BN_CLICKED(IDC_NewMonster, OnNewMonster)
	ON_BN_CLICKED(IDC_EditDSAList, OnEditDSAList)
	ON_BN_CLICKED(IDC_NewDoorDecoration, OnNewDoorDecoration)
	ON_BN_CLICKED(IDC_ReplaceDoorDecoration, OnReplaceDoorDecoration)
	ON_BN_CLICKED(IDC_ReplaceFloor, OnReplaceFloor)
	ON_BN_CLICKED(IDC_ReplaceWall, OnReplaceWall)
	ON_BN_CLICKED(IDC_ReplaceMonster, OnReplaceMonster)
	ON_BN_CLICKED(IDC_WallDecorationsEverywhere, OnWallDecorationsEverywhere)
	ON_BN_CLICKED(IDC_WallDecorationsNowhere, OnWallDecorationsNowhere)
	ON_BN_CLICKED(IDC_FloorDecorationsEverywhere, OnFloorDecorationsEverywhere)
	ON_BN_CLICKED(IDC_FloorDecorationsNowhere, OnFloorDecorationsNowhere)
	ON_WM_HELPINFO()
	ON_BN_CLICKED(IDC_EditPalette, OnEditPalette)
	ON_EN_CHANGE(IDC_CurrentLevel, OnChangeCurrentLevel)
	ON_NOTIFY(UDN_DELTAPOS, IDC_SpinLevel, OnDeltaposSpinLevel)
	ON_BN_CLICKED(IDC_ExportLevel, OnExportLevel)
	ON_BN_CLICKED(IDC_Import, OnImport)
	ON_BN_CLICKED(IDC_EditSpecialLocations, OnEditSpecialLocations)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// ViewLevelInfo message handlers

void ViewLevelInfo::OnOK() 
{
	// TODO: Add extra validation here
  UpdateData(true);
	EncodeDataIntoRecord();
	CDialog::OnOK();
}

void ViewLevelInfo::OnAddColumn() 
{
	// TODO: Add your control notification handler code here
  if (!pDoc->VerifyLevelData())
  {
    DungeonError();
    return;
  };
	if (!pDoc->AddColumn(m_currentLevel))
  {
    EnlargeError();
    return;
  };
  ExtractDataFromRecord();
  UpdateData(false);
  pDoc->UpdateAllViews(NULL);
}

void ViewLevelInfo::OnAddRow() 
{
	// TODO: Add your control notification handler code here
  if (!pDoc->VerifyLevelData())
  {
    DungeonError();
    return;
  };
	if (!pDoc->AddRow(m_currentLevel))
  {
    EnlargeError();
    return;
  };
  ExtractDataFromRecord();
  UpdateData(false);
  pDoc->UpdateAllViews(NULL);
}

BOOL ViewLevelInfo::OnInitDialog() 
{
	CDialog::OnInitDialog();
	m_currentLevel = pDoc->CurrentDisplayLevel();
  ExtractDataFromRecord();
	UpdateData(false);
	// TODO: Add extra initialization here
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void ViewLevelInfo::PopListItem(int offset,
                                CListBox *box, 
                                char *msg)
{
  int index;
  CString MSG;
  UpdateData(true);
  index = box->GetCurSel();
  if (index < 0) 
  {
    MSG.Format("You must select a %s",msg);
    return;
    Error(MSG);
  };
  pDoc->RotateDecoration(m_currentLevel,offset,index);
  UpdateData(false);
}

/*
void ViewLevelInfo::OnDeleteFloor() 
{
	// TODO: Add your control notification handler code here
	
}
*/
/*
void ViewLevelInfo::OnDeleteMonster() 
{
	// TODO: Add your control notification handler code here
	
}
*/
/*
void ViewLevelInfo::OnDeleteWall() 
{
	// TODO: Add your control notification handler code here
	
}
*/

void ViewLevelInfo::OnNewFloor() 
{
	// TODO: Add your control notification handler code here
	FLOORTYPE ft;
  SelectFloorDlg dlg;
  UpdateData(true);
  if (dlg.DoModal() == IDCANCEL) return;
  pDoc->AddFloorDecoration(m_currentLevel, dlg.m_floorType);
  UpdateData(false);
}

void ViewLevelInfo::OnReplaceFloor() 
{
	// TODO: Add your control notification handler code here
  int index;
	FLOORTYPE ft;
  SelectFloorDlg dlg;
  UpdateData(true);
	// TODO: Add your control notification handler code here
  index = m_FloorList.GetCurSel();
  if (index < 0)
  {
    pDoc->Sorry("You have not selected\na Floor Decoration to replace");
    return;
  };
  if (dlg.DoModal() == IDCANCEL) return;
  pDoc->ReplaceFloorDecoration(m_currentLevel, dlg.m_floorType, index+1);
  UpdateData(false);	
}

void ViewLevelInfo::OnNewWall() 
{
	// TODO: Add your control notification handler code here
	
	WALLTYPE wt;
  SelectWall dlg;
  UpdateData(true);
  if (dlg.DoModal() == IDCANCEL) return;
  pDoc->AddWallDecoration(m_currentLevel, dlg.m_wallType);
  UpdateData(false);
}

/*
void ViewLevelInfo::OnPopFloor() 
{
	// TODO: Add your control notification handler code here
  PopListItem(pDoc->NumMonsterType(m_currentLevel)
                 +pDoc->NumWallDecorations(m_currentLevel),
              &m_FloorList,
              "Floor Decoration");
	
}
*/
/*
void ViewLevelInfo::OnPopMonster() 
{
	// TODO: Add your control notification handler code here
  PopListItem(0,
              &m_MonsterList,
              "Monster Type");
	
}
*/
/*
void ViewLevelInfo::OnPopWall() 
{
	// TODO: Add your control notification handler code here
  PopListItem(pDoc->NumMonsterType(m_currentLevel),
              &m_WallDecorations,
              "Wall Decoration");
}
*/

void ViewLevelInfo::OnNewMonster() 
{
	// TODO: Add your control notification handler code here
	MONSTERTYPE mt;
  SelectMonster dlg;
  UpdateData(true);
  if (dlg.DoModal() == IDCANCEL) return;
  pDoc->AddMonsterType(m_currentLevel, dlg.m_monsterType);
  UpdateData(false);
	
}
/*
void ViewLevelInfo::OnDeleteDoorDecoration() 
{
	// TODO: Add your control notification handler code here
	
}
*/
void ViewLevelInfo::OnNewDoorDecoration() 
{
	// TODO: Add your control notification handler code here
	ORNATETYPE ot;
  SelectDoorDecoration dlg;
  UpdateData(true);
  if (dlg.DoModal() == IDCANCEL) return;
  pDoc->AddDoorDecoration(m_currentLevel, dlg.m_ornateType);
  UpdateData(false);
	
}

void ViewLevelInfo::OnReplaceDoorDecoration() 
{
  int index;
	ORNATETYPE ot;
  SelectDoorDecoration dlg;
  UpdateData(true);
	// TODO: Add your control notification handler code here
  index = m_doorDecoration.GetCurSel();
  if (index < 0)
  {
    pDoc->Sorry("You have not selected\na Door Decoration to replace");
    return;
  };
  if (dlg.DoModal() == IDCANCEL) return;
  pDoc->ReplaceDoorDecoration(m_currentLevel, dlg.m_ornateType, index);
  UpdateData(false);
}

/*
void ViewLevelInfo::OnPopDoorDecoration() 
{
	// TODO: Add your control notification handler code here
	
  PopListItem(pDoc->NumMonsterType(m_currentLevel)
                 +pDoc->NumWallDecorations(m_currentLevel)
                 +pDoc->NumFloorDecorations(m_currentLevel),
              &m_doorDecoration,
              "Door Decoration");
}
*/
void ViewLevelInfo::OnEditDSAList() 
{
	// TODO: Add your control notification handler code here
	EditDSAList dlg(m_currentLevel);
  dlg.DoModal();
}

void ViewLevelInfo::OnReplaceWall() 
{
	// TODO: Add your control notification handler code here
  int index;
	WALLTYPE wt;
  SelectWall dlg;
  UpdateData(true);
	// TODO: Add your control notification handler code here
  index = m_WallDecorations.GetCurSel();
  if (index < 0)
  {
    pDoc->Sorry("You have not selected\na Wall Decoration to replace");
    return;
  };
  if (dlg.DoModal() == IDCANCEL) return;
  pDoc->ReplaceWallDecoration(m_currentLevel, dlg.m_wallType, index);
  UpdateData(false);
}

void ViewLevelInfo::OnReplaceMonster() 
{
	// TODO: Add your control notification handler code here
  int index, result;
	MONSTERTYPE oldmt;
  SelectMonster dlg;
  UpdateData(true);
  index = m_MonsterList.GetCurSel();

  if (index < 0)
  {
    pDoc->Sorry("You have not selected\na Monster to replace");
    return;
  };
  if (dlg.DoModal() == IDCANCEL) return;
  oldmt = pDoc->MonsterType(m_currentLevel, index);
  pDoc->ReplaceMonsterType(m_currentLevel, dlg.m_monsterType, index);
  result = MessageBox(
             "Should we replace all instances\nof these montsters on this level?",
             "Question",
             MB_YESNO);
  if (result == IDYES)
  {
    LOCATIONREL locr;
    RN RNmon;
    DB4 *pMon;
    locr.x(0);
    locr.y(0);
    for (locr.l(m_currentLevel); (unsigned)locr.l()==m_currentLevel; locr.Next())
    {
      for (RNmon = pDoc->FindFirstMonster(locr); RNmon!=RNeof; RNmon=pDoc->FindNextMonster(RNmon))
      {
        pMon = pDoc->GetRecordAddressDB4(RNmon);
        if (pMon->monsterType() == oldmt) pMon->monsterType(dlg.m_monsterType);
      };
    };
  };
  UpdateData(false);	
}

void ViewLevelInfo::RandomDecorations(ROOMTYPE rt, bool place)
{
  int row, col;
  CELLFLAG cf;
  pDoc->ClearCellCheckpoints();
  for (col=0; col<=pDoc->LastColumn(m_currentLevel); col++)
  {
    for (row=0; row<=pDoc->LastRow(m_currentLevel); row++)
    {
      cf = pDoc->GetCellFlag(m_currentLevel,col,row);
      if (RoomType(cf) == roomSTONE)
      {
        if (rt == roomSTONE)
        {
          if (place)
          {
            int newDecorations;
            newDecorations = 0xf; //Assume all sides
            if (RoomType(m_currentLevel, col, row-1) == roomDOOR) newDecorations &= 0x7;
            if (RoomType(m_currentLevel, col+1, row) == roomDOOR) newDecorations &= 0xb;
            if (RoomType(m_currentLevel, col, row+1) == roomDOOR) newDecorations &= 0xd;
            if (RoomType(m_currentLevel, col-1, row) == roomDOOR) newDecorations &= 0xe;
            if (RoomType(m_currentLevel, col, row-1) == roomSTAIRS) newDecorations &= 0x7;
            if (RoomType(m_currentLevel, col+1, row) == roomSTAIRS) newDecorations &= 0xb;
            if (RoomType(m_currentLevel, col, row+1) == roomSTAIRS) newDecorations &= 0xd;
            if (RoomType(m_currentLevel, col-1, row) == roomSTAIRS) newDecorations &= 0xe;
            cf = (CELLFLAG)(cf | newDecorations);
          }
          else       cf = (CELLFLAG)(cf & 0xf0);
        };
      }
      else if (RoomType(cf) == roomOPEN)
      {
        if (rt == roomOPEN)
        {
          if (place) cf = (CELLFLAG)(cf | 0x8);
          else       cf = (CELLFLAG)(cf & 0xf7);
        };
      };
      pDoc->SetCellFlag(m_currentLevel, col, row, cf);
    };
  };
}

void ViewLevelInfo::OnWallDecorationsEverywhere() 
{
	// TODO: Add your control notification handler code here
  int ans;
	ans = AreYouSure("Put Random Wall Decorations EVERYWHERE\n"
                   "in this level?");
  if (ans == IDNO) return;
  RandomDecorations(roomSTONE, true);
}

void ViewLevelInfo::OnWallDecorationsNowhere() 
{
	// TODO: Add your control notification handler code here
	
  int ans;
	ans = AreYouSure("Remove Random Wall Decorations EVERYWHERE\n"
                   "in this level?");
  if (ans == IDNO) return;
  RandomDecorations(roomSTONE, false);
}

void ViewLevelInfo::OnFloorDecorationsEverywhere() 
{
	// TODO: Add your control notification handler code here
	
  int ans;
	ans = AreYouSure("Put Random Floor Decorations EVERYWHERE\n"
                   "in this level?");
  if (ans == IDNO) return;
  RandomDecorations(roomOPEN, true);
}

void ViewLevelInfo::OnFloorDecorationsNowhere() 
{
	// TODO: Add your control notification handler code here
	
  int ans;
	ans = AreYouSure("Remove Random Floor Decorations EVERYWHERE\n"
                   "in this level?");
  if (ans == IDNO) return;
  RandomDecorations(roomOPEN, false);
}

BOOL ViewLevelInfo::OnHelpInfo(HELPINFO* pHelpInfo) 
{
	// TODO: Add your message handler code here and/or call default
	
	CString s;
  int id;
  switch (pHelpInfo->iCtrlId)
  {
  case IDC_MonsterList: id = IDH_LevelInfo_MonsterList; break;
  default: id = IDH_Context_None;
  };
	s.LoadString(id);
  HelpMsg(s);
  return true;
}

void ViewLevelInfo::OnEditPalette() 
{
	// TODO: Add your control notification handler code here
	EditPalette dlg;
  PALETTE palette;
  dlg.m_pPalette = &palette;
  dlg.DoModal();
}

void ViewLevelInfo::OnChangeCurrentLevel() 
{
	// TODO: If this is a RICHEDIT control, the control will not
	// send this notification unless you override the CDialog::OnInitDialog()
	// function and call CRichEditCtrl().SetEventMask()
	// with the ENM_CHANGE flag ORed into the mask.
	
  int oldLevel, newLevel;
  oldLevel = m_currentLevel;
	UpdateData(true);
  newLevel = m_currentLevel;
  m_currentLevel = oldLevel;
  EncodeDataIntoRecord();
  m_currentLevel = newLevel;
  if (m_currentLevel >= (unsigned)pDoc->NumLevel()) m_currentLevel = 0;
  ExtractDataFromRecord();
  UpdateData(false);
	
}

void ViewLevelInfo::OnDeltaposSpinLevel(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_UPDOWN* pNMUpDown = (NM_UPDOWN*)pNMHDR;
  int newLevel;
	// TODO: Add your control notification handler code here
  UpdateData(true);
	pNMUpDown->iDelta = -pNMUpDown->iDelta;
  newLevel = m_currentLevel + pNMUpDown->iDelta;
  if (newLevel < 0 ) return;
  if (newLevel >= pDoc->NumLevel()) return;
  EncodeDataIntoRecord();
  m_currentLevel = newLevel;
  ExtractDataFromRecord();
  UpdateData(false);
	*pResult = 0;
}

void ViewLevelInfo::OnExportLevel() 
{
	// TODO: Add your control notification handler code here
	
  FILE *f;
  CString defaultFname;
  defaultFname.Format("ExportLevel%02d.txt", this->m_currentLevel);
	CFileDialog dlg(false,"txt",defaultFname);
  dlg.m_ofn.lpstrTitle = "Export Level to text file";
  if (dlg.DoModal() != IDOK) return;
  f = fopen(dlg.GetPathName(),"w");
  if (f == NULL)
  {
    Error("I cannot open that file for writing");
    return;
  };
  ExportLevel(f);
  fclose(f);
}

void ViewLevelInfo::OnImport() 
{
	// TODO: Add your control notification handler code here
	
  FILE *f;
  CString defaultFname;
  defaultFname.Format("ExportLevel%02d.txt", this->m_currentLevel);
	CFileDialog dlg(true,"txt",defaultFname);
  dlg.m_ofn.lpstrTitle = "Import Level Data from text file";
  if (dlg.DoModal() != IDOK) return;
  f = fopen(dlg.GetPathName(),"r");
  if (f == NULL)
  {
    Error("I cannot open that file for reading");
    return;
  };
  ImportLevel(f);
  fclose(f);
}

void ViewLevelInfo::OnEditSpecialLocations() 
{
	// TODO: Add your control notification handler code here
	LevelLocations dlg(m_currentLevel);
  dlg.DoModal();
}
