#include "stdafx.h"
//#include "types.h"
#include "CSBuildDoc.h"
#include "Actuator.h"
#include "Misc.h"

void ExportTarget(FILE *f, 
                  const EXPORTLOC& src, 
                  const LOCATIONREL& dst);

bool AutoEnlarge = false;
bool RecordDesignOption = false;
bool DMRulesOption = false;
bool invisibleMonsters = false;
bool drawAsSize4Monsters = false;
bool extendedPortraits = false;
char ExtendedFeaturesVersion = '@';
int  ExtendedFeaturesSize = 0;
char *gameInfo = NULL;

char *teleporterFacing[8] =
{
  "None", "90", "180", "270",
  "North", "East", "South", "West"
};

char *teleporterWhat[8] =
{
  "Objects", "Monsters", "PartyObjects", "Anything"
};




DATABASES db;

i32 dbEntrySizes[16] =
{
  sizeof (DB0),
  sizeof (DB1),
  sizeof (DB2),
  sizeof (DB3),
  sizeof (DB4),
  sizeof (DB5),
  sizeof (DB6),
  sizeof (DB7),
  sizeof (DB8),
  sizeof (DB9),
  sizeof (DB10),
  sizeof (DB11),
  sizeof (DB12),
  sizeof (DB13),
  sizeof (DB14),
  sizeof (DB15)
};

char *dbNames[16] =
{
  "Doors",        //0
  "Teleporters",  //1
  "Text",         //2
  "Actuators",    //3
  "Monsters",     //4
  "Weapons",      //5
  "Clothes",      //6
  "Scrolls",      //7
  "potions",      //8
  "chests",       //9
  "food, misc",   //10
  "Not Used",     //11
  "Not Used",     //12
  "Not Used",     //13
  "Missiles",     //14
  "Clouds"        //15
};

DATABASES::DATABASES (void)
{
  i32 i;
  for (i=0; i<16; i++)
  {
    m_Address[i] = NULL;
    m_numEnt[i] = 0;
  };
}

DATABASES::~DATABASES (void)
{
  i32 i;
  for (i=0; i<16; i++) DeAllocate(i);
}

const RN DBCOMMON::link(void) const
{
  return m_link;
}

void DBCOMMON::link(RN rn)
{
  m_link = rn;
}

bool DATABASES::IsDBIndexOK(DBTYPE dbType, int dbIndex)
{
  return dbIndex < m_numEnt[dbType];
}

DBCOMMON *DATABASES::GetCommonAddress(DBTYPE dbType, int dbIndex)
{
  ASSERT(dbIndex < m_numEnt[dbType]);
  return (DBCOMMON *)
       (((char *)m_Address[dbType]) 
            + dbIndex * dbEntrySizes[dbType]);
}

DBCOMMON *DATABASES::GetCommonAddress(RN object)
{
  i32 dbNum, index;
  ASSERT((object!=RNempty) && !object.IsSpell());
  dbNum = object.dbNum();
  index = object.dbIndex();
  return GetCommonAddress(DBTYPE(dbNum), index);
}

void DATABASES::SetAddress(DBTYPE dbType, DBCOMMON *address)
{
  m_Address[dbType] = address;
}

i32 DATABASES::NumEntry(int dbNum)
{
  return m_numEnt[dbNum];
}

void DATABASES::NumEntry(int dbNum, int numEntry)
{
  m_numEnt[dbNum] = numEntry;
  pDoc->SetDBSize(dbNum, numEntry);
}

CString DATABASES::DBName(int dbNum)
{
  return CString(dbNames[dbNum]);
}

i32 DATABASES::NumUsedEntry(DBTYPE dbType)
{
  int result = 0;
  int i, dbNum;
  DBCOMMON *entryAddr;
  i32 DBsize, entrySize;
  dbNum = dbType;
  DBsize = db.m_numEnt[dbNum]; // # entries
  DBsize -=3; //Always leave a bit or room for engine.
  entrySize = dbEntrySizes[dbNum];
  entryAddr = db.m_Address[dbNum];
  for (i=0; i<m_numEnt[dbNum]; i++)
  {
    if (entryAddr->link() != RNnul)
    {
      result++;
    };
    entryAddr = (DBCOMMON *)((ui8 *)entryAddr + entrySize);
  };
  return result;
}

void DATABASES::Allocate(i32 dbNum, i32 numEnt)
{
  DeAllocate(dbNum); // In case something is there already.
  ASSERT((dbNum & 0xfffffff0) == 0);
  dbNum &= 15;
  ASSERT(numEnt <= ((dbNum==15) ? 768 : ((dbNum==3)?16000:8000)));
  m_numEnt[dbNum] = numEnt;
  if (numEnt > 0)
  {
    m_Address[dbNum] = (DBCOMMON *)
             malloc(dbEntrySizes[dbNum] * numEnt,MALLOC003);
    if (m_Address[dbNum]==NULL)
    {
      DIE("Cannot allocate memory");
    };
    ui8 *p = (ui8 *)m_Address[dbNum];
    for (i32 i=0; i<m_numEnt[dbNum]; i++)
    {
      *(RN *)p = RNnul; // Mark entry as empty
      p += dbEntrySizes[dbNum];
    }
  }
  else
  {
    m_Address[dbNum] = NULL;
  };
}

void DATABASES::DeAllocate(i32 dbNum)
{
  ASSERT((dbNum & 0xfffffff0) == 0);
  dbNum &= 15;
  if (m_Address[dbNum] != NULL) free(m_Address[dbNum]);
  m_Address[dbNum] = NULL;
  m_numEnt[dbNum] = 0;
}

RN DATABASES::Enlarge(int dbNum)
{
  i32 max, iNew, i, newSize;
  i32 result;
  DBCOMMON *pC;
  char msg[1000];
  ASSERT( (dbNum & 0xfffffff0) == 0);
  if (m_noEnlarge & (1 << dbNum)) return RNnul;
  max = (dbNum==15) ? 767 : ((dbNum==3)?16000:MAXDBSIZE);
  if (m_numEnt[dbNum] >= max)
  {
    return RNnul;//Already as big as possible.
  };

  newSize = m_numEnt[dbNum] + 20;
  result = IDOK;
  if (newSize > max-500)
  {
    sprintf(msg,
             "Database Number %d (containing %s)\n"
             "has become full.  There is a rather arbitrary\n"
             "limit of %d items in this database. This limit\n"
             "is being approached.\n"
             "\n"
             "Do you want me to enlarge the database?"
             ,dbNum
             ,dbNames[dbNum]
             ,MAXDBSIZE);
    if (AutoEnlarge) result = IDYES;
    else result = MessageBox(NULL,msg,"Error",MB_YESNO);
  };
  if (result == IDNO)
  {
    //m_noEnlarge |= 1 << dbNum;
    return RNnul;
  };
  m_Address[dbNum] =
     (DBCOMMON *)realloc(m_Address[dbNum], newSize*dbEntrySizes[dbNum],MALLOC004);
  if (m_Address[dbNum] == NULL)
  {
    DIE("Cannot allocate memory");
  };
  pDoc->SetDBSize(dbNum, newSize);
  // We need to clear all the entries.
  iNew = m_numEnt[dbNum]; // First entry to clear
  m_numEnt[dbNum] = newSize;
  for (i=iNew; i<newSize; i++)
  {
    pC = GetCommonAddress(DBTYPE(dbNum),i);
    memset((char *)pC,0,dbEntrySizes[dbNum]);
    pC->link(RNnul);
  };
  pC = GetCommonAddress(DBTYPE(dbNum),iNew);
  pC->link(RNeof);
  return RN(DBTYPE(dbNum),iNew);
}

bool DATABASES::IsDBType(DBCOMMON *pDB, DBTYPE type)
{
  if (    (pDB >= m_Address[type])
       && (pDB <  m_Address[type] + dbEntrySizes[type] * m_numEnt[type])) return true;
  return false;
}

RN DATABASES::AllocateSegment(int segment, DBTYPE dbType)
{
  int entrySize;
  ASSERT (m_Address[segment] == NULL);
  ASSERT (m_numEnt[segment] == 0);
  entrySize = dbEntrySizes[dbType];
  m_Address[segment] = (DBCOMMON *)malloc(2*entrySize, MALLOC027);
  m_numEnt[segment] = 2;
  pDoc->SetDBSize(segment, 2);
  db.GetCommonAddress(DBTYPE(segment), 0)->link(RNnul);
  pDoc->SetModifiedFlag();
  return RN(DBTYPE(segment),1);
}

// TAG00a0b0
RN DATABASES::FindEmptyDBEntry(DBTYPE dbType)
{//(RN)
  DBCOMMON *entryAddr;
  RN result;
  int dbNum;
  i32 j;
  i32 DBsize, entrySize;
  ASSERT((dbType&0x7fff) < 16);
  result = RNnul;
  entrySize = dbEntrySizes[dbType];
  dbNum = dbType;
  DBsize = db.m_numEnt[dbNum]; // # entries
  DBsize -=3; //Always leave a bit or room for engine.
  entryAddr = (DBCOMMON *)(((ui8 *)db.m_Address[dbNum]) + entrySize);
  for (j=1; j<DBsize; j++)
  {
    if (entryAddr->link() == RNnul)
    {
      result = RN(DBTYPE(dbNum),j);
      break; // Eentry is empty
    };
    entryAddr = (DBCOMMON *)((ui8 *)entryAddr + entrySize);
  };
  if (result == RNnul)
  { //First try to enlarge existing segments
    dbNum = dbType;
    result = Enlarge(dbNum);
  };
  if (result == RNnul) return result;
  entryAddr = GetCommonAddress(result);
  memset((ui8 *)entryAddr, 0, entrySize);
  entryAddr->link(RNeof);
  pDoc->SetModifiedFlag();
  return result;
}

void DATABASES::swapDOOR(int dbNum)
{
  DB0 *pDB0;
  pDB0 = m_Address[dbNum]->CastToDB0();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB0[i].swap();
  };
}

void DATABASES::swapTELEPORTER(int dbNum)
{
  DB1 *pDB1;
  pDB1 = m_Address[dbNum]->CastToDB1();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB1[i].swap();
  };
}

void DATABASES::swapTEXT(int dbNum)
{
  DB2 *pDB2;
  pDB2 = m_Address[dbNum]->CastToDB2();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB2[i].swap();
  };
}

void DATABASES::swapACTUATOR(int dbNum)
{
  DB3 *pDB3;
  pDB3 = m_Address[dbNum]->CastToDB3();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB3[i].swap();
  };
}

void DATABASES::swapMONSTER(int dbNum)
{
  DB4 *pDB4;
  pDB4 = m_Address[dbNum]->CastToDB4();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB4[i].swap();
  };
}

void DATABASES::swapWEAPON(int dbNum)
{
  DB5 *pDB5;
  pDB5 = m_Address[dbNum]->CastToDB5();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB5[i].swap();
  };
}

void DATABASES::swapCLOTHING(int dbNum)
{
  DB6 *pDB6;
  pDB6 = m_Address[dbNum]->CastToDB6();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB6[i].swap();
  };
}

void DATABASES::swapSCROLL(int dbNum)
{
  DB7 *pDB7;
  pDB7 = m_Address[dbNum]->CastToDB7();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB7[i].swap();
  };
}

void DATABASES::swapPOTION(int dbNum)
{
  DB8 *pDB8;
  pDB8 = m_Address[dbNum]->CastToDB8();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB8[i].swap();
  };
}

void DATABASES::swapCHEST(int dbNum)
{
  DB9 *pDB9;
  pDB9 = m_Address[dbNum]->CastToDB9();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB9[i].swap();
  };
}

void DATABASES::swapMISC(int dbNum)
{
  DB10 *pDB10;
  pDB10 = m_Address[dbNum]->CastToDB10();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB10[i].swap();
  };
}

void DATABASES::swapEXPOOL(int dbNum)
{
  DB11 *pDB11;
  pDB11 = m_Address[dbNum]->CastToDB11();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB11[i].swap();
  };
}

void DATABASES::swapMISSILE(int dbNum)
{
  DB14 *pDB14;
  pDB14 = m_Address[dbNum]->CastToDB14();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB14[i].swap();
  };
}

void DATABASES::swapCLOUD(int dbNum)
{
  DB15 *pDB15;
  pDB15 = m_Address[dbNum]->CastToDB15();
  for (i32 i=0; i<m_numEnt[dbNum]; i++)
  {
    pDB15[i].swap();
  };
}




void DATABASES::swap(int dbNum)
{
  switch (dbNum)
  {
    // We need to littleEndian some of the words in these
    // miscellaneous databases.
  case dbDOOR:       swapDOOR(dbNum);       break;
  case dbTELEPORTER: swapTELEPORTER(dbNum); break;
  case dbTEXT:       swapTEXT(dbNum);       break;
  case dbACTUATOR:   swapACTUATOR(dbNum);   break;
  case dbMONSTER:    swapMONSTER(dbNum);    break;
  case dbWEAPON:     swapWEAPON(dbNum);     break;
  case dbCLOTHING:   swapCLOTHING(dbNum);   break;
  case dbSCROLL:     swapSCROLL(dbNum);     break;
  case dbPOTION:     swapPOTION(dbNum);     break;
  case dbCHEST:      swapCHEST(dbNum);      break;
  case dbMISC:       swapMISC(dbNum);       break;
  case dbEXPOOL:     swapEXPOOL(dbNum);     break;
  case 12:
  case 13:                             break;
  case dbMISSILE:    swapMISSILE(dbNum);    break;
  case dbCLOUD  :    swapCLOUD(dbNum);      break;
  default: ASSERT(0);
  };
}

void DATABASES::Dump(FILE *f)
{
  DBCOMMON *pCommon;
  DBTYPE dbType;
  ROOMTYPE *pActuatorRoomTypes = NULL;
  int i, n;
  bool title;
  for (dbType = (DBTYPE)0; dbType < 16; dbType = (DBTYPE)(dbType+1))
  {
    n = NumEntry(dbType);
    title = true;
    if (dbType == dbACTUATOR)
    {
      pActuatorRoomTypes = pDoc->DetermineActuatorRoomTypes();
    };
    for (i=0; i<n; i++)
    {
      pCommon = GetCommonAddress(dbType, i);
      if (pCommon->link() == RNnul) continue;
      switch (dbType)
      {
      case dbDOOR:
        pCommon->CastToDB0()->Dump(f, i, title);
        break;
      case dbTEXT:
        pCommon->CastToDB2()->Dump(f, i, title);
        break;
      case dbSCROLL:
        pCommon->CastToDB7()->Dump(f, i, title);
        break;
      case dbTELEPORTER:
        pCommon->CastToDB1()->Dump(f, i, title);
        break;
      case dbMONSTER:
        pCommon->CastToDB4()->Dump(f, i, title);
        break;
      case dbWEAPON:
        pCommon->CastToDB5()->Dump(f, i, title);
        break;
      case dbCLOTHING:
        pCommon->CastToDB6()->Dump(f, i, title);
        break;
      case dbPOTION:
        pCommon->CastToDB8()->Dump(f, i, title);
        break;
      case dbCHEST:
        pCommon->CastToDB9()->Dump(f, i, title);
        break;
      case dbMISC:
        pCommon->CastToDB10()->Dump(f, i, title);
        break;
      case dbMISSILE:
        pCommon->CastToDB14()->Dump(f, i, title);
        break;
      case dbCLOUD:
        pCommon->CastToDB15()->Dump(f, i, title);
        break;
      case dbACTUATOR:
        pCommon->CastToDB3()->Dump(f, i, title, pActuatorRoomTypes[i]);
        break;
      default: NotImplemented();
      };
      title = false;
    };
    if (pActuatorRoomTypes != NULL)
    {
      free (pActuatorRoomTypes);
      pActuatorRoomTypes = NULL;
    };
  };
}

bool DBCOMMON::IsDBType(DBTYPE dbtype)
{
  return db.IsDBType(this, dbtype);
}

DB0 *DBCOMMON::CastToDB0(void)
{
  ASSERT(IsDBType(dbDOOR));
  return (DB0 *)this;
}

DB1 *DBCOMMON::CastToDB1(void)
{
  ASSERT(IsDBType(dbTELEPORTER));
  return (DB1 *)this;
}

DB2 *DBCOMMON::CastToDB2(void)
{
  ASSERT(IsDBType(dbTEXT));
  return (DB2 *)this;
}

DB3 *DBCOMMON::CastToDB3(void)
{
  ASSERT(IsDBType(dbACTUATOR));
  return (DB3 *)this;
}

DB4 *DBCOMMON::CastToDB4(void)
{
  ASSERT(IsDBType(dbMONSTER));
  return (DB4 *)this;
}

DB5 *DBCOMMON::CastToDB5(void)
{
  ASSERT(IsDBType(dbWEAPON));
  return (DB5 *)this;
}

DB6 *DBCOMMON::CastToDB6(void)
{
  ASSERT(IsDBType(dbCLOTHING));
  return (DB6 *)this;
}

DB7 *DBCOMMON::CastToDB7(void)
{
  ASSERT(IsDBType(dbSCROLL));
  return (DB7 *)this;
}

DB8 *DBCOMMON::CastToDB8(void)
{
  ASSERT(IsDBType(dbPOTION));
  return (DB8 *)this;
}

DB9 *DBCOMMON::CastToDB9(void)
{
  ASSERT(IsDBType(dbCHEST));
  return (DB9 *)this;
}

DB10 *DBCOMMON::CastToDB10(void)
{
  ASSERT(IsDBType(dbMISC));
  return (DB10 *)this;
}

DB11 *DBCOMMON::CastToDB11(void)
{
  ASSERT(IsDBType(dbEXPOOL));
  return (DB11 *)this;
}

DB14 *DBCOMMON::CastToDB14(void)
{
  ASSERT(IsDBType(dbMISSILE));
  return (DB14 *)this;
}

DB15 *DBCOMMON::CastToDB15(void)
{
  ASSERT(IsDBType(dbCLOUD));
  return (DB15 *)this;
}

void DBCOMMON::swapLink(void)
{
  m_link.swap();
}

DB0::DB0(void)
{
  InternalError();
}

void DB0::swap(void)
{
  swapLink();
  m_word2 = LE16(m_word2);
}

void DB0::Clear(void)
{
  m_link = RNeof;
  m_word2 = 0x0060; //  Up/down with switch
}

bool DB0::operator != (DB0& db0)
{
  if (m_word2 != db0.m_word2) return true;
  return false;
}

CString DB0::Name(void)
{
  return CString("Door");
}

void DB0::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  //CString result;
  if (pDoc->DoorType(locr.l(),doorType()) >= pDoc->NumDoorTypes())
  {
    result->NewError(locr.Format());
    (*result) += "Door type not defined on this level\n";
  };
  if (ornateness() > pDoc->NumDoorDecorations(locr.l()))
  {
    result->NewError(locr.Format());
    (*result) += "Door decoration not defined on this level\n";
  };
}

void DB0::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n## Database 0 = Doors\n");
    fprintf(f,"#\n#  index  link type ornate mode switch openWithFireball OpenWithAxe\n");

  };
  fprintf(f,"%8d %5d %4d %5d %5d  %s %12s %14s\n",
          index,
          link().ConvertToInteger(),
          doorType(),
          ornateness(),
          mode(),
          doorSwitch()?"yes":"no ",
          canOpenWithFireball()?"yes":"no ",
          canOpenWithAxe()?"yes":"no ");
}

DB1::DB1(void)
{
  InternalError();
}

void DB1::swap(void)
{
  swapLink();
  m_word2 = LE16(m_word2); // x, y, flags
  m_word4 = LE16(m_word4); // level
}

void DB1::Clear(void)
{
  m_link = RNeof;
  m_word2 = 0;
  m_word4 = 0;
}

CString DB1::Name(void)
{
  CString result;
  CString What;
  switch (what())
  {
  case 0: What = "Objects"; break;
  case 1: What = "Monsters"; break;
  case 2: What = "Party/objects"; break;
  case 3: What = "Anything"; break;
  };

  result.Format("Teleporter(%s)",What);
  return result;
}

    bool DB1::operator != (DB1& db1)
{
  if (m_word2 != db1.m_word2) return true;
  if (m_word4 != db1.m_word4) return true;
  return false;
}

void DB1::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  bool legal = true;
  ROOMTYPE rt;
  if (destLevel() >= pDoc->NumLevel()) legal = false;
  else
  {
    if (destX() > pDoc->LastColumn(destLevel())) legal = false;
    if (destY() > pDoc->LastRow(destLevel())) legal = false;
  };
  if (legal)
  {
    rt = RoomType(pDoc->GetCellFlag(destLevel(),destX(),destY()));
    if (rt==roomSTONE)
    {
      if ((what()==2) || (what()==3))
      {
        result->NewError(locr.Format());
        (*result) += "Teleports party into a wall";
        return;
      };
    };
    return;
  }
  else
  {
    result->NewError(locr.Format());
    (*result) += "Bogus Teleporter destination\n";
    return;
  };
}

void DB1::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n## Database 1 = Teleporters\n");
    fprintf(f,"#\n#  index  link\n");

  };
  fprintf(f,"%8d %5d \n",
          index,
          link().ConvertToInteger());
}

void DB1::dest(LOCATIONREL locr)
{
  destX(locr.x());
  destY(locr.y());
  destLevel(locr.l());
}

void DB1::Export(FILE *f, const EXPORTLOC& eloc)
{
  LOCATIONREL target;
  int faceIndex;
  fprintf(f, "  TeleporterRec");
  if (audible()) fprintf(f," audible");
  faceIndex = facingMode() ? 4 : 0;
  faceIndex += rotation();
  if (faceIndex!=0)
        fprintf(f," %s", teleporterFacing[faceIndex]);
  fprintf(f," %s",teleporterWhat[what()]);
  target = dest();
  ExportTarget(f,eloc, target);
  fprintf(f,";\n");
}

LOCATIONREL DB1::dest(void)
{
  LOCATIONREL locr;
  locr.l(destLevel());
  locr.x(destX());
  locr.y(destY());
  locr.p(rotation());
  return locr;
}

void DB1::SwapDestLevel(int L1, int L2)
{
  if (destLevel() == L1)
  {
    destLevel(L2);
    return;
  };
  if (destLevel() == L2)
  {
    destLevel(L1);
    return;
  };
}

DB2::DB2(void)
{
  InternalError();
}

void DB2::swap(void)
{
  swapLink();
  m_word2 = LE16(m_word2); // x, y, flags
}

void DB3::swap(void)
{
  swapLink();
  m_word2 = LE16(m_word2); // x, y, flags
  m_word4 = LE16(m_word4);
  m_word6 = LE16(m_word6);
}

void DB2::Clear(void)
{
  m_link = RNeof;
  m_word2 = 0;
  index(0xffff);
  show(true);
}

static bool IsVisibleWall(LOCATIONREL locr)
{
  ROOMTYPE rt;
  rt = RoomType(locr);
  if (rt == roomSTONE) return true;
  if (rt == roomFALSEWALL)
  {
    CELLFLAG cf;
    cf = pDoc->GetCellFlag(locr);
    if ((cf & 4) ==0) return true;
  };
  return false;
}

void DB2::Verify(LOCATIONREL locr, 
                 CHKERRMSG *result,
                 int containerMask)
{
  int i, len;
  char text[205];
  i = index();
  if (i >= pDoc->NumWordsInTextArray())
  {
    result->NewError("Bogus text index.");
    return;
  };
  //If this is wall text make sure it is not more than 70
  // characters.
  if (    IsVisibleWall(locr)
       && ((containerMask & 
             (1<<dbSCROLL)
           ) == 0))
  {
    len = pDoc->ExpandText(text,this,TEXTTYPE_WALL,200);
    if (len > 68)
    {
      result->NewError(locr.Format());
      (*result) += " Wall Text Too Long";
    };
  };
  return;
}

void DB2::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n##Database 2 = Text\n");
    fprintf(f,"#\n#index  link   textIndex  show\n");
  };
  fprintf(f,"%5d %6d %9d %6s\n",
          index,
          link().ConvertToInteger(),
          this->index(),
          show()?"yes":"no ");
}

DB3::DB3(void)
{
  InternalError();
}

bool DB3::operator != (DB3& db3)
{
  if (m_word2 != db3.m_word2) return true;
  if (m_word4 != db3.m_word4) return true;
  if (m_word6 != db3.m_word6) return true;
  return false;
}

void DB3::Clear(void)
{
  m_link = RNeof;
  m_word2 = 0;
  m_word4 = 0;
  m_word6 = 0;
  m_word8 = 0;
}

void DB3::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  ACTUATORCLASS ac;
  ROOMTYPE rt, rtText;
  LOCATIONREL textLocr;
  textLocr = locr;
  rt = RoomType(locr);
  ac = this->actuatorClass(rt);
  switch (ac)
  {
  case AC_PORTRAIT:
    if (   (rt == roomSTONE)
        && Cast2AC_Portrait(rt)->MightBeActiveAtRuntime())
    {
      switch (locr.p())
      {
      case 0: textLocr.y(textLocr.y()-1); break;
      case 1: textLocr.x(textLocr.x()+1); break;
      case 2: textLocr.y(textLocr.y()+1); break;
      case 3: textLocr.x(textLocr.x()-1); break;
      };
      rtText = RoomType(textLocr);
      if (   (rtText == roomOPEN)
          || (rtText == roomFALSEWALL)
          || (rtText == roomTELEPORTER))
      {
        RN RNtext;
        RNtext = pDoc->FindFirstTextRecord(textLocr);
        if (RNtext != RNeof)
        {
          break;
        };
      };
      {
        CString msg;
        LOCATIONABS loca;
        loca = locr;
        msg.Format("%02d(%02d,%02d)%d Portrait without text",
                  loca.l, loca.x, loca.y, loca.p);
        result->NewError(msg);
        result->Location(locr);
      };
    };
    break;
  };
  return;
}

void DB3::Dump(FILE *f, int index, bool title, ROOMTYPE rt)
{
  ACTUATORCLASS ac;
  if (title)
  {
    fprintf(f,"#\n##Database 3 = Actuators\n");
    fprintf(f,"#\n# index link type     Options\n");
  };
  ac = actuatorClass(rt);
  fprintf(f,"%6d %5d <%s> ",
            index,
            link().ConvertToInteger(),
            this->Name(rt));
  switch (ac)
  {
  case AC_ENDGAME:
    this->Cast2AC_EndGame(rt)->Dump(f);
    break;
  case AC_ANDOR:
    this->Cast2AC_AndOr(rt)->Dump(f);
    break;
  case AC_PRESSUREPAD:
    this->Cast2AC_PressurePad(rt)->Dump(f);
    break;
  case AC_SWAPOBJECT:
    this->Cast2AC_SwapObject(rt)->Dump(f);
    break;
  case AC_COUNTER:
    this->Cast2AC_Counter(rt)->Dump(f);
    break;
  case AC_LAUNCHER:
    this->Cast2AC_Launcher(rt)->Dump(f);
    break;
  case AC_PUSHBUTTON:
    this->Cast2AC_PushButton(rt)->Dump(f);
    break;
  case AC_MONSTERGENERATOR:
    this->Cast2AC_MonsterGenerator(rt)->Dump(f);
    break;
  case AC_PORTRAIT:
    this->Cast2AC_Portrait(rt)->Dump(f);
    break;
  case AC_OBJECTHOLDER:
    this->Cast2AC_ObjectHolder(rt)->Dump(f);
    break;
  case AC_DISABLED:
    this->Cast2AC_Disabled(rt)->Dump(f);
    break;
  case AC_DSA:
    this->Cast2AC_DSA(rt)->Dump(f);
    break;
  default:
    NotImplemented();
    break;
  };
}

void DB4::swap(void)
{
  swapLink();
  possession2.swap();
  hitPoints6[0]  = LE16(hitPoints6[0]);
  hitPoints6[1]  = LE16(hitPoints6[1]);
  hitPoints6[2]  = LE16(hitPoints6[2]);
  hitPoints6[3]  = LE16(hitPoints6[3]);
  word14 = LE16(word14);
}

void DB4::Clear(void)
{
  MONSTERTYPE mt;
  mt = "Screamer";
  hitPoints6[0] = 1;
  hitPoints6[1] = 1;
  hitPoints6[2] = 1;
  hitPoints6[3] = 1;
  monsterType(mt);
  possession(RNeof);
  numMonster(1);
  w14b10(0);
  facing(0);
  groupIndex(0xe4); //Put monster in center
}

bool DB4::operator != (DB4& db4)
{
  if (possession2 == RNeof)
  {
    if (db4.possession2 != RNeof) return true;
  }
  else
  {
    if (db4.possession2 == RNeof) return true;
    else if (!possession2.IsSameObjectAs(db4.possession2)) return true;
  };
  if (uByte4 != db4.uByte4) return true;
  if (uByte5 != db4.uByte5) return true;
  if (hitPoints6[0] != db4.hitPoints6[0]) return true;
  if (hitPoints6[1] != db4.hitPoints6[1]) return true;
  if (hitPoints6[2] != db4.hitPoints6[2]) return true;
  if (hitPoints6[3] != db4.hitPoints6[3]) return true;
  if (word14 != db4.word14) return true;
  return false;
}

void DB4::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  if (  (locr.l() != -1)
      &&(RoomType(pDoc->GetCellFlag(locr)) == roomSTONE))
  {
    result->NewError(locr.Format());
    (*result) += "Monster encased in stone?";
  };
  return;
}

void DB4::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n## Database 4 = Monsters\n");
    fprintf(f,"#\n#  index  link type\n");

  };
  fprintf(f,"%8d %5d %4d\n",
          index,
          link().ConvertToInteger(),
          monsterType());
}

void DB4::Export(FILE *f, const LOCATIONREL& /*locr */)
{
  MONSTERTYPE mt;
  mt = this->monsterType();
  fprintf(f,"  MonsterRec num=%d", numMonster());
  fprintf(f," type=%s",mt.Name());
  fprintf(f," %s", teleporterFacing[facing()+4]);
  if (Invisible()) fprintf(f," Invisible");
  if (DrawAsSize4()) fprintf(f," DrawAsSize4");
  fprintf(f," HP=%d,%d,%d,%d",
              hitPoints(0),
              hitPoints(1),
              hitPoints(2),
              hitPoints(3));
  fprintf(f, " Position=%d,%d,%d,%d",
               positions() >> 0 &3,
               positions() >> 2 &3,
               positions() >> 4 &3,
               positions() >> 6 &3);
  fprintf(f,";\n");
}

DB5::DB5(void)
{
  InternalError();
}

void DB5::swap(void)
{
  swapLink();
  word2  = LE16(word2);
}

void DB5::Clear(void)
{
  m_link = RNeof;
  word2 = 0;
  important(true);
}

bool DB5::operator != (DB5& db5)
{
  if (word2 != db5.word2) return true;
  return false;
}

void DB5::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  if (    !important() 
       && ((pDoc->EditingOptions()&EDIT_OPT_AllowUnimportantObjects)==0)
     )
  {
    result->NewError(locr.Format());
    (*result) += "Weapon is not important?";
  };
}

void DB5::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n## Database 5 = Weapons\n");
    fprintf(f,"#\n#  index  link type\n");

  };
  fprintf(f,"%8d %5d %4d\n",
          index,
          link().ConvertToInteger(),
          weaponType());
}

void DB5::Export(FILE *f, const LOCATIONREL& /* locr */)
{
  WEAPONTYPE wt;
  wt = this->weaponType();
  fprintf(f,    "  Weapon = %s\n", wt.Name());
  fprintf(f,    "  EndWeapon\n");
}

DB6::DB6(void)
{
  InternalError();
}

void DB6::swap(void)
{
  swapLink();
  word2  = LE16(word2);
}

void DB6::Clear(void)
{
  m_link = RNeof;
  word2 = 0;
  important(true);
}


void DB6::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  if (    !important()
       && ((pDoc->EditingOptions()&EDIT_OPT_AllowUnimportantObjects)==0)
     )
  {
    result->NewError(locr.Format());
    (*result) += "Clothing is not important?";
  };
}

void DB6::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n## Database 6 = Clothes\n");
    fprintf(f,"#\n#  index  link type\n");

  };
  fprintf(f,"%8d %5d %4d\n",
          index,
          link().ConvertToInteger(),
          clothingType());
}

void DB7::swap(void)
{
  swapLink();
  m_RNtext.LittleEndian();
  m_word4 = LE16(m_word4);
}

void DB7::Clear(void)
{
  RN RNtext;
  m_word4 = 0;
  RNtext = pDoc->FindEmptyDB2Entry(TEXTTYPE_SCROLL);
  text(RNtext);
  open(false);
}

void DB7::Verify(LOCATIONREL locr, 
                 CHKERRMSG *result,
                 int containerMask)
{
  RN RNtext;
  DB2 *pText;
  RNtext = text();
  if (RNtext.dbType() != dbTEXT)
  {
    result->NewError(locr.Format());
    (*result) += "A Scroll contains something other than text";
    result->ErrorType(CEMT_SCROLLNONTEXT);
    result->DBCommonAddress(this);
    return;
  };
  //RNtext.pos(locr.p);
  //pDoc->MarkObjectReferenced(RNtext);
  pText = pDoc->GetRecordAddressDB2(RNtext);
  pText->Verify(locr, result, containerMask|(1<<dbSCROLL));
}

void DB7::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n##Database 7 = Scrolls\n");
    fprintf(f,"#\n# index  link  text  open\n");
  };
  fprintf(f,"%7d %5d %5d  %s\n",
            index,
            link().ConvertToInteger(),
            text().ConvertToInteger(),
            open()?"yes":"no ");
}

DB8::DB8(void)
{
  InternalError();
}

void DB8::swap(void)
{
  swapLink();
  m_word2  = LE16(m_word2);
}

void DB8::Clear(void)
{
  m_word2 = 0;
  important(true);
}

void DB8::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  POTIONTYPE pt;
  pt = potionType();
  if (!pt.IsValid())
  {
    result->NewError(locr.Format());
    (*result) += " Illegal ",
    (*result) += pt.Name();
    return;
  };
  if (    !important()
       && ((pDoc->EditingOptions()&EDIT_OPT_AllowUnimportantObjects)==0)
     )
  {
    result->NewError(locr.Format());
    (*result) += " Potion is not important?";
  };
}

void DB8::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n## Database 8 = Potions\n");
    fprintf(f,"#\n#  index  link type\n");

  };
  fprintf(f,"%8d %5d %4d\n",
          index,
          link().ConvertToInteger(),
          potionType());
}

void DB8::Export(FILE *f, const LOCATIONREL& /* locr */)
{
  POTIONTYPE pt;
  fprintf(f, "  Potion ");
  pt = this->potionType();
  fprintf(f, "potion=%s", pt.Name());
  if (important()) fprintf(f, " important");
  fprintf(f," strength=%d;\n", strength());
  fprintf(f, ";\n");
}

void DB9::swap(void)
{
  swapLink();
  m_contents.swap();
  m_word4 = LE16(m_word4);
}

void DB9::Clear(void)
{
  m_link = RNeof;
  m_contents = RNeof;
  m_word4 = 0;
  m_uByte6[0] = 0;
  m_uByte6[1] = 0;
}

void DB9::Verify(LOCATIONREL /*locr*/, CHKERRMSG * /*result*/)
{
  //RN RNcontents;
  //for (RNcontents=contents(); 
  //     RNcontents!=RNeof; 
  //     RNcontents=pDoc->GetDBRecordLink(RNcontents))
  //{
  //  pDoc->MarkObjectReferenced(RNcontents);
  //};
}

bool DB9::operator != (DB9& db9)
{
  if ((m_contents==RNeof)&&(db9.m_contents==RNeof)) return false;
  if (m_contents==RNeof) return true;
  if (db9.m_contents==RNeof) return true;
  if (m_contents.IsSameObjectAs(db9.m_contents)) return true;
  if (m_word4 != db9.m_word4) return true;
  if (m_uByte6[0] != db9.m_uByte6[0]) return true;
  if (m_uByte6[1] != db9.m_uByte6[1]) return true;
  return false;
}

void DB9::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n## Database 9 = Chests\n");
    fprintf(f,"#\n#  index  link contents\n");

  };
  fprintf(f,"%8d %5d %4d\n",
          index,
          link().ConvertToInteger(),
          contents().ConvertToInteger());
}

/*
DB9& DB9::operator = (DB9& db9)
{
  if (db9.m_contents != RNeof)
  {
    m_contents = pDoc->Copy(db9.m_contents);
  };
  m_word4 = db9.m_word4;
  m_uByte6[0] = db9.m_uByte6[0];
  m_uByte6[1] = db9.m_uByte6[1];
  return *this;
}
*/

DB10::DB10(void)
{
  InternalError();
}

void DB10::swap(void)
{
  swapLink();
  m_word2  = LE16(m_word2);
}
/*
DB10& DB10::operator = (DB10& db10)
{
  m_word2 = db10.m_word2;
  return *this;
}
*/
void DB10::Clear(void)
{
  m_link = RNeof;
  m_word2 = 0;
  important(true);
}

void DB10::Verify(LOCATIONREL locr, CHKERRMSG *result)
{
  if (    !important()
       && ((pDoc->EditingOptions()&EDIT_OPT_AllowUnimportantObjects)==0)
     )
  {
    result->NewError(locr.Format());
    (*result) += "Miscellaneous item is not important?";
  };
}

bool DB10::operator != (DB10& db10)
{
  if (m_word2 != db10.m_word2) return true;
  return false;
}

void DB10::Dump(FILE *f, int index, bool title)
{
  if (title)
  {
    fprintf(f,"#\n##Database 10 - Miscellaneous food, etc.");
    fprintf(f,"#\n#index  link  type  important cursed value\n");
  };
  fprintf(f,"%6d %5d %5d     %s     %s  %5d\n",
            index,
            link().ConvertToInteger(),
            miscType(),
            important()?"yes":"no ",
            cursed()?"yes":"no ",
            value());
}

void DB11::swap(void)
{
  int i;
  swapLink();
  size = LE16(size);
  for (i=0; i<63; i++)
  {
    d[i] = LE32(d[i]);
  };
}

void DB14::swap(void)
{
  swapLink();
  m_obj2.swap();
  m_timerIndex  = LE16(m_timerIndex);
}

void DB14::Clear(void)
{
  m_link = RNeof;
  m_obj2 = RNeof;
  m_energyRemaining = 0;
  m_energyRemaining2 = 0;
  m_timerIndex = 0;
}

void DB14::Verify(LOCATIONREL /*locr*/, CHKERRMSG * /*result*/)
{
}

void DB14::Dump(FILE * /*f*/, int /*index*/, bool title)
{
  if (title)
  {
    NotImplemented();
  };
}

void DB15::swap(void)
{
  swapLink();
  m_word2  = LE16(m_word2);
}

void DB15::Clear(void)
{
  m_link = RNeof;
  m_word2 = 0;
}

void DB15::Verify(LOCATIONREL /*locr*/, CHKERRMSG * /*result*/)
{
}

void DB15::Dump(FILE * /*f*/, int /*index*/, bool title)
{
  if (title)
  {
    NotImplemented();
  };
}


EXPOOL expool;


EXPOOL::EXPOOL(void)
{
  m = NULL;
}

EXPOOL::~EXPOOL(void)
{
}

void EXPOOL::Setup(void)
{
  if (db.NumEntry(dbEXPOOL) == 0)
  {
    m = (ui32 *)malloc(256, MALLOC094);
    memset(m, 0 , 256);
    db.SetAddress(dbEXPOOL, (DBCOMMON *)m);
    ((DB11 *)m)->link(RNeof);
    ((DB11 *)m)->size = 0;
    db.NumEntry(dbEXPOOL, 1);
  }
  else
  {
    m = (ui32 *)db.GetCommonAddress(dbEXPOOL, 0);
  };
}

void EXPOOL::Write(ui32 key, ui32 *record, ui32 size)
{
  ui32 hash = key * 0xbb40e62d;
  ui32 hashi = 32 + (hash >> 27);
  ui32 result;
  ASSERT(size <= 29);
  if (m[size+2] == 0) enlarge(size+2);
  result = m[size+2];
  m[size+2] = m[result];
  if ((m[hashi] & 0x80000000) != 0)
  {
    hashi = (m[hashi] & 0x7fffffff) + ((hash >> 21) & 0x3f);
  };
  m[result] = m[hashi];
  m[hashi] = result;
  m[result+1] = key;
  memcpy(m+result+2, record, 4*size);
}

i32 EXPOOL::Read(ui32 key, ui32 *record, ui32 max)
{ //Returns size of record (or zero)
  ui32 hash = key * 0xbb40e62d;
  ui32 hashi = 32 + (hash >> 27);
  ui32 p, size, len;
  if ((m[hashi] & 0x80000000) != 0)
  {
    hashi = (m[hashi] & 0x7fffffff) + ((hash >> 21) & 0x3f);
  };
  while (m[hashi] != 0)
  {
    p = m[hashi];
    if (m[p+1] == key)
    {
      size = ((DB11 *)(m+(p & 0xffffffc0)))->size;
      len = size-2;
      if (len > max) len = max;
      if (len > 0) memcpy(record, m+p+2, 4*(len));
      m[hashi] = m[p];
      m[p] = m[size];
      m[size] = p;
      return size-2;
    };
    hashi = m[hashi];
  };
  return -1;
}

i32 EXPOOL::Locate(ui32 key, ui32 **record)
{ //Returns size of record (or -1)
  ui32 hash = key * 0xbb40e62d;
  ui32 hashi = 32 + (hash >> 27);
  ui32 p, size;
  if ((m[hashi] & 0x80000000) != 0)
  {
    hashi = (m[hashi] & 0x7fffffff) + ((hash >> 21) & 0x3f);
  };
  while (m[hashi] != 0)
  {
    p = m[hashi];
    if (m[p+1] == key)
    {
      size = ((DB11 *)(m+(p & 0xffffffc0)))->size;
      *record = m+p+2;
      //memcpy(record, m+p+2, 4*(size-2));
      //m[hashi] = m[p];
      //m[p] = m[size];
      //m[size] = p;
      return size-2;
    };
    hashi = m[hashi];
  };
  return -1;
}

void EXPOOL::enlarge(ui32 size)
{
  ui32 n, e, p, i;
  ui32 numBlock;
  DB11 *pDB11;
  numBlock = db.NumEntry(dbEXPOOL);
  numBlock++;
  m = (ui32 *)realloc(m, 256*numBlock, MALLOC095);
  p = 64*(numBlock-1) + 1;
  e = p+64;
  n = 63 / size;
  pDB11 = (DB11 *)m + numBlock-1;
  pDB11->link(RNeof);
  pDB11->Size(size);
  for (i=0; i<n; i++, p+=size)
  {
    m[p] = p+size;
  };
  m[p-size] = m[size];
  m[size] = 64*(numBlock-1) + 1;
  db.NumEntry(dbEXPOOL, numBlock);
  db.SetAddress(dbEXPOOL, (DBCOMMON *)m);
}

bool GetExtendedCellFlag(
       i32 level, i32 x, i32 y, EXPOOL_DATA_TYPE ftype)
{
  ui32 *record;
  i32 len;
  len = expool.Locate(
                (EDT_ExtendedCellFlags<<24) + (level<<5) + x, 
                &record);
  if (len < 8) return false;
  return (record[ftype] & (1<<y)) != 0;
}



void DUNGEONDATINDEX::Swap(void)
{
  int i;
  if (pDoc->m_littleEndian)
  {
    sentinel = (sentinel);
    word2 = (word2);
    word4 = LE16(word4);
    word6 = (word6);
    word8 = (word8);
    word10 = (word10);
    for (i=0; i<16; i++) dbSizes[i] = (dbSizes[i]);
  }
  else
  {
    sentinel = LE16(sentinel);
    word2 = LE16(word2);
    word4 = LE16(word4);
    word6 = LE16(word6);
    word8 = LE16(word8);
    word10 = LE16(word10);
    for (i=0; i<16; i++) dbSizes[i] = LE16(dbSizes[i]);
  };
}

void CCSBuildDoc::DumpObjectIndirectIndexTable(FILE *f)
{
  int i;
  fprintf(f,"#\n##ObjectIndirectReferenceTable\n");
  fprintf(f,"#\n# index  db#  pos dbIndex \n");
  for (i=1; i<dataMapLength(); i++)
  {
    if (this->m_dataIndexMap[i] == 0xffff) continue;
    fprintf(f,"%5d     %2d    %d   %5d\n",
              i,
              m_dataTypeMap[i] & 15,
              (m_dataTypeMap[i] >> 4) & 3,
              m_dataIndexMap[i]);
  };
}

void CCSBuildDoc::DumpDatabases(FILE *f)
{
  db.Dump(f);
}

void CCSBuildDoc::DumpLevelInfo(FILE *f)
{
  int level, i, m;
  fprintf(f,"#\n##Level Information\n");
  fprintf(f,"#\n#level width height x0 y0 numRanWall numRanFloor ExpMult doora doorb\n");
  for (level=0; level<NumLevel(); level++)
  {
    fprintf(f," %5d %5d %5d %2d %2d %5d %11d %9d %7d %5d\n",
              level,
              LastColumn(level),
              LastRow(level),
              OffsetX(level),
              OffsetY(level),
              NumRandomWallDecorations(level),
              NumRandomFloorDecorations(level),
              ExperienceMultiplier(level),
              DoorType(level, 0),
              DoorType(level, 1));
  };
  fprintf(f,"#\n##Available Monsters\n");
  fprintf(f,"#\n#Level     Monster List\n");
  for (level=0; level<NumLevel(); level++)
  {
    fprintf(f,"%3d         ", level);
    for (i=1; i<16; i++)
    {
      if (i>NumMonsterType(level)) m = -1;
      else m = MonsterType(level,i-1).Integer();
      fprintf(f,"%3d", m);
    };
    fprintf(f,"\n");
  };
  fprintf(f,"#\n##Available Wall Decorations\n");
  fprintf(f,"#\n#Level     Wall Decoration List\n");
  for (level=0; level<NumLevel(); level++)
  {
    fprintf(f,"%3d         ", level);
    for (i=1; i<16; i++)
    {
      if (i>NumWallDecorations(level)) m = -1;
      else m = WallDecorationType(level,i-1).Integer();
      fprintf(f,"%3d ", m);
    };
    fprintf(f,"\n");
  };
  fprintf(f,"#\n##Available Floor Decorations\n");
  fprintf(f,"#\n#Level     Floor Decoration List\n");
  for (level=0; level<NumLevel(); level++)
  {
    fprintf(f,"%3d         ", level);
    for (i=1; i<16; i++)
    {
      if (i>NumFloorDecorations(level)) m = -1;
      else m = FloorDecorationType(level,i).Integer();
      fprintf(f,"%3d", m);
    };
    fprintf(f,"\n");
  };
  fprintf(f,"#\n##Available Door Decorations\n");
  fprintf(f,"#\n#Level     Door Decoration List\n");
  for (level=0; level<NumLevel(); level++)
  {
    fprintf(f,"%3d         ", level);
    for (i=1; i<16; i++)
    {
      if (i>NumDoorDecorations(level)) m = -1;
      else m = DoorDecorationType(level,i-1).Integer();
      fprintf(f,"%3d", m);
    };
    fprintf(f,"\n");
  };
  fprintf(f,"#\n##Available DSAs\n");
  fprintf(f,"#\n#Level     DSA List\n");
  for (level=0; level<NumLevel(); level++)
  {
    fprintf(f,"%3d     ", level);
    for (i=0; i<32; i++)
    {
      m = m_DSALevelIndex[level][i];
      if (m == 65535) m = -1;
      fprintf(f," %3d", m);
    };
    fprintf(f,"\n");
  };
}

void CCSBuildDoc::DumpRawText(FILE *f)
{
  int i, compressedTextIndex, nibble;
  char c;
  ui16 *pWord, word;
  fprintf(f,"#\n##Raw Text\n");
  fprintf(f,"#\n#Index  text\n");
  for (i=0; i<NumWordsInTextArray(); i++)
  {
    compressedTextIndex = pDoc->CompressedTextIndex(i);
    if (compressedTextIndex == -1) continue;
    if (!pDoc->IsValidCompressedTextIndex(compressedTextIndex)) continue;
    fprintf(f,"%5d  \"",i);
    pWord = m_compressedText + compressedTextIndex;
    do
    { 
      c = 0;
      for (word=LE16(*(pWord++)),nibble=0; 
          (nibble<3) && (c!=31); 
          nibble++, word<<=5)
      {
        c = (char)((word>>10) & 0x1f);
        switch(c)
        {
        case 31: break;
        case 26: fprintf(f,"%c",' '); break;
        default: fprintf(f,"%c",c+'A'); break;
        };
      };
    } while (c != 31);
    fprintf(f,"\"\n");
  };
}


SKIN_CACHE skinCache;


i32 SKIN_CACHE::Load(i32 level, i32 x)
{
  ui32 *pExpoolRecord;
  i32 i, column, recId, recSize;
  if (level != m_level) 
  {
    m_level = level;
    for (i=0; i<16; i++)
    {
      m_size[i] = -1;
    };
  };
  column = x / 2;
  if (m_size[column] == 0) return column;
  if (m_size[column] == -1)
  {
    recId = (EDT_Skins << 24) + (level << 4) + (x / 2);
    recSize = 4* expool.Locate(recId, &pExpoolRecord);
    if (recSize <= 0) recSize = 0;
    m_size[column] = (i8)recSize;
    if (m_size[column] == 0) return column;
    if (m_size[column] > 64) m_size[column] = 64;
    memcpy(m_columns[column], pExpoolRecord, m_size[column]);
  };   
  return column;
}

ui8 SKIN_CACHE::GetSkin(i32 level, i32 x, i32 y)
{
  i32 column, index;
  column = Load(level, x);
  if (m_size[column] == 0) return 0;
  index = (2 * y) + (x & 1);
  if (index >= m_size[column]) return 0;
  return m_columns[column][index];
}

ui8 SKIN_CACHE::GetSkin(const LOCATIONREL& locr)
{
  return GetSkin(locr.l(), locr.x(), locr.y());
}

void SKIN_CACHE::SetSkin(i32 level, i32 x, i32 y, i8 skinNum)
{
  i32 i, column, index, recId, recSize;
  column = Load(level, x);
  index = (2 * y) + (x & 1);
  if ( index < m_size[column])
  {
    if (skinNum == m_columns[column][index]) return;
  }
  else
  {
    if (skinNum == 0) return;
  };
  pDoc->SetModifiedFlag();
  recId = (EDT_Skins << 24) + (level << 4) + (x / 2);
  recSize = 4 * expool.Read(recId, (ui32 *)m_columns[column], 64); // read and delete old record
  if (recSize < 0) recSize = 0;
  if (recSize > 64) recSize = 64;
  if (recSize < 64) memset(m_columns[column]+recSize, 0, 64-recSize);
  m_columns[column][index] = skinNum;
  for (i=63; i>=0; i--) 
  {
    if (m_columns[column][i] != 0) break;
  };
  if (i < 0) return;
  expool.Write(recId, (ui32 *)m_columns[column], (i+4)/4);
  m_size[column] = (i8)(i+1);
}













ui8 SKIN_CACHE::GetDefaultSkin(i32 level)
{
  ui32 *pExpoolRecord;
  ui32 recId;
  i32 recSize;
  recId = (EDT_Skins << 24) + 0x800000;
  recSize = 4 * expool.Locate(recId, &pExpoolRecord);
  if (recSize <= level) return 0;
  return ((ui8 *)pExpoolRecord)[level];
}

void SKIN_CACHE::SetDefaultSkin(i32 level, ui8 skinNum)
{
  ui8 expoolRecord[64];
  ui32 recId;
  i32 recSize;
  memset(expoolRecord,0,64);
  recId = (EDT_Skins << 24) + 0x800000;
  recSize = 4 * expool.Read(recId, (ui32 *)expoolRecord, 16); // read and delete old record
  if (expoolRecord[level] != skinNum)
  {
    expoolRecord[level] = skinNum;
    pDoc->SetModifiedFlag();
  };
  expool.Write(recId, (ui32 *)expoolRecord, 16);
}






